# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008, 2014 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls JMD.
#
# mpuls JMD is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls JMD is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls JMD. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Torsten Irländer <torsten.irlaender@intevation.de>
# Sascha L. Teichmann <teichmann@intevation.de>
#

import datetime
import logging
from re import match

import formencode
from formencode.validators import FormValidator, Int

from pylons import url
from pylons.controllers.util import redirect, abort

from mpulsweb.lib.base import c, render, request, render_htmlfill
from mpulsweb.lib.security import checkRole
from mpulsweb.lib.db import db
from mpulsweb.lib.validators import BaseFormValidator

from mpulsweb.controllers.case import CaseController as MpulsCaseController

from jmdstrukturweb.model.case import find_unused_years, year_used
from formed.meta.data import ChoiceNode, BoolLeaf, IntLeaf
from formed.meta import UNKNOWN_INT

log = logging.getLogger(__name__)


class UnusedYear(FormValidator):

    messages = {
        "alreadyused": u'Für das Jahr existiert bereits eine Statistik.'
        }

    def _to_python(self, value, state):
        return int(value)

    def validate_python(self, value, state):
        with db.cursor() as cur:
            if year_used(cur, value):
                raise formencode.Invalid(self.message("alreadyused", state),
                                         value, state)



class CaseController(MpulsCaseController):

    def _get_validator(self):
        '''Set specific validator'''
        return BaseFormValidator(year=UnusedYear())

    @checkRole('cm')
    def new2(self):
        return self._year_form('/casemanagement/new.mako')

    def _year_form(self, template, form_errors=None, form_defaults=None):
        this_year = datetime.date.today().year
        with db.cursor() as cur:
            c.free_years = find_unused_years(cur, this_year - 1, this_year + 3)
        if form_errors is None:
            form_errors = {}
        if form_defaults is None:
            form_defaults = {}
        if c.free_years:
            form_defaults["year"] = c.free_years[0]
        form = render(template)
        return render_htmlfill(form, defaults=form_defaults, errors=form_errors,
                               auto_insert_errors=False)

    def _year_dates(self, year):
        """Return a dict with the start/end dates for a new case.
        The dictionary can be passed directly to an instance tree's
        setData method. Therefore the values are strings.
        """
        return dict(datum_begin_zeitraum="1.1.%s" % year,
                    datum_ende_zeitraum="31.12.%s" % year)

    def _createCase(self, form_result):
        """Convert the form_result and pass it to the base class method.
        """
        year = form_result["year"]
        return super(CaseController, self)._createCase(self._year_dates(year))

    @checkRole('cm')
    def copy(self, id):
        return self._year_form('/casemanagement/copy.mako',
                               form_defaults=dict(case_id=self._checkInt(id)))

    @checkRole('cm')
    def copyAction(self):
        validator = BaseFormValidator(case_id=Int(),
                                      year=UnusedYear())
        try:
            form_result = validator.to_python(request.params)
        except formencode.Invalid, error:
            log.debug("Invalid form data: %r", error)
            return self._year_form('/casemanagement/copy.mako',
                                   form_defaults = error.value,
                                   form_errors = error.error_dict)

        case = self._loadCase(form_result["case_id"], read_only=True)
        new_values_dict = self._year_dates(form_result["year"])
        new_values_dict['eingabe_vollstaendig'] = ''
        new_values_dict['phase'] = '0'
        new_values_dict['entstand_aus_kopie'] = '1'
        new_it = case.copy(new_values=new_values_dict)
        self._clean_invisible(new_it)
        self._clean_unkown_int(new_it)
        self._clean_checked_copy_fields(new_it)
        redirect(url(controller="case", action="select",
                     id=new_it.getRootNode().getIdentifier()))

    def _clean_checked_copy_fields(self, new_it):
        change = False
        for leaf in self._walk_tree_for_leafs(new_it.getRootNode()):
            if match('kopie_geprueft_p',leaf.getIdentifier()):
                new_it.setData({leaf.getIdentifier(): ''})
                change = True
        if change:
            new_it.commit()

    def _clean_invisible(self, new_it):
        choice_leafs = []
        for leaf in self._walk_tree_for_leafs(new_it.getRootNode()):
            (choice, inv) = self._check_leaf(leaf)
            if choice:
                choice_leafs.append((leaf, inv))

        change = False
        for (leaf, inv) in choice_leafs:
            change = self._clean_field(new_it, leaf, inv) or change
        if change:
            new_it.commit()

    def _clean_field(self, new_it, leaf, inlist):
        change = False
        if leaf.getValue() in inlist:
            new_it.setData({leaf.getIdentifier(): "-1"})
            change = True
        return change

    def _walk_tree_for_leafs(self, node):
        res = []
        if node.isLeaf():
            res.append(node)
        else:
            for child in node.getChildren():
                res.extend(self._walk_tree_for_leafs(child))
        return res

    def _check_leaf(self, leaf):
        meta = leaf.getMeta()
        if isinstance(meta,ChoiceNode):
            return (True, self._check_invisible(meta))
        return (False, [])

    def _check_invisible(self, meta):
        inv_list = []
        for walker in meta.walk():
            if isinstance(walker, BoolLeaf) and walker.isInvisible():
                inv_list.append(walker.getValue())
        return inv_list

    def _clean_unkown_int(self, new_it):
        int_leafs = []
        for leaf in self._walk_tree_for_leafs(new_it.getRootNode()):
            meta = leaf.getMeta()
            if isinstance(meta, IntLeaf):
                if leaf.getValue() == UNKNOWN_INT:
                    int_leafs.append(leaf)
        for leaf in int_leafs:
            new_it.setData({leaf.getIdentifier(): ''})
        if int_leafs:
            log.debug("at copy clean int %s", " ".join(
                     [leaf.getIdentifier() for leaf in int_leafs]))
            new_it.commit()

    @checkRole('cm')
    def markForDelete(self, id, confirmed):
        case = self._loadCase(id)
        if case.getPhase() == 1:
             abort(401)
        return super(CaseController, self).markForDelete(id, confirmed)

    @checkRole('admin')
    def delete(self, id, confirmed):
        case = self._loadCase(id)
        if case.getPhase() == 1:
             abort(401)
        return super(CaseController, self).delete(id, confirmed)

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
