## -*- coding: utf-8 -*-
# Copyright 2007, 2008, 2014 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls JMD.
#
# mpuls JMD is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls JMD is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls JMD. If not, see <http://www.gnu.org/licenses/>.
#

import logging
import lxml
import datetime
import formencode

from pylons import response, app_globals as g

from chosoutil import expand_odt, set_meta_odt, get_doc_file

from libmpuls.evaluation.export import EvaluationExportXML
from libmpuls.evaluation.evaluation import EvaluationSet
from libmpuls.evaluation.config import EvaluationConfig

from mpulsweb.lib.base import request
from mpulsweb.lib.base import c, render, render_htmlfill
from mpulsweb.lib.base import BaseController
from mpulsweb.lib.security import checkRole
from mpulsweb.lib.db import db
from mpulsweb.model.agencysettings import Agency
from mpulsweb.controllers.evaluate import get_configfile
from mpulsweb.controllers.evaluate import get_phase_filter_for_evaluation

from jmdstrukturweb.controllers.evaluate import EvaluateController
from jmdweb.model.struktur import get_db


log = logging.getLogger(__name__)


class ReportController(BaseController):

    @checkRole('cm')
    def doc(self):
        year = datetime.datetime.now().year
        c.years = range(year - 2, year + 1)
        form_defaults = {'year': year}
        form_errors = {}

        form = render('report/report.mako')
        return render_htmlfill(form, defaults=form_defaults, errors=form_errors,
                               auto_insert_errors=False)

    @checkRole('cm')
    def exportAction(self):
        '''
        Returns a DOC file to be filed as a year report by the jmd agencies
        '''

        params = formencode.variabledecode.variable_decode(request.params)
        year = int(params['year'])

        # Gather the needed data from the evaluations
        template_vars = {}

        #
        # Struktur- und Personaldaten
        #
        eval_structure = self._get_evaluation(1, year=year)
        root = lxml.etree.fromstring(eval_structure)

        for element in root.iter("evaluation"):
            element_id = element.get("id")

            # Hauptamtliche, Stellentanteile, Gesamt
            if element_id == u'5':
                template_vars['II-1a-g'] = "%0.2f" % float(self._set_single_table_value(element))

            # Hauptamtliche - mit Migrationshintergrund
            if element_id == u'13':
                template_vars['II-1-dmg'] = self._set_single_table_value(element)

            # Hauptamtliche - gesamt
            if element_id == u'10':
                template_vars['II-1b-g'] = self._set_single_table_value(element)

            # Hauptamtliche - maennlich
            if element_id == u'11':
                template_vars['II-1b-m'] = self._set_single_table_value(element)

            # Hauptamtliche - weiblich
            if element_id == u'12':
                template_vars['II-1b-w'] = self._set_single_table_value(element)

            # Honorarkraefte, Stellenanteile, Gesamt
            if element_id == u'34':
                template_vars['II-2a-g'] = "%0.2f" % float(self._set_single_table_value(element))

            # Honorarkraefte, Personen, Gesamt
            if element_id == u'6':
                template_vars['II-2b-g'] = self._set_single_table_value(element)

            # Honorarkraefte, Personen, weiblich
            if element_id == u'15':
                template_vars['II-2b-w'] = self._set_single_table_value(element)

            # Honorarkraefte, Personen, maennlich
            if element_id == u'14':
                template_vars['II-2b-m'] = self._set_single_table_value(element)

            # Honorarkraefte, Personen, mit Migrationshintergrund
            if element_id == u'16':
                template_vars['II-2b-dmg'] = self._set_single_table_value(element)

            # Eherenamtliche, Personen, Gesamt
            if element_id == u'7':
                template_vars['II-3-g'] = self._set_single_table_value(element)

            # Eherenamtliche, Personen, weiblich
            if element_id == u'18':
                template_vars['II-3-w'] = self._set_single_table_value(element)

            # Eherenamtliche, Personen, männlich
            if element_id == u'17':
                template_vars['II-3-m'] = self._set_single_table_value(element)

            # Eherenamtliche, Personen, Migration Gesamt
            if element_id == u'19':
                template_vars['II-3-mg'] = self._set_single_table_value(element)


            #
            # Prozentuale Aufteilung der Arbeit - sonstiges
            #

            # Prozentuale Aufteilung der Arbeit - Individuelle Begleitung des jungen Menschen im CM
            if element_id == u'8':
                template_vars['PVA-CM'] = self._set_single_table_value(element)

            # Prozentuale Aufteilung der Arbeit - Gruppenangebote
            if element_id == u'22':
                template_vars['PVA-GA'] = self._set_single_table_value(element)

            # Prozentuale Aufteilung der Arbeit - Netzwerkarbeit
            if element_id == u'23':
                template_vars['PVA-NA'] = self._set_single_table_value(element)

            # Prozentuale Aufteilung der Arbeit - Dokumentation
            if element_id == u'26':
                template_vars['PVA-MVD'] = self._set_single_table_value(element)

            # Prozentuale Aufteilung der Arbeit - Beratung
            if element_id == u'31':
                template_vars['PVA-B'] = self._set_single_table_value(element)

            # Prozentuale Aufteilung der Arbeit - sonstiges
            if element_id == u'27':
                template_vars['PVA-S'] = self._set_single_table_value(element)

            # Prozentuale Aufteilung der Arbeit - Elternarbeit
            if element_id == u'30':
                template_vars['PVA-E'] = self._set_single_table_value(element)


            #
            # Bildungsberatung Garantiefonds Hochschule
            #

            if element_id == u'9':
                try:
                    if int(element.find('table').find('rows')[1].find('col').text) > 0:
                        template_vars['BBGH-V'] = 'Ja'
                    else:
                        template_vars['BBGH-V'] = 'Nein'
                except IndexError:
                    template_vars['BBGH-V'] = 'Nein'

            if element_id == u'29':
                template_vars['BBGH-BER'] = self._set_single_table_value(element)

            #
            # Elternberatung
            #
            if element_id == u'32':
                template_vars['V4-EB-A'] = self._set_single_table_value(element)

            if element_id == u'33':
                template_vars['V4-EB-S'] = self._set_single_table_value(element)



        #
        # Netzwerkaktivitäten und Kooperationen
        #
        eval_network = self._get_evaluation(2, year=year)
        root = lxml.etree.fromstring(eval_network)
        for element in root.iter("evaluation"):
            element_id = element.get("id")

            # Mitarbeit an Jugendhilfekonferenz / AG nach §78 SGB VIII (1)
            if element_id == u'1':
                try:
                    template_vars['$JHK-JA$'] = element.find('table').find('rows').find('row')[2].text
                except IndexError:
                    template_vars['$JHK-JA$'] = 0

                try:
                    template_vars['$JHK-NEIN$'] = element.find('table').find('rows').find('row')[1].text
                except IndexError:
                    template_vars['$JHK-NEIN$'] = 0

                try:
                    template_vars['$JHK-KA$'] = element.find('table').find('rows').find('row')[0].text
                except IndexError:
                    template_vars['$JHK-KA$'] = 0

            # Mitarbeit an Jugendkonferenz SGB II/III (2)
            if element_id == u'2':
                try:
                    template_vars['$JH-JA$'] = element.find('table').find('rows').find('row')[2].text
                except IndexError:
                    template_vars['$JH-JA$'] = 0

                try:
                    template_vars['$JH-NEIN$'] = element.find('table').find('rows').find('row')[1].text
                except IndexError:
                    template_vars['$JH-NEIN$'] = 0

                try:
                    template_vars['$JH-KA$'] = element.find('table').find('rows').find('row')[0].text
                except IndexError:
                    template_vars['$JH-KA$'] = 0

            # Zielgruppenspezifisches Netzwerk (3)
            if element_id == u'3':

                # Keine Angabe
                try:
                    template_vars['ZGS-KA'] = element.find('table').find('rows')[0].find('col').text
                except IndexError:
                    template_vars['ZGS-KA'] = 0

                # existiert nicht
                try:
                    template_vars['ZGS-EN'] = element.find('table').find('rows')[1].find('col').text
                except IndexError:
                    template_vars['ZGS-EN'] = 0

                # existiert, aber ohne regionale Koordinierungsstelle
                try:
                    template_vars['ZGS-EOH'] = element.find('table').find('rows')[2].find('col').text
                except IndexError:
                    template_vars['ZGS-EOH'] = 0

                # existiert, wird durch Koordinierungs- oder Regiestelle unterstützt
                try:
                    template_vars['ZGS-EKR'] = element.find('table').find('rows')[3].find('col').text
                except IndexError:
                    template_vars['ZGS-EKR'] = 0

            template_vars['MAN-KA'] = 0
            # Mitarbeit am Netzwerk
            if element_id == u'4':

                # Keine Angabe
                try:
                    template_vars['MAN-KA'] = element.find('table').find('rows')[0].find('col').text
                except IndexError:
                    template_vars['MAN-KA'] = 0

                # keine Mitarbeit durch JMD
                try:
                    template_vars['MAN-KM'] = element.find('table').find('rows')[1].find('col').text
                except IndexError:
                    template_vars['MAN-KM'] = 0

                # Punktuelle Mitarbeit im Netzwerk
                try:
                    template_vars['MAN-PM'] = element.find('table').find('rows')[2].find('col').text
                except IndexError:
                    template_vars['MAN-PM'] = 0

                # Regelmäßige Mitarbeit im Netzwerk
                try:
                    template_vars['MAN-RM'] = element.find('table').find('rows')[3].find('col').text
                except IndexError:
                    template_vars['MAN-RM'] = 0

                # Federführend im Netzwerk
                try:
                    template_vars['MAN-FN'] = element.find('table').find('rows')[4].find('col').text
                except IndexError:
                    template_vars['MAN-FN'] = 0

            # Erstellung der Netzwerkkarte
            if element_id == u'5':
                try:
                    template_vars['$ja$'] = element.find('table').find('rows')[2].find('col').text
                except IndexError:
                    template_vars['$ja$'] = 0

                try:
                    template_vars['$nein$'] = element.find('table').find('rows')[1].find('col').text
                except IndexError:
                    template_vars['$nein$'] = 0

                try:
                    template_vars['$ka$'] = element.find('table').find('rows')[0].find('col').text
                except IndexError:
                    template_vars['$ka$'] = 0

            # öffentliche und freie Jugendhilfe - Davon sind Kooperationspartner
            if element_id == u'8':
                rows = element.find('table').find('rows')

                # Einrichtungen der schulbezogenen Jugendsozialarbeit
                try:
                    template_vars['V3c-01'] = rows[2].find('col').text
                except IndexError:
                    template_vars['V3c-01'] = 0

                # Ausbildungszentren / Jugendwerkstätten
                try:
                    template_vars['V3c-02'] = rows[3].find('col').text
                except IndexError:
                    template_vars['V3c-02'] = 0

                # Einrichtungen für berufsvorbereitende Maßnahmen
                try:
                    template_vars['V3c-03'] = rows[4].find('col').text
                except IndexError:
                    template_vars['V3c-03'] = 0

                # Einrichtungen für berufsbegleitende Maßnahmen
                try:
                    template_vars['V3c-04'] = rows[5].find('col').text
                except IndexError:
                    template_vars['V3c-04'] = 0

                # Beratungsstellen der öffentlichen Jugendhilfe
                try:
                    template_vars['V3c-05'] = rows[6].find('col').text
                except IndexError:
                    template_vars['V3c-05'] = 0

                # Einrichtungen der Heimerziehung
                try:
                    template_vars['V3c-06'] = rows[7].find('col').text
                except IndexError:
                    template_vars['V3c-06'] = 0

                # Jugendgerichtshilfe
                try:
                    template_vars['V3c-07'] = rows[8].find('col').text
                except IndexError:
                    template_vars['V3c-07'] = 0

                # Jugendzentren / offene Jugendarbeit
                try:
                    template_vars['V3c-08'] = rows[9].find('col').text
                except IndexError:
                    template_vars['V3c-08'] = 0

                # Jugendamt
                try:
                    template_vars['V3c-09'] = rows[1].find('col').text
                except IndexError:
                    template_vars['V3c-09'] = 0

                # STÄRKEN vor Ort
                try:
                    template_vars['V3c-70'] = rows[10].find('col').text
                except IndexError:
                    template_vars['V3c-70'] = 0

                # Kompetenzagenturen
                try:
                    template_vars['V3c-10'] = rows[11].find('col').text
                except IndexError:
                    template_vars['V3c-10'] = 0

                # Koordinierungsstelle Schulverweigerung ? Die 2. Chance
                try:
                    template_vars['V3c-11'] = rows[12].find('col').text
                except IndexError:
                    template_vars['V3c-11'] = 0

                # Sonstige
                try:
                    template_vars['V3c-62'] = rows[13].find('col').text
                except IndexError:
                    template_vars['V3c-62'] = 0

                # Koordinierungsstelle des Modellprogramms JUGEND STÄRKEN: Aktiv in der Region
                try:
                    template_vars['V3c-61'] = rows[14].find('col').text
                except IndexError:
                    template_vars['V3c-61'] = 0

            # Integrationsbezogene Einrichtungen - Davon sind Kooperationspartner (11)
            if element_id == u'11':
                rows = element.find('table').find('rows')

                # Jugendmigrationsdienste
                try:
                    template_vars['V3c-12'] = rows[1].find('col').text
                except IndexError:
                    template_vars['V3c-12'] = 0

                # Kommunale Migrationsdienste
                try:
                    template_vars['V3c-13'] = rows[2].find('col').text
                except IndexError:
                    template_vars['V3c-13'] = 0

                # Migrationsberatung (MBE)
                try:
                    template_vars['V3c-14'] = rows[3].find('col').text
                except IndexError:
                    template_vars['V3c-14'] = 0

                # anderweitig geförderte Erwachsenenmigrationsdienste
                try:
                    template_vars['V3c-15'] = rows[4].find('col').text
                except IndexError:
                    template_vars['V3c-15'] = 0

                # Übergangswohnheime / Gemeinschaftsunterkünfte / Aufnahmeeinrichtungen
                try:
                    template_vars['V3c-16'] = rows[5].find('col').text
                except IndexError:
                    template_vars['V3c-16'] = 0

                # Interkulturelle Zentren / Vereine
                try:
                    template_vars['V3c-17'] = rows[6].find('col').text
                except IndexError:
                    template_vars['V3c-17'] = 0

                # Integrationskursträger
                try:
                    template_vars['V3c-20'] = rows[7].find('col').text
                except IndexError:
                    template_vars['V3c-20'] = 0

                # Integrations-, Aussiedler/innen-, Ausländer/innenbeautragte/r
                try:
                    template_vars['V3c-21'] = rows[8].find('col').text
                except IndexError:
                    template_vars['V3c-21'] = 0

                # Kommunale Referate für Integrationsfragen
                try:
                    template_vars['V3c-22'] = rows[9].find('col').text
                except IndexError:
                    template_vars['V3c-22'] = 0

                # Fördereinrichtungen für schulpflichtige Migrant/en/innen
                try:
                    template_vars['V3c-24'] = rows[10].find('col').text
                except IndexError:
                    template_vars['V3c-24'] = 0

                # Vereine von Migrant/en/innen
                try:
                    template_vars['V3c-26'] = rows[11].find('col').text
                except IndexError:
                    template_vars['V3c-26'] = 0

                # Länderspezifische Integrationsangebote
                try:
                    template_vars['V3c-64'] = rows[12].find('col').text
                except IndexError:
                    template_vars['V3c-64'] = 0

                # Sonstige
                try:
                    template_vars['V3c-27'] = rows[13].find('col').text
                except IndexError:
                    template_vars['V3c-27'] = 0

                # Regionalkoordinator/en/innen des BAMF
                try:
                    template_vars['V3c-19'] = rows[14].find('col').text
                except IndexError:
                    template_vars['V3c-19'] = 0

                # Flüchtlingsorganisationen
                try:
                    template_vars['V3c-18'] = rows[15].find('col').text
                except IndexError:
                    template_vars['V3c-18'] = 0


            # Schulen - Davon sind Kooperationspartner (14)
            if element_id == u'14':
                rows = element.find('table').find('rows')

                # Allgemeinbildende Schulen
                try:
                    template_vars['V3c-29'] = rows[1].find('col').text
                except IndexError:
                    template_vars['V3c-29'] = 0

                # Weiterführende Schulen
                try:
                    template_vars['V3c-30'] = rows[2].find('col').text
                except IndexError:
                    template_vars['V3c-30'] = 0

                # Berufliche Schulen / Berufsschulen
                try:
                    template_vars['V3c-31'] = rows[3].find('col').text
                except IndexError:
                    template_vars['V3c-31'] = 0

                # Schulamt / Schulverwaltung
                try:
                    template_vars['V3c-32'] = rows[4].find('col').text
                except IndexError:
                    template_vars['V3c-32'] = 0

                # Schulsozialarbeit
                try:
                    template_vars['V3c-65'] = rows[5].find('col').text
                except IndexError:
                    template_vars['V3c-65'] = 0

                # Schulpsychologischer Dienst
                try:
                    template_vars['V3c-33'] = rows[6].find('col').text
                except IndexError:
                    template_vars['V3c-33'] = 0

                # Sonstige
                try:
                    template_vars['V3c-34'] = rows[7].find('col').text
                except IndexError:
                    template_vars['V3c-34'] = 0

                # Förderschulen
                try:
                    template_vars['V3c-69'] = rows[8].find('col').text
                except IndexError:
                    template_vars['V3c-69'] = 0

            # Weitere öffentliche Einrichtungen - Davon sind Kooperationspartner (17)
            if element_id == u'17':
                rows = element.find('table').find('rows')

                # Jobcenter
                try:
                    template_vars['V3c-66'] = rows[1].find('col').text
                except IndexError:
                    template_vars['V3c-66'] = 0

                # Ausländerbehörde
                try:
                    template_vars['V3c-36'] = rows[2].find('col').text
                except IndexError:
                    template_vars['V3c-36'] = 0

                # Ausgleichsamt / Vertriebenenamt
                try:
                    template_vars['V3c-37'] = rows[3].find('col').text
                except IndexError:
                    template_vars['V3c-37'] = 0

                # Sozialamt
                try:
                    template_vars['V3c-39'] = rows[4].find('col').text
                except IndexError:
                    template_vars['V3c-39'] = 0

                # Gesundheitsamt / Krankenhäuser / Ärzte
                try:
                    template_vars['V3c-41'] = rows[5].find('col').text
                except IndexError:
                    template_vars['V3c-41'] = 0

                # Stadtverwaltung / Gemeindeverwaltung / Bürgeramt / Meldestelle
                try:
                    template_vars['V3c-42'] = rows[6].find('col').text
                except IndexError:
                    template_vars['V3c-42'] = 0

                # Kreisverwaltung / Landratsamt / Regierungsbezirke
                try:
                    template_vars['V3c-43'] = rows[7].find('col').text
                except IndexError:
                    template_vars['V3c-43'] = 0

                # Polizei
                try:
                    template_vars['V3c-44'] = rows[8].find('col').text
                except IndexError:
                    template_vars['V3c-44'] = 0

                # Jugendstrafanstalt
                try:
                    template_vars['V3c-45'] = rows[9].find('col').text
                except IndexError:
                    template_vars['V3c-45'] = 0

                # Sonstige
                try:
                    template_vars['V3c-47'] = rows[10].find('col').text
                except IndexError:
                    template_vars['V3c-47'] = 0

            # Andere Organisationen - Davon sind Kooperationspartner (20)
            if element_id == u'20':
                rows = element.find('table').find('rows')

                # Arbeitgeber/innen / Betriebe
                try:
                    template_vars['V3c-48'] = int(rows[1].find('col').text)
                except IndexError:
                    template_vars['V3c-48'] = 0

                # Handwerks-/ Industrie- und Handelskammer
                try:
                    template_vars['V3c-49'] = int(rows[2].find('col').text)
                except IndexError:
                    template_vars['V3c-49'] = 0

                # Bildungsträger für Aus- und Weiterbildung
                try:
                    template_vars['V3c-50'] = int(rows[3].find('col').text)
                except IndexError:
                    template_vars['V3c-50'] = 0

                # Träger der politischen Bildung
                try:
                    template_vars['V3c-51'] = int(rows[4].find('col').text)
                except IndexError:
                    template_vars['V3c-51'] = 0

                # Begegnungsstätten / Kulturzentren
                try:
                    template_vars['V3c-67'] = int(rows[5].find('col').text)
                except IndexError:
                    template_vars['V3c-67'] = 0

                # Kirchengemeinden
                try:
                    template_vars['V3c-52'] = int(rows[6].find('col').text)
                except IndexError:
                    template_vars['V3c-52'] = 0

                # Medien (Presse, Funk und Fernsehen)
                try:
                    template_vars['V3c-53'] = int(rows[7].find('col').text)
                except IndexError:
                    template_vars['V3c-53'] = 0

                # Stiftungen
                try:
                    template_vars['V3c-54'] = int(rows[8].find('col').text)
                except IndexError:
                    template_vars['V3c-54'] = 0

                # Selbsthilfevereine
                try:
                    template_vars['V3c-55'] = int(rows[9].find('col').text)
                except IndexError:
                    template_vars['V3c-55'] = 0

                # Mehrgenerationenhäuser
                try:
                    template_vars['V3c-56'] = int(rows[10].find('col').text)
                except IndexError:
                    template_vars['V3c-56'] = 0

                # Beratungsstellen (Schuldnerberatung, Drogenberatung, Bewährungshilfe, etc.)
                try:
                    template_vars['V3c-57'] = int(rows[11].find('col').text)
                except IndexError:
                    template_vars['V3c-57'] = 0

                # Religions- und Glaubensgemeinschaften
                try:
                    template_vars['V3c-59'] = int(rows[12].find('col').text)
                except IndexError:
                    template_vars['V3c-59'] = 0

                # JUGEND STÄRKEN: 1000 Chancen
                # (old title: "Wirtschaftsjunioren (Projekt JUGEND STÄRKEN:
                # Junge Wirtschaft macht mit!)", see mpuls/issue5256)
                try:
                    template_vars['V3c-60'] = int(rows[13].find('col').text)
                except IndexError:
                    template_vars['V3c-60'] = 0

                # JUGEND STÄRKEN im Quartier
                # see issue6160 (JMD: JUSTiQ als Auswahlmöglichkeit hinzufügen)
                try:
                    template_vars['V3c-71'] = int(rows[15].find('col').text)
                except IndexError:
                    template_vars['V3c-71'] = 0

                # Sonstige
                try:
                    template_vars['V3c-68'] = rows[14].find('col').text
                except IndexError:
                    template_vars['V3c-68'] = 0

        #
        # Gruppenangebote
        #
        eval_groups = self._get_evaluation(3, year=year)
        root = lxml.etree.fromstring(eval_groups)

        for element in root.iter("evaluation"):
            element_id = element.get("id")

            # KJP Gruppenangebote
            if element_id == u'1':

                # Sonstiges
                try:
                    template_vars['V2a8'] = element.find('table').find('rows')[0].find('col').text
                except IndexError:
                    template_vars['V2a8'] = 0

                # Keine Angabe
                try:
                    template_vars['V2ka'] = element.find('table').find('rows')[1].find('col').text
                except IndexError:
                    template_vars['V2ka'] = 0

                # Orientierungshilfen zu gesellschaftlichen und politischen Themen
                try:
                    template_vars['V2a1'] = element.find('table').find('rows')[2].find('col').text
                except IndexError:
                    template_vars['V2a1'] = 0

                # Orientierungshilfen im Bildungs- und Ausbildungssystem
                try:
                    template_vars['V2a2'] = element.find('table').find('rows')[3].find('col').text
                except IndexError:
                    template_vars['V2a2'] = 0

                # Ergänzendes Sprach- und Kommunikationstraining
                try:
                    template_vars['V2a3'] = element.find('table').find('rows')[4].find('col').text
                except IndexError:
                    template_vars['V2a3'] = 0

                # IT-Seminare/Medienpädagogik
                try:
                    template_vars['V2a4'] = element.find('table').find('rows')[5].find('col').text
                except IndexError:
                    template_vars['V2a4'] = 0

                # Beratung und Informationen zu weiteren Maßnahmen
                try:
                    template_vars['V2a5'] = element.find('table').find('rows')[6].find('col').text
                except IndexError:
                    template_vars['V2a5'] = 0

                # Training sozialer Kompetenzen
                try:
                    template_vars['V2a6'] = element.find('table').find('rows')[7].find('col').text
                except IndexError:
                    template_vars['V2a6'] = 0

                # Elternarbeit
                try:
                    template_vars['V2a7'] = element.find('table').find('rows')[8].find('col').text
                except IndexError:
                    template_vars['V2a7'] = 0

                # Qualifizierungsangebote im Rahmen der Freiwilligenarbeit
                try:
                    template_vars['V2a9'] = element.find('table').find('rows')[9].find('col').text
                except IndexError:
                    template_vars['V2a9'] = 0

                # Gesamt
                try:
                    template_vars['V2ag'] = element.find('table').find('rows')[10].find('col').text
                except IndexError:
                    template_vars['V2ag'] = 0


            # KJP-geförderte Gruppenangebote, Anzahl der TeilnehmerInnen - gesamt
            if element_id == u'2':
                template_vars['V2-TN-G'] = self._set_single_table_value(element)

            # KJP-geförderte Gruppenangebote, Anzahl der TeilnehmerInnen - männlich
            if element_id == u'3':
                template_vars['V2-TN-M'] = self._set_single_table_value(element)

            # KJP-geförderte Gruppenangebote, Anzahl der TeilnehmerInnen - weiblich
            if element_id == u'4':
                template_vars['V2-TN-W'] = self._set_single_table_value(element)

            # KJP-geförderte Gruppenangebote, Anzahl der TeilnehmerInnen nach Alter
            if element_id == u'5':

                value_map = {'12 bis 14 Jahre':'V2-TN-14',
                             '15 bis 18 Jahre':'V2-TN-18',
                             '19 bis 27 Jahre':'V2-TN-27'}

                for et in element.findall('table'):
                    try:
                        template_vars[value_map[et.get('ylabel')]] = \
                            et.find('rows').find('row').find('col').text
                    except AttributeError:
                        template_vars[value_map[et.get('ylabel')]] = 0

            #
            # Nicht-KJP, Gruppenangebote
            #

            # Nicht-KJP, Teilenehmer, weiblich
            if element_id == u'11':
                try:
                    template_vars['V2b-w'] = self._set_single_table_value(element)
                except IndexError:
                    template_vars['V2b-w'] = 0

            # Nicht-KJP, Teilenehmer, maennlich
            if element_id == u'10':
                try:
                    template_vars['V2b-m'] = self._set_single_table_value(element)
                except IndexError:
                    template_vars['V2b-m'] = 0

            # Nicht-KJP, Teilenehmer, gesamtt
            if element_id == u'9':
                try:
                    template_vars['V2b-g'] = self._set_single_table_value(element)
                except IndexError:
                    template_vars['V2b-g'] = 0

            # Nicht-KJP-geförderte Gruppenangebote, Anzahl der TeilnehmerInnen nach Alter
            if element_id == u'12':

                value_map = {'12 bis 14 Jahre':'V2-TN-NKJP-14',
                             '15 bis 18 Jahre':'V2-TN-NKJP-18',
                             '19 bis 27 Jahre':'V2-TN-NKJP-27'}

                for et in element.findall('table'):
                    try:
                        template_vars[value_map[et.get('ylabel')]] = \
                            et.find('rows').find('row').find('col').text
                    except AttributeError:
                        template_vars[value_map[et.get('ylabel')]] = 0


            # Art der Gruppenangebote
            if element_id == u'8':

                # Sonstiges
                try:
                    template_vars['V2-NKJP-5'] = element.find('table').find('rows')[0].find('col').text
                except IndexError:
                    template_vars['V2-NKJP-5'] = 0

                # Keine Angabe
                try:
                    template_vars['V2-NKJP-7'] = element.find('table').find('rows')[1].find('col').text
                except IndexError:
                    template_vars['V2-NKJP-7'] = 0

                # freizeitpädagogische Angebote
                try:
                    template_vars['V2-NKJP-1'] = element.find('table').find('rows')[2].find('col').text
                except IndexError:
                    template_vars['V2-NKJP-1'] = 0

                # Zusammenführung mit gleichaltrigen jungen Menschen
                try:
                    template_vars['V2-NKJP-2'] = element.find('table').find('rows')[3].find('col').text
                except IndexError:
                    template_vars['V2-NKJP-2'] = 0

                # außerschulische Fördermaßnahmen
                try:
                    template_vars['V2-NKJP-3'] = element.find('table').find('rows')[4].find('col').text
                except IndexError:
                    template_vars['V2-NKJP-3'] = 0

                # Vermittlung in Hilfen nach SGB VIII
                try:
                    template_vars['V2-NKJP-4'] = element.find('table').find('rows')[5].find('col').text
                except IndexError:
                    template_vars['V2-NKJP-4'] = 0

                # Gesamt
                try:
                    template_vars['V2-NKJP-6'] = element.find('table').find('rows')[6].find('col').text
                except IndexError:
                    template_vars['V2-NKJP-6'] = 0



            #
            # Interkulturelle Öffnung in Zahlen
            #

            # Interkulturelle Öffnung in Zahlen - Vorträge und Referate
            if element_id == u'19':
                try:
                    template_vars['V4-1'] = self._set_single_table_value(element)
                except IndexError:
                    template_vars['V4-1'] = 0

            # Interkulturelle Öffnung in Zahlen - Informationsveranstaltungen
            if element_id == u'15':
                try:
                    template_vars['V4-2'] = self._set_single_table_value(element)
                except IndexError:
                    template_vars['V4-2'] = 0

            # Interkulturelle Öffnung in Zahlen - Workshops
            if element_id == u'16':
                try:
                    template_vars['V4-3'] = self._set_single_table_value(element)
                except IndexError:
                    template_vars['V4-3'] = 0

            # Interkulturelle Öffnung in Zahlen - interkulturelle Trainings
            if element_id == u'17':
                try:
                    template_vars['V4-4'] = self._set_single_table_value(element)
                except IndexError:
                    template_vars['V4-4'] = 0

            # Interkulturelle Öffnung in Zahlen - Veröffentlichungen
            if element_id == u'18':
                try:
                    template_vars['V4-5'] = self._set_single_table_value(element)
                except IndexError:
                    template_vars['V4-5'] = 0

            # Interkulturelle Öffnung in Zahlen - Sonstiges
            if element_id == u'20':
                try:
                    template_vars['V4-6'] = self._set_single_table_value(element)
                except IndexError:
                    template_vars['V4-6'] = 0

        # Interkulturelle Öffnung in Zahlen - gesamt
        template_vars['V4-g'] = 0
        for i in range(1, 7):
            template_vars['V4-g'] = template_vars['V4-g'] + int(template_vars['V4-%s' % i])

        # Replace empty placeholders
        template_vars['$jahr$'] = str(year)

        agency = Agency(db=get_db("casedb"))
        template_vars['$adresse$'] = agency.getAddress()
        template_vars['$name$'] = agency.getName()

        # Process the file
        template = g.report_template
        odt_doc = expand_odt(template, template_vars)
        odt_doc = set_meta_odt(odt_doc, u"Berichtsexport", u"impuls JMD",
                               u"impuls JMD")
        content = get_doc_file(odt_doc)

        response.headers['Content-Type'] = 'application/msword; charset=utf8'
        response.headers['Content-Disposition'] = \
                       'attachment; filename=JMD-Bericht-%s.doc' % year
        return content

    def _set_single_table_value(self, element):
        # evaluation
        # - table
        #   - rows
        #     - row
        #       - row
        #         - col
        try:
            return element.find('table').find('rows').find('row').find('col').text
        except AttributeError:
            return 0


    def _get_evaluation(self, eval_id, year):
        '''
        Perform an evaluation for the given year and ID
        '''

        init_params = {
            'id': eval_id,
            'start_date': datetime.date(year, 1, 1),
            'end_date': datetime.date(year, 12, 31),
            'typelist': [unicode(i) for i in range(1, 35)],
            'phase': [u'-1', u'0'],
            'type_ending': []
            }
        eval = EvaluateController()
        params = eval._get_evalparams(init_params, id_field=None,
                                      sql_where=None)

        conn = db.getConnection()
        try:
            evalconfig = EvaluationConfig(get_configfile(eval_id),
                                          conn,
                                          params['start_date'],
                                          params['end_date'],
                                          params['start_date_field'],
                                          params['end_date_field'],
                                          None,
                                          params['sql'],
                                          params['typelist'])
        finally:
            # Rollback any changes made by libmpuls. libmpuls creates a
            # temporary table for the evaluation but does not clean up
            # after itself, apparently, so we have to do that here,
            # otherwise we cannot perform multiple evaluations with the
            # same database connection. An alternative to the rollback
            # would be to use new connections for each evaluation, but
            # that does not really improve things. In any case, we have
            # to assume that the report is not generated as part of a
            # larger transaction.
            conn.rollback()
            db.recycleConnection(conn, None)

        # Massage the filter to make it human readable
        selected_phases = get_phase_filter_for_evaluation(params.get('phase'))
        evalconfig.set_filter({"phases": selected_phases})

        # Set second parameter to True to enable debugging
        evalset = EvaluationSet(evalconfig)
        evalset.evaluate()
        return evalset.export(EvaluationExportXML())
