# -*- coding: utf-8 -*-
import logging

import formencode
from formencode import Pipe, ForEach, NoDefault
from formencode.validators import Int, OneOf, Regex

from pylons import app_globals as g, tmpl_context as c

from libmpuls.evaluation.export import EvaluationExportHTML, \
    EvaluationExportXML, EvaluationExportCSV, EvaluationExportXLS

from mpulsweb.controllers.evaluate import EvaluateController \
                                          as MpulsEvaluateController
from mpulsweb.model.agencysettings import Agency
from mpulsweb.lib.validators import EvaluationFormValidator
from mpulsweb.lib.base import session


log = logging.getLogger(__name__)


def int_option_validator(options):
    return Pipe(Int(), OneOf([item[0] for item in options]))


class JMDExportMixin(object):

    def __init__(self, *args, **kw):
        self.zielgruppe = kw.pop("zielgruppe", None)
        super(JMDExportMixin, self).__init__(*args, **kw)

        if self.zielgruppe is not None:
            desc = self.parameter_desc[:]
            for i, (label, path) in enumerate(self.parameter_desc):
                if label == "Erstellungsdatum":
                    desc.insert(i, ("Zielgruppe", "jmd-zielgruppe"))
            self.parameter_desc = desc


    def _get_application_parameters(self):
        if self.zielgruppe is not None:
            return u"<jmd-zielgruppe>%s</jmd-zielgruppe>" % (self.zielgruppe,)
        return u""


class JMDEvaluationExportHTML(JMDExportMixin, EvaluationExportHTML):

    pass


class JMDEvaluationExportXML(JMDExportMixin, EvaluationExportXML):

    pass


class JMDEvaluationExportCSV(JMDExportMixin, EvaluationExportCSV):

    pass


class JMDEvaluationExportXLS(JMDExportMixin, EvaluationExportXLS):

    pass



class EvaluatePhaseValidator(formencode.validators.FancyValidator):

    def _to_python(self, value, state):
        if value == "9999":
            return ["1", "2"]
        return [value]

    def _from_python(self, value, state):
        if value == ["1", "2"]:
            return "9999"
        if len(value) == 1:
            return value[0]
        elif len(value) == 0:
            return None
        else:
            raise formencode.Invalid("value must be a list of"
                                     " exactly one element", value, state)

    def validate_python(self, value, state):
        for v in value:
            if not isinstance(v, (str, unicode)):
                raise formencode.Invalid("Element %r is not a string" % (v,),
                                         value, state)
            try:
                map(int, value)
            except TypeError, e:
                raise formencode.Invalid("Value %r cannot be converted to int"
                                         % (v,), value, state)


class EvaluateController(MpulsEvaluateController):

    def _render_evaluate_page(self, eval_config, eval_config_id, *args, **kw):
        c.jmd2start_zielgruppen = self._jmd2start_zielgruppe_options()

        # We only want to show the "Aufenthaltspapier" evaluation to
        # jmd2start agencies, so we remove it for other agencies.
        c.evaluations = eval_config.get_evaluations()
        if eval_config_id == "1" and not self._is_jmd2start():
            c.evaluations = [i for i in c.evaluations
                             if not "Aufenthaltspapier" \
                                    in eval_config.get_name(i)]

        return super(EvaluateController, self)._render_evaluate_page(
            eval_config, eval_config_id, *args, **kw)

    def _jmd2start_zielgruppe_options(self):
        """Return the options allowed for the jmd2start_zielgruppe field.
        This field is only available as an evaluation option when the
        agency takes part in jmd2start and when the application is not
        an evaluation server. The evaluation server for jmd2start only
        contains jmd2start files so the option is not necessary.
        """
        if (self._is_jmd2start()
            and not g.mpuls_config.is_enabled('module', 'agency-overview')):
            nc = g.formedTree.findByName("zielgruppe")
            return [opt for opt in nc.get_options() if opt[0] >= 0]
        return []

    def _is_jmd2start(self):
        return Agency().get_value("jmd2start")

    def _get_validator(self):
        extra_validators = {
            "phase": EvaluatePhaseValidator(not_empty=True)
        }

        options = self._jmd2start_zielgruppe_options()
        if options:
            extra_validators["jmd2start_zielgruppe"] \
                = int_option_validator(options)
        return EvaluationFormValidator(**extra_validators)


    def appspecific_clause(self, params):
        jmd2start_zielgruppe = params.get("jmd2start_zielgruppe")
        if self._is_jmd2start() and jmd2start_zielgruppe is not None:
            return ("m.zielgruppe = %(zielgruppe)s",
                    {"zielgruppe": jmd2start_zielgruppe})
        else:
            return super(EvaluateController, self).appspecific_clause(params)

    def _get_options(self, *args, **kw):
        soptions = session.get('evaluation.options')
        jmd2start_zielgruppe = 0
        if soptions:
            jmd2start_zielgruppe = soptions.get("jmd2start_zielgruppe",
                                                jmd2start_zielgruppe)
        return (super(EvaluateController, self)._get_options(*args, **kw)
                + [dict(jmd2start_zielgruppe=jmd2start_zielgruppe)])

    def _new_html_exporter(self, eval_params, *args, **kw):
        return self._new_jmd_exporter(JMDEvaluationExportHTML, eval_params,
                                      *args, **kw)

    def _new_xml_exporter(self, *args, **kw):
        return self._new_jmd_exporter(JMDEvaluationExportXML,
                                      session.get('evaluation.params'),
                                      *args, **kw)

    def _new_xls_exporter(self, *args, **kw):
        return self._new_jmd_exporter(JMDEvaluationExportXLS,
                                      session.get('evaluation.params'),
                                      *args, **kw)

    def _new_csv_exporter(self, *args, **kw):
        return self._new_jmd_exporter(JMDEvaluationExportCSV,
                                      session.get('evaluation.params'),
                                      *args, **kw)

    def _new_jmd_exporter(self, cls, eval_params, *args, **kw):
        zielgruppe_options = self._jmd2start_zielgruppe_options()
        if zielgruppe_options:
            zielgruppe_num = eval_params["sqlparams"]["zielgruppe"]
            for opt, desc in zielgruppe_options:
                if opt == zielgruppe_num:
                    kw["zielgruppe"] = desc
                    break

        return cls(*args, **kw)
