# -*- coding: utf-8 -*-
#
# Copyright 2007-2018 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Ludwig Reiter (ludwig.reiter@intevation.de)
# Dirk Massmann (dirk.massmann@intevation.de)

import logging
import datetime

from mpulsweb.lib.base import session, g
from mpulsweb.lib.db import db
from mpulsweb.lib.translation import _
from mpulsweb.model.appointment import AppointmentOverview, \
        CaseAppointmentFactory
from mpulsweb.model.case import MpulsCaseOverview

log = logging.getLogger(__name__)

class AgeReminderOverview(AppointmentOverview):
    """Generic overview controlling age of person."""

    AGE_CONDITION = ""
    AGE_MESSAGE = ""
    AGE_MESSAGE_TITLE = ""

    def __init__(self):

        self.appointment_list = []
        cases = MpulsCaseOverview()

        user_id = self.getUserId()
        sm = g.case_factory.get_state_machine()
        cases.search({"own": user_id, "standin": user_id,
                      "state": [1, 2],
                      "phase": sm.get_search_item_ids()})

        case_id_list = []
        with db.cursor() as cur:
            cur.execute(self.AGE_CONDITION,
                        {'editor': self.getUserId()})
            for r in cur.fetchall():
                case_id_list.append(int(r[0]))

        case_list = [case for case in cases.getDatasets()
                          if case.id in case_id_list]
        for case in case_list:
            appointment = self.makeAppointment(case,
                    self.AGE_MESSAGE,
                    self.AGE_MESSAGE_TITLE)
            self.appointment_list.append(appointment)

    def makeAppointment(self, case, description, title):
        factory = CaseAppointmentFactory()
        appointment_type = 2

        appointment = factory.createByData(None,
                                           title,
                                           datetime.datetime.now(),
                                           None,
                                           description,
                                           case.id,
                                           appointment_type)
        return appointment

    def getUserId(self):
        """
        Return the User-ID. This function can be overwritten by applications
        that have a "structure" part where the user that owns the cases is not
        the user logged in.
        """
        user = session['USER_AUTHORIZED']
        return user.id

class MinAgeReminderOverview(AgeReminderOverview):
    """JMD specific overview. If a young person hasn't reached an age of
    12 years, the cm should be informed."""

    AGE_CONDITION = """
    SELECT m.id
    FROM master_tbl_view m
    WHERE m.bearbeiter_id = %(editor)s
    AND age(m.geburtsdatum) < interval '12 years'
    """
    AGE_MESSAGE = _('The age of the young person is under 12.')
    AGE_MESSAGE_TITLE = _('Age under 12')

class MaxAgeReminderOverview(AgeReminderOverview):
    """JMD specific overview. If a young person reaches an age of 27, the cm
    should be informed."""

    AGE_CONDITION = """
    SELECT m.id
    FROM master_tbl_view m
    WHERE m.bearbeiter_id = %(editor)s
    AND age(m.geburtsdatum) > interval '27 years'
    """
    AGE_MESSAGE = _('The age of the young person is above 27.')
    AGE_MESSAGE_TITLE = _('Age above 27')

class JustBeforeMaxAgeReminderOverview(AgeReminderOverview):
    """JMD specific overview. If a young person reaches an age of 27 within the
    next 6 months, the cm should be informed."""

    AGE_CONDITION = """
    SELECT m.id
    FROM master_tbl_view m
    WHERE m.bearbeiter_id = %(editor)s
    AND age(m.geburtsdatum) > interval '318 months'
    AND age(m.geburtsdatum) < interval '324 months'
    """
    AGE_MESSAGE = _('The age of the young person will be 27 within the next 6 months.')
    AGE_MESSAGE_TITLE = _('Reaching age of 27 soon')
