# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008, 2010 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls JMD
#
# mpuls JMD is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls JMD is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls JMD. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls JMD has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
# Authors:
# Bernhard Herzog <bh@intevation.de>

from pylons import config

from mpulsweb.lib.db import DB
from mpulsweb.lib.security import get_db_parameters
from mpulsweb.model.user import UserListObject, parse_dblogin
from mpulsweb.lib.base import session


def struktur_part_available():
    """Returns whether the struktur part is available.
    The struktur part is available if the struktur database connection
    is configured."""
    return config.get('jmd.structuredb.host') is not None


def get_db(partname="structuredb"):
    current_user = session['USER_AUTHORIZED']
    return DB(**get_db_parameters(current_user.login,
                                  current_user.password,
                                  config.get('jmd.%s.host' % partname),
                                  config.get('jmd.%s.port' % partname),
                                  config.get('jmd.%s.database' % partname),
                                  config.get('jmd.%s.user' % partname)))


def get_struktur_users():
    """Return a list with the ids of the users with access to the struktur part.
    """
    all_users = UserListObject()
    user_map = dict((user.login, user.id) for user in all_users.getUserList())

    with get_db().cursor() as cur:
        cur.execute("SELECT login FROM jmd_struktur_nutzer_view")
        user_ids = []
        while True:
            row = cur.fetchone()
            if not row:
                break
            user_id = user_map.get(parse_dblogin(row[0])[0])
            if user_id is not None:
                user_ids.append(user_id)

    return user_ids


def remove_struktur_user_by_login(user_login, user_agency):
    """
    Deletes user from the ka_benutzer_tbl on the structure part.
    """
    struktur_db = get_db()
    con = cur = None
    with get_db().transaction() as cur:
        cur.execute("SELECT id FROM ka_benutzer_tbl WHERE login = %s",
                    ("ka_%s_%s" % (user_agency, user_login),))
        row = cur.fetchone()
        if row:
            cur.execute("SELECT jmd_delete_struktur_nutzer(%s)", (row[0],))


def set_struktur_users(user_ids):
    """Define the group of users who may access the struktur part.
    All users in user_ids and only those users, are given permission to
    access the struktur data.
    """
    user_ids = set(user_ids)

    all_users = UserListObject()

    new_logins = set(user.full_login for user in all_users.getUserList()
                     if user.id in user_ids)

    with get_db().transaction() as cur:
        cur.execute("SELECT b.id, b.login"
                    "  FROM ka_benutzer_tbl_view b"
                    "  JOIN nm_benutzer_gruppe_tbl_view g"
                    "    ON g.benutzer_id = b.id"
                    " WHERE g.gruppe_id ="
                    "       (SELECT id FROM ka_benutzergruppe_tbl_view"
                    "         WHERE name = 'struktur-nutzer')")
        old_users = cur.fetchall()
        old_logins = set(r[1] for r in old_users)

        to_delete = old_logins - new_logins
        to_add = new_logins - old_logins

        cur.executemany("SELECT jmd_delete_struktur_nutzer(%s)",
                        [(uid,) for uid, login in old_users
                         if login in to_delete])

        for user in all_users.getUserList():
            if user.full_login in to_add:
                cur.execute("SELECT jmd_add_struktur_nutzer(%s, %s, %s, %s)",
                            (user.getAgency(), user.getLogin(),
                             user.first_name, user.last_name))


def is_struktur_user():
    """Return whether the current user may access the struktur part."""
    with get_db().cursor() as cur:
        cur.execute("SELECT jmd_check_struktur_permission()")
        user_ids = []
        row = cur.fetchone()
        if not row:
            return False
        return row[0]
