"""Pylons middleware initialization"""

import logging
import locale

from webob.exc import HTTPMovedPermanently, HTTPInternalServerError
from beaker.middleware import CacheMiddleware, SessionMiddleware
from paste.cascade import Cascade
from paste.registry import RegistryManager
from paste.urlparser import StaticURLParser
from paste.deploy.converters import asbool
from pylons.util import class_name_from_module_name
from pylons.middleware import ErrorHandler, StatusCodeRedirect
from pylons.wsgiapp import PylonsApp
from routes.middleware import RoutesMiddleware
import formencode

from mpulsweb.config.environment import load_environment
from mpulsweb.config.importer import import_overridable_module


log = logging.getLogger(__name__)


def log_wsgi_request_data_read_error(app):
    """Middleware to detect and log mod_wsgi's 'request data read error'.

    When using mod_wsgi with daemon processes it can happen that the
    request is not fully transmitted to the daemon process for various
    reasons (timeouts, clients that close connections prematurely, and
    more). The error log will contain lines with the message 'Unable to
    get bucket brigade for request' which come from mod_wsgi in the
    apache process. In the daemon process this will lead to a failure to
    read data from the apache process which mod_wsgi in the daemon
    process reports as an IOError with the message 'request data read
    error'.

    This middleware catches those exceptions, logs the message and
    returns an HTTPInternalServerError. Without this, the exception will
    usually be logged with a full traceback, filling up the log files
    with somewhat useless messages that distract from more serious
    errors. The daemon process cannot do anything about the problem
    anyway and it's not a defect in mpuls.

    A middleware stack might contain multiple places where this middle
    is useful. It should usually be used inside a generic error handler
    middleware---so that it's not logged by that middleware and turned
    into an internal server error---and probably as the outermost
    middleware, too, because there might be other middlewares around the
    error handler middleware that try to read the request data.
    """
    def call_app(*args, **kw):
        try:
            return app(*args, **kw)
        except IOError, exc:
            msg = str(exc)
            if "request data read error" in msg:
                log.error("Error while reading request from mod_wsgi: %r",
                          msg)
                return HTTPInternalServerError()(*args, **kw)
            else:
                raise
    return call_app


class SimpleExceptionLogger(object):

    """Simply log any exception raised while a request is handled.

    If an exception is raised during a request, the exception is logged
    with log.exception and reraised without any further processing.
    This can be used to make sure that the actual exception information
    is logged even if the more sophisticated error handlers of
    e.g. Pylons fail.  For that use-case it has to be as near to the
    actual application in the application stack as possible.
    """

    def __init__(self, application):
        self.application = application

    def __call__(self, *args, **kw):
        try:
            return self.application(*args, **kw)
        except:
            log.exception("Exception while handling request")
            raise


class FrameOptions(object):

    """Middleware that simply sets the X-Frame-Options header field to DENY.

    mpuls does not use any frames, so it doesn't hurt to tell clients
    not to allow using an mpuls application in a frame, and it improves
    security by preventing click-jacking attacks.
    """

    def __init__(self, application):
        self.application = application

    def __call__(self, environ, start_response):
        def start_response_wrapper(status, response_headers, exc_info=None):
            headers = [("X-Frame-Options", "DENY")] + response_headers
            return start_response(status, headers, exc_info)

        return self.application(environ, start_response_wrapper)


class NoCacheHeaderMiddleware(object):

    """ Middleware to disable browser caching of html content """

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        def start_response_wrapper(status, headers, exec_info=None):
            # no-cache should be set by default in pylons already but add it
            # here too because we shouldn't rely on pylons
            # with no-cache the browser should not store the webpage in its
            # cache. max-age=0 is set mostly set for convenience
            headers.append(("Cache-Control", "no-cache"))
            headers.append(("Cache-Control", "max-age=0"))
            # only store webpage in volatile memory
            headers.append(("Cache-Control", "no-store"))
            # always reload page. This one is most important for the csrf token
            # validation
            headers.append(("Cache-Control", "must-revalidate"))
            return start_response(status, headers, exec_info)
        return self.app(environ, start_response_wrapper)


class MyPylonsApp(PylonsApp):

    def find_controller(self, controller):
        """Override this method to look for the controller in two places.

        First, the controller is searched in the MPuls specialization's
        controllers package, and if it cannot be found there, it's
        imported from the mpulsweb.
        """
        # Check to see if we've cached the class instance for this name
        if controller in self.controller_classes:
            return self.controller_classes[controller]

        # Hide the traceback here if the import fails (bad syntax and such)
        __traceback_hide__ = 'before_and_this'

        module = import_overridable_module("controllers."
                                           + controller.replace('/', '.'),
                                           self.config)

        if hasattr(module, '__controller__'):
            mycontroller = getattr(module, module.__controller__)
        else:
            module_name = controller.split('/')[-1]
            class_name = class_name_from_module_name(
                module_name) + 'Controller'
            if self.log_debug:
                log.debug("Found controller, module: %r, class: %r",
                          module, class_name)
            mycontroller = getattr(module, class_name)
        self.controller_classes[controller] = mycontroller
        return mycontroller


def make_app(global_conf, full_stack=True, static_files=True, **app_conf):
    """Create a Pylons WSGI application and return it

    ``global_conf``
        The inherited configuration for this application. Normally from
        the [DEFAULT] section of the Paste ini file.

    ``full_stack``
        Whether this application provides a full WSGI stack (by default,
        meaning it handles its own exceptions and errors). Disable
        full_stack when this application is "managed" by another WSGI
        middleware.

    ``static_files``
        Whether this application serves its own static files; disable
        when another web server is responsible for serving them.

    ``app_conf``
        The application's local configuration. Normally specified in
        the [app:<name>] section of the Paste ini file (where <name>
        defaults to main).

    """
    # Configure the Pylons environment
    config = load_environment(global_conf, app_conf)

    # Set locale default locale to de_DE.utf8
    locale.setlocale(locale.LC_ALL, global_conf.get('locale', 'de_DE.utf8'))
    formencode.api.set_stdtranslation(domain="FormEncode", languages=["de"])

    # The Pylons WSGI app
    app = MyPylonsApp(config=config)

    # Pylon's exception handler (at least the debug handler) sometimes
    # has problems with exceptions whose messages are byte-strings
    # containing non-ascii characters, which can happen when
    # e.g. psycopg2 is used in a German locale.  To avoid loss of
    # information, we put the SimpleExceptionLogger immediately around
    # the pylons app.
    if config.get("mpuls.app.simple-exception-handler", False):
        log.debug("Using MPuls SimpleExceptionLogger")
        app = SimpleExceptionLogger(app)
    else:
        log.debug("Not using MPuls SimpleExceptionLogger")

    # Routing/Session/Cache Middleware
    app = RoutesMiddleware(app, config['routes.map'])
    app = SessionMiddleware(app, config)
    app = CacheMiddleware(app, config)

    # CUSTOM MIDDLEWARE HERE (filtered by error handling middlewares)
    app = NoCacheHeaderMiddleware(app)

    # Add log_wsgi_request_data_read_error middleware here, so that it
    # catches the "request data read error" before it's handled by the
    # ErrorHandler middleware.
    app = log_wsgi_request_data_read_error(app)

    if asbool(full_stack):
        # Handle Python exceptions
        app = ErrorHandler(app, global_conf, **config['pylons.errorware'])

        # Display error documents for 401, 403, 404 status codes (and
        # 500 when debug is disabled)
        if asbool(config['debug']):
            app = StatusCodeRedirect(app, [400, 401, 403, 404, 405])
        else:
            app = StatusCodeRedirect(app, [400, 401, 403, 404, 405, 500])

    # Establish the Registry for this application
    app = RegistryManager(app)

    if asbool(static_files):
        # Serve static files
        static_apps = [StaticURLParser(path)
                       for path in config['pylons.paths']['static_files']]
        app = Cascade(static_apps + [app])

    app = FrameOptions(app)

    # Add log_wsgi_request_data_read_error middleware here, so that it
    # catches the "request data read error" that might be raised by e.g.
    # the Cascade middleware and perhaps other middleware used here.
    app = log_wsgi_request_data_read_error(app)

    app.config = config
    return app


def make_redirector(global_conf, **app_conf):
    """Create a WSGI Application that permanently redirects to a specific URL.
    The URL can be specified with the location setting in the configuration.

    This is useful for e.g. a composite application with two WSGI
    applications under different prefixes.  In this case one may want to
    specify a default application to use when accessing the root URL of
    the server.
    """
    # We need to create a new instance of HTTPMovedPermanently for every
    # request because HTTPMovedPermanently instances modify themselves
    # in-place when they process a request by storing the absolute
    # redirection URL in self.location and that value will be as the
    # location the responses sent for all following requests.
    location = app_conf["location"]
    def redirector(environ, start_response):
        return HTTPMovedPermanently(location=location)(environ, start_response)
    return redirector
