# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Torsten Irländer <torsten.irlaender@intevation.de>
# Sascha L. Teichmann <teichmann@intevation.de>

import logging
from datetime import datetime

from pylons.controllers.util import redirect, abort

import formencode

from mpulsweb.lib.translation import _

from mpulsweb.lib.base import BaseController, c, render, request, \
    render_htmlfill, restrict_to_POST, abort_unless_POST
from mpulsweb.lib.helpers import url, dd_mm_YYYY, HH_MM
from mpulsweb.lib.security import checkRole
from mpulsweb.model.appointment import GlobalAppointmentOverview, \
    GlobalAppointmentFactory, MissingAppointmentError

from mpulsweb.lib.validators import CreateAppointmentForm

log = logging.getLogger(__name__)


class AppointmentController(BaseController):

    def _load_global_appointment(self, id):
        try:
            return GlobalAppointmentFactory().loadById(id)
        except MissingAppointmentError:
            abort(404)

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def index(self):
        redirect(url(controller='appointment', action='overview'))

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def overview(self):
        c.appointmentlist = GlobalAppointmentOverview()
        return render('/appointments/overview.mako')

    @checkRole(('admin'))
    def create(self):
        c.form_errors = {}
        c.form_result = {}
        cd = datetime.now()
        c.form_defaults = {'start_date': dd_mm_YYYY(cd),
                           'start_time': HH_MM(cd),
                           'case_id': None}
        form = render('/appointments/new.mako')
        return render_htmlfill(form, defaults=c.form_defaults)

    @checkRole(('admin'))
    @restrict_to_POST
    def createAction(self):
        validator = CreateAppointmentForm()
        factory = GlobalAppointmentFactory()
        c.form_errors = {}
        c.form_result = {}
        try:
            c.form_result = validator.to_python(request.params)
            # Create appointment
            appointment = factory.createNew()
            appointment.setData(c.form_result)
            appointment.store()
            c.dialog_title = _(u"Appointment created")
            c.dialog_text = _(u"The appointment was successfully created.") + \
                " " + \
                _("Please click OK to continue to the appointment overview")
            c.url_ok = url(controller='appointment', action='overview')
            return render(
                '/appointments/dialogs/success_create_appointment.mako')
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            form = render('/appointments/new.mako')
            return render_htmlfill(form, defaults=c.form_result,
                                   auto_insert_errors=False,
                                   errors=c.form_errors)

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def show(self, id):
        c.appointment = self._load_global_appointment(self._checkInt(id))
        return render('/appointments/show.mako')

    @checkRole(('admin'))
    def edit(self, id):
        app = self._load_global_appointment(self._checkInt(id))

        c.form_result = {'start_date': dd_mm_YYYY(app.start_date),
                         'start_time': HH_MM(app.start_date),
                         'end_date': dd_mm_YYYY(app.end_date),
                         'end_time': HH_MM(app.end_date),
                         'title': app.title,
                         'id': app.id,
                         'description': app.description}

        c.form_errors = {}
        form = render('/appointments/edit.mako')
        return render_htmlfill(form, defaults=c.form_result,
                               auto_insert_errors=False, errors=c.form_errors)

    @checkRole(('admin'))
    @restrict_to_POST
    def editAction(self):
        # validate the appointment ID. It's a hidden input field managed
        # by mpuls and not editable by the user. If this validation
        # fails, it's a defect in mpuls and therefore an internal server
        # error should be OK.
        appointment_id = self._checkInt(request.params["id"])

        validator = CreateAppointmentForm()
        factory = GlobalAppointmentFactory()
        c.form_errors = {}
        c.form_result = {}
        try:
            form_result = validator.to_python(request.params)
            # Load appointment and set data
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            form = render('/appointments/edit.mako')
            return render_htmlfill(form, defaults=c.form_result,
                                   auto_insert_errors=False,
                                   errors=c.form_errors)

        appointment = self._load_global_appointment(appointment_id)
        appointment.setData(form_result)
        appointment.store()
        c.dialog_title = _(u"Appointment edited")
        c.dialog_text = _(u"The appointment was successfully edited") + \
            " " + \
            _("Please click OK to continue to the appointment overview")
        c.url_ok = url(controller='appointment', action='overview')
        return render('/appointments/dialogs/success_edit_appointment.mako')

    @checkRole(('admin'))
    def delete(self, id, confirmed=False):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        if confirmed == 1:
            abort_unless_POST()
            appointment = self._load_global_appointment(id)
            appointment.delete()
            c.dialog_title = _(u'Appointment deleted')
            c.dialog_text = _(u"The appointment was successfully deleted.") + \
                " " + \
                _("Please click OK to continue to the appointment overview")
            c.url_ok = url(controller='appointment', action='overview')
            return render(
                '/appointments/dialogs/success_delete_appointment.mako')
        else:
            c.context = "../main.mako"
            c.dialog_title = _(u'Delete appointment?')
            c.dialog_text = _(u"Do you really want to delete the appointment"
                              u"\nand irrevocably lose all the data?")
            c.url_yes = url(controller='appointment', action='delete',
                            id=id, confirmed=1)
            c.url_no = url(controller='appointment', action='overview')
            return render('/appointments/dialogs/confirm.mako')

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def save(self):
        return render('/appointments/new.mako')

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
