# -*- coding: utf-8 -*-

import logging
import formencode

import paste

from pylons import request, tmpl_context as c, url
from pylons.controllers.util import redirect, abort
from mpulsweb.lib.translation import _

from mpulsweb.lib.base import BaseController, render, session, g, \
    render_htmlfill, restrict_to_POST, abort_unless_POST
from mpulsweb.lib.validators import UploadCaseDocumentForm, \
    UploadCaseDocumentWDescForm, \
    RenameDocumentForm, \
    RenameDocumentWDescForm

from mpulsweb.lib.security import checkRole
import mpulsweb.lib.helpers as h

from mpulsweb.model.document import Document, deleteDocument, renameDocument, \
    listDocuments, DocumentException


log = logging.getLogger(__name__)


class CasedocumentController(BaseController):

    @checkRole(('admin', 'cm'))
    def index(self, id):
        return self.overview(id)

    @checkRole(('admin', 'cm'))
    def overview(self, id):
        id = self._checkInt(id)
        case = self._loadCase(id, read_only=True)
        c.files = case.getDocuments()
        return render('/documents/case_overview.mako')

    @checkRole('cm')
    def new(self, id):
        id = self._checkInt(id)
        self._ensure_session_case(id)
        return render('/documents/case_new.mako')

    @checkRole('cm')
    @restrict_to_POST
    def newAction(self):
        case_id = self._checkInt(request.params.get("case"))
        self._ensure_session_case(case_id)
        if not session["case"].is_editable_state():
            abort(401)

        if g.mpuls_config.is_enabled("module", "documents-description"):
            validator_cls = UploadCaseDocumentWDescForm
        else:
            validator_cls = UploadCaseDocumentForm
        validator = validator_cls(
            max_size=g.mpuls_config.get("case", "max-document-size"),
            max_size_text=g.mpuls_config.get("case", "max-document-size-text")
            )
        try:
            result = validator.to_python(request.params)
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            log.debug(c.form_result)
            log.warn("Invalid form data for case document %s" % c.form_errors)
            form = render('/documents/case_new.mako')
            return render_htmlfill(form, defaults=c.form_result,
                                   auto_insert_errors=False,
                                   errors=c.form_errors)

        doc = Document()
        filename = result['file']['filename']
        content = result['file']['content']
        case = result['case']
        if result['name']:
            # Make sure the new name has a (3 letter) extension
            if '.' not in result['name'][-4:]:
                extension = filename.rpartition('.')[2]
                filename = result['name'] + '.' + extension
            else:
                filename = result['name']
        if g.mpuls_config.is_enabled("module", "documents-description"):
            doc.create(filename, content, case,
                       description=result['description'])
        else:
            doc.create(filename, content, case)
        c.dialog_title = _(u"Attachment added!")
        c.dialog_text = _(u"Please click OK to continue.")
        c.url_ok = h.url(controller="casedocument", action="index",
                         id=case)
        return render('/documents/dialogs/success_attachment.mako')

    @checkRole(('admin', 'cm'))
    def show(self, id):
        id = self._checkInt(id)
        try:
            document = Document(id)
            document.load(True)
            data = paste.fileapp.DataApp(document.data,
                                         content_type=document.mime,
                                         cache_control='must-revalidate',
                                         pragma='public')
            return data(request.environ, self.start_response)
        except Exception, e:
            log.exception(e)
            redirect(url(controller="case", action="documents"))

    def _load_editable_case_document(self, doc_id):
        doc_id = self._checkInt(doc_id)
        doc = Document(doc_id)
        try:
            doc.load(asCase=True, load_contents_immediately=False)
        except DocumentException:
            # The most likely reason for an exception apart from defects
            # in the code is that the document simply doesn't exist. In
            # that case a 404 error seems an appropriate response.
            log.exception("Cannot load document for id %r", doc_id)
            abort(404)

        self._ensure_session_case(doc.case)
        if not session["case"].is_editable_state():
            abort(401)

        return doc

    @checkRole('cm')
    def delete(self, id, confirmed):
        doc = self._load_editable_case_document(id)

        confirmed = self._checkBool(confirmed)
        if confirmed == 1:
            abort_unless_POST()
            try:
                deleteDocument(doc.id, True)
            except Exception, e:
                log.exception(e)
            c.url_ok = h.url(controller="casedocument", action="index",
                             id=doc.case)
            c.success_for = _(u"Attachment deleted!")
            c.success_text = _(u"Please click OK to continue.")
            return render('/casemanagement/dialogs/'
                          'success_delete_attachment.mako')
        else:
            c.context = "../main.mako"
            c.confirm_for = _(u"Delete attachment?")
            c.question = _(u"Do you want to irrevocably delete the attachment"
                           u" '%s'?") % doc.name
            c.url_yes = h.url(controller='casedocument', action='delete',
                              id=id, confirmed=1)
            c.url_no = h.url(controller="casedocument", action="index",
                             id=doc.case)
            return render('/casemanagement/dialogs/'
                          'confirm_delete_attachment.mako')

    @checkRole('cm')
    def rename(self, id, confirmed='0'):
        doc = self._load_editable_case_document(id)

        c.id = doc.id
        c.name = doc.getName()
        if not c.name:
            c.name = ""
        c.description = doc.description
        if not c.description:
            c.description = ""

        return render('/casemanagement/rename_attachment.mako')

    @checkRole('cm')
    @restrict_to_POST
    def renameAction(self, id):
        doc = self._load_editable_case_document(id)
        if g.mpuls_config.is_enabled("module", "documents-description"):
            validator = RenameDocumentWDescForm()
        else:
            validator = RenameDocumentForm()
        try:
            result = validator.to_python(request.params)
        except formencode.Invalid, error:
            c.id = doc.id
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            log.debug(c.form_result)
            form = render('/casemanagement/rename_attachment.mako')
            return render_htmlfill(form, defaults=c.form_result,
                                   auto_insert_errors=False,
                                   errors=c.form_errors)

        filename = result['name']
        desc = result.get('description', '')
        renameDocument(doc.id, filename, desc, True)

        redirect(url(controller="casedocument", action="index", id=doc.case))
