# -*- coding: utf-8 -*-
#
# Authors:
# Ludwig Reiter <ludwig.reiter@intevation.de>
# Bernhard Herzog <bernhard.herzog@intevation.de>

import thread
import logging


class DBNameContextFilter(logging.Filter):

    """Logging filter that adds the databasename to log records.

    The database name is different for each request and multiple
    requests may be handled simultaneously in multiple threads, so
    determining the database name needs to be done carefully. Simply
    using the getDbName function from the security module doesn't work
    properly, because, even though it would return the correct result,
    it logs messages using the logging system, so the filter could be
    called recursively.

    To avoid such problems, the class provides a way to store and
    retrieve thread-specific database names with the class methods
    set_thread_dbname and get_thread_dbname. Some other code, e.g. the
    base controller, calls set_thread_dbname with the appropriate
    database name at the start of a request and sets it to None after
    the request has been finished.

    The filter always succeeds, i.e. it never suppresses any record, but
    it modifies the records it filters. It adds the database name to the
    record object as the attribute mpuls_dbname. With that name it can
    be referenced in format string of a formatter. If no database name
    has been set, the value of mpuls_dbname is None.
    """

    dbnames = {}

    @classmethod
    def set_thread_dbname(cls, dbname):
        cls.dbnames[thread.get_ident()] = dbname

    @classmethod
    def get_thread_dbname(cls):
        return cls.dbnames.get(thread.get_ident())

    def filter(self, record):
        record.mpuls_dbname = self.get_thread_dbname()
        return True


class FilteredHandler(object):

    """Creation of this class returns the base_handler with a 
    DBNameContextFilter.

    In Python 2.6 a handler object needs to be a class.
    Overwritting __new__ is a way to add the filter and return the
    base_handler.
    """

    def __new__(cls, base_handler, *args):
        handler = base_handler(*args)
        handler.addFilter(DBNameContextFilter())
        return handler
