#
# Author: Ludwig Reiter
import os
import sqlite3
import datetime

INSERT_SUCCESS = """INSERT INTO login_tbl (login, last_success_login)
                    VALUES (:login, datetime('now', 'localtime'))"""
UPDATE_SUCCESS = """UPDATE login_tbl
                    SET last_success_login = datetime('now', 'localtime')
                    WHERE login = :login"""
INSERT_FAILED = """INSERT INTO login_tbl (login, last_failed_login)
                   VALUES (:login, datetime('now', 'localtime'))"""
UPDATE_FAILED = """UPDATE login_tbl
                   SET last_failed_login = datetime('now', 'localtime')
                   WHERE login = :login"""

CREATE_VERSION_TBL = """CREATE TABLE version_tbl (version INTEGER);"""
INSERT_VERSION = """INSERT INTO version_tbl (version) VALUES (1);"""
CREATE_LOGIN_TBL = """CREATE TABLE login_tbl (
                      login VARCHAR(60) PRIMARY KEY,
                      last_failed_login TIMESTAMP,
                      last_success_login TIMESTAMP);"""


class LoginDBError(StandardError):
    pass


class LoginDatabase(object):

    def __init__(self, path, check=False):
        self.dbname = path
        if not os.path.isfile(path):
            self.create_tables()
        if check:
            self.check()

    def check(self):
        self.work_with_database(self.check_helper, False)

    def create_tables(self):
        self.work_with_database(self.create_tables_helper, True)

    def store_success_login(self, login):
        self.work_with_database(self.store_login, True, login, INSERT_SUCCESS,
                                UPDATE_SUCCESS)

    def store_failed_login(self, login):
        self.work_with_database(self.store_login, True, login, INSERT_FAILED,
                                UPDATE_FAILED)

    def get_failed_login(self, login, mindate=None):
        row = self.work_with_database(self.get_login, False, login)
        if not row or not row[2]:
            return None
        if mindate:
            if row[1] >= row[2]:
                return None
        return datetime.datetime.strptime(row[2], "%Y-%m-%d %H:%M:%S")

    def get_success_login(self, login):
        row = self.work_with_database(self.get_login, False, login)
        if not row or not row[1]:
            return None
        return datetime.datetime.strptime(row[1], "%Y-%m-%d %H:%M:%S")

    def work_with_database(self, func, commit, *args):
        con = sqlite3.connect(self.dbname)
        try:
            res = func(con.cursor(), *args)
            if commit:
                con.commit()
            return res
        finally:
            con.close()

    def get_login(self, cur, login):
        cur.execute("""SELECT login, last_success_login, last_failed_login
                       FROM login_tbl WHERE login=:login""",
                    {"login": login})
        return cur.fetchone()

    def store_login(self, cur, login, insert, update):
        cur.execute("SELECT * FROM login_tbl WHERE login=:login",
                    {'login': login})
        if cur.fetchone():
            cur.execute(update, {'login': login})
        else:
            cur.execute(insert, {'login': login})

    def check_helper(self, cur):
        cur.execute("SELECT version FROM version_tbl")
        version = cur.fetchone()[0]
        if not (version == 1):
            raise LoginDBError(("Wrong version of login db."
                               " Version is %s except 1") % version)
        cur.execute("SELECT COUNT(*) FROM login_tbl")

    def create_tables_helper(self, cur):
        cur.execute(CREATE_VERSION_TBL)
        cur.execute(INSERT_VERSION)
        cur.execute(CREATE_LOGIN_TBL)
