# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Torsten Irländer <torsten.irlaender@intevation.de>
#

import logging
from datetime import datetime, timedelta
import psycopg2.extras

import mpulsweb.lib.helpers as h
from mpulsweb.lib.base import session, g
from mpulsweb.lib.db import db
from mpulsweb.lib.translation import _, ungettext
from mpulsweb.lib.exceptions import MpulsMissingItemException
from mpulsweb.model.case import MpulsCaseOverview
from mpulsweb.model.agencysettings import Agency

log = logging.getLogger(__name__)


class MissingAppointmentError(MpulsMissingItemException):

    """Indicates that an apppointment is missing.

    When this exception is raised the appointment that was to be loaded
    does not exist or the user is not allowed to see that appointment.
    """


class AppointmentFactory(object):

    def loadById(self, id):
        with db.cursor() as cur:
            cur.execute(self.get_sql, {'id': id})
            row = cur.fetchone()
            if row is not None:
                return self.createByDBRow(row)
            else:
                raise MissingAppointmentError("%s: no appointment found with"
                                              " ID %r"
                                              % (self.__class__.__name__,
                                                 id))

    def createByData(self, id=None, title=None, sdate=None, edate=None,
                     desc=None, case_id=None, type=0, case_subject=None):
        appointment = self.constructor()
        appointment.id = id
        appointment.start_date = sdate
        appointment.end_date = edate
        appointment.title = title
        appointment.description = desc
        appointment.case_id = case_id
        appointment.type = type
        appointment.case_subject = case_subject
        return appointment

    def createByDBRow(self, row, with_case_subject=False):
        id = row[0]
        master_id = row[1]
        sdate = row[2]
        edate = row[3]
        name = h.ensure_unicode(row[4])
        desc = h.ensure_unicode(row[5])
        type = row[6]
        case_subject = None
        if with_case_subject:
            case_subject = self.get_case_subject_from_row(row)
        return self.createByData(id, name, sdate, edate, desc, master_id,
                                 type, case_subject)

    def createNew(self, case_id=None):
        with db.cursor() as cur:
            cur.execute(self.create_sql,
                        {'id': case_id} if case_id is not None else {})
            return self.createByData(cur.fetchone()[0])

    def get_case_subject_from_row(self, row):
        result = {}
        result.update(row)
        result["id"] = result['master_id']
        result["s_id"] = None
        result["status"] = None
        result["zugriff"] = None
        result["phase"] = None
        result["tags"] = None
        result["editor_id"] = None
        result["bearbeiter_id"] = None
        return g.case_factory.loadFromQueryResult(result).get_subject()


class CaseAppointmentFactory(AppointmentFactory):

    GET_CASE_APPOINTMENT_SQL = """
    SELECT
    id, master_id, sdatum, edatum, name,
    beschreibung, art
    FROM ka_fall_termine_tbl_view
    WHERE id = %(id)s """
    CREATE_CASE_APPOINTMENT_SQL = "SELECT create_fall_termin_ds(%(id)s)"

    def __init__(self):
        self.get_sql = self.GET_CASE_APPOINTMENT_SQL
        self.create_sql = self.CREATE_CASE_APPOINTMENT_SQL
        self.constructor = CaseAppointment


class CaseReminderFactory(AppointmentFactory):

    GET_CASE_APPOINTMENT_SQL = """
    SELECT
    id, master_id, sdatum, edatum, name,
    beschreibung, art
    FROM ka_fall_termine_tbl_view
    WHERE id = %(id)s"""
    CREATE_CASE_APPOINTMENT_SQL = "SELECT create_fall_termin_ds(%(id)s)"

    def __init__(self):
        self.get_sql = self.GET_CASE_APPOINTMENT_SQL
        self.create_sql = self.CREATE_CASE_APPOINTMENT_SQL
        self.constructor = CaseReminder


class GlobalAppointmentFactory(AppointmentFactory):

    GET_GLOBAL_APPOINTMENT_SQL = """
    SELECT
    id, master_id, sdatum, edatum,
    name, beschreibung, art
    FROM ka_global_termine_tbl_view
    WHERE id = %(id)s"""
    CREATE_GLOBAL_APPOINTMENT_SQL = "SELECT create_global_termin_ds()"

    def __init__(self):
        self.get_sql = self.GET_GLOBAL_APPOINTMENT_SQL
        self.create_sql = self.CREATE_GLOBAL_APPOINTMENT_SQL
        self.constructor = GlobalAppointment


class AppointmentOverview(object):

    def _load(self, factory, sql, values, with_case_subject=False):
        with db.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
            cur.execute(sql, values)
            self.appointment_list = \
                [factory.createByDBRow(row, with_case_subject=with_case_subject)
                 for row in cur.fetchall()]

    def getDatasets(self):
        return self.appointment_list

    def setDatasets(self, list):
        self.appointment_list = list

    def join(self, overview):
        self.appointment_list.extend(overview.getDatasets())

    def sort(self, cmp=None):
        self.appointment_list = sorted(self.appointment_list)


class CaseAppointmentOverview(AppointmentOverview):

    GET_CASE_APPOINTMENTLIST_SQL = """
    SELECT
    id, master_id, sdatum, edatum,
    name, beschreibung, art
    FROM ka_fall_termine_tbl_view
    WHERE master_id = %(id)s AND coalesce(edatum, sdatum)::date+1
    >= %(date)s::date
    AND art = 0
    ORDER BY sdatum"""

    GET_ALLCASES_APPOINTMENTLIST_SQL = """
    SELECT a.id, a.master_id, a.sdatum,
    a.edatum, a.name, a.beschreibung, a.art, {casefields}
    FROM ka_fall_termine_tbl_view a
    JOIN master_tbl_view m ON m.id = a.master_id
    JOIN ka_status_tbl_view s ON m.id = s.master_id
    WHERE coalesce(edatum, sdatum)::date+1 >= %(date)s::date
    AND art = 0
    AND m.bearbeiter_id = %(editor)s
    AND s.status = 1 ORDER BY
    sdatum"""

    def __init__(self, caseid=None):
        afactory = CaseAppointmentFactory()
        today = datetime.now().replace(hour=0, minute=0, second=0,
                                       microsecond=0)
        if caseid:
            values = {'id': caseid, 'date': today}
            sql = self.GET_CASE_APPOINTMENTLIST_SQL
            with_case_subject = False
        else:
            user = session['USER_AUTHORIZED']
            retrieve_fields = g.case_factory.get_search_retrieve_fields()
            retrieve_fields_str = ", ".join([
                ("m.%s") % field for field in retrieve_fields])
            sql = self.GET_ALLCASES_APPOINTMENTLIST_SQL.format(
                    casefields=retrieve_fields_str)
            values = {'date': today, 'editor': user.id}
            with_case_subject = True

        self._load(afactory, sql, values, with_case_subject=with_case_subject)


class GlobalAppointmentOverview(AppointmentOverview):

    GET_GLOBAL_APPOINTMENTLIST_SQL = """
    SELECT
    id, master_id, sdatum, edatum,
    name, beschreibung, art
    FROM ka_global_termine_tbl_view
    WHERE coalesce(edatum, sdatum)::date+1 >= %(date)s::date
    ORDER BY sdatum"""

    def __init__(self):
        afactory = GlobalAppointmentFactory()
        today = datetime.now().replace(hour=0, minute=0, second=0,
                                       microsecond=0)
        values = {'date': today}
        sql = self.GET_GLOBAL_APPOINTMENTLIST_SQL
        self._load(afactory, sql, values)


class CaseReminderOverview(AppointmentOverview):

    GET_CASE_REMINDERLIST_SQL = """
    SELECT
    id, master_id, sdatum, edatum,
    name, beschreibung, art
    FROM ka_fall_termine_tbl_view
    WHERE master_id = %(id)s
    AND art = 1
    ORDER BY sdatum"""

    GET_ALLCASES_REMINDERLIST_SQL = """
    SELECT a.id, a.master_id, a.sdatum,
    a.edatum, a.name, a.beschreibung, a.art, {casefields}
    FROM ka_fall_termine_tbl_view a
    JOIN master_tbl_view m ON m.id = a.master_id
    JOIN ka_status_tbl_view s ON m.id = s.master_id
    WHERE art = 1
    AND m.bearbeiter_id = %(editor)s
    AND s.status = 1
    ORDER BY sdatum"""

    def __init__(self, caseid=None):
        afactory = CaseReminderFactory()
        today = datetime.now().replace(hour=0, minute=0, second=0,
                                       microsecond=0)
        if caseid:
            values = {'id': caseid, 'date': today}
            sql = self.GET_CASE_REMINDERLIST_SQL
            with_case_subject = False
        else:
            user = session['USER_AUTHORIZED']
            retrieve_fields = g.case_factory.get_search_retrieve_fields()
            rfstr = ", ".join([("m.%s") % field for field in retrieve_fields])
            values = {'date': today, 'editor': user.id}
            sql = self.GET_ALLCASES_REMINDERLIST_SQL.format(casefields=rfstr)
            with_case_subject = True
        self._load(afactory, sql, values, with_case_subject=with_case_subject)


class Appointment(object):

    def __init__(self):
        self.id = None
        self.title = None
        self.start_date = None
        self.end_date = None
        self.description = None
        self.case_id = None
        # available types
        # 0: normal appointment
        # 1: normal Reminder
        # 2: automatic Reminder
        self.type = 0
        self.case_subject = None

    def __cmp__(self, other):
        return cmp(self.start_date, other.start_date)

    def _convertStr2Date(self, str, format=None):
        str = str.strip()
        date, time = str.split(' ')
        try:
            d, m, Y = date.split('.')
        except:
            return None
        try:
            H, M = time.split(':')
        except:
            H, M = (0, 0)
        return datetime(int(Y), int(m), int(d), int(H), int(M))

    def _convert2Datetime(self, mdate, mtime, ddate=None, dtime=None):
        log.debug("%s, %s" % (mtime, type(mtime)))
        try:
            date, time = (False, False)
            try:
                y = mdate.year
                m = mdate.month
                d = mdate.day
                date = True
            except:
                date = False

            try:
                H = mtime.hour
                M = mtime.minute
                S = mtime.second
                time = True
            except:
                time = False

            if date and time:
                return datetime(y, m, d, H, M, S)
            elif date:
                H = dtime.hour
                M = dtime.minute
                S = dtime.second
                return datetime(y, m, d, H, M, S)
            elif time:
                y = ddate.year
                m = ddate.month
                d = ddate.day
                return datetime(y, m, d, H, M, S)

            return None

        except Exception, e:
            log.error(e)
            return None

    def getStartDate(self):
        return self.start_date

    def getEndDate(self):
        return self.end_date

    def getDuration(self, empty=''):
        if self.end_date:
            return h.human_timedelta(self.end_date - self.start_date)
        return empty

    def setData(self, data):
        """Sets values of the appointment to data.
        Data is a dict returned from formencode
        """
        self.title = data.get('title', self.title)
        self.type = data.get('type', self.type)

        self.start_date = self._convert2Datetime(data.get('start_date'),
                                                 data.get('start_time'))

        end_date = self._convert2Datetime(data.get('end_date'),
                                          data.get('end_time'),
                                          data.get('start_date'),
                                          data.get('start_time'))

        # If convert of enddate fail
        if not end_date:
            self.end_date = self.start_date + timedelta(hours=24)
        else:
            self.end_date = end_date

        self.description = data.get('description', self.description)
        self.case_id = data.get('case_id', self.case_id)

    def store(self):
        try:
            fields = {'id': self.id, 'case_id': self.case_id, 'type': self.type,
                      'title': self.title, 'description': self.description,
                      'start_date': self.start_date, 'end_date': self.end_date,
                      }
            db.execute(self.store_sql, fields)
            return True
        except Exception:
            log.exception("Error while storing appointment %r for case %r"
                          % (self.id, self.case_id))
        return False

    def isGlobal(self):
        return self.case_id is None

    def isReminder(self):
        return self.type == 1

    def delete(self):
        db.execute(self.delete_sql, {'id': self.id})
        return True

    def getCaseSubject(self):
        if self.case_subject is not None:
            return self.case_subject
        if self.case_id is not None:
            case = g.case_factory.loadById(self.case_id)
            return case.get_subject()
        return ""


class GlobalAppointment(Appointment):

    DELETE_GLOBAL_APPOINTMENT_SQL = "SELECT delete_global_termin_ds(%(id)s)"
    STORE_GLOBAL_APPOINTMENT_SQL = """
    UPDATE ka_global_termine_tbl_view
    SET sdatum = %(start_date)s::timestamp,
    edatum = %(end_date)s::timestamp,
    name = %(title)s,
    beschreibung = %(description)s,
    art = %(type)s
    WHERE id = %(id)s"""

    def __init__(self):
        Appointment.__init__(self)
        self.type = 0
        self.store_sql = self.STORE_GLOBAL_APPOINTMENT_SQL
        self.delete_sql = self.DELETE_GLOBAL_APPOINTMENT_SQL


class CaseAppointment(Appointment):

    DELETE_CASE_APPOINTMENT_SQL = "SELECT delete_fall_termin_ds(%(id)s)"
    STORE_CASE_APPOINTMENT_SQL = """
    UPDATE ka_fall_termine_tbl_view
    SET sdatum = %(start_date)s::timestamp,
    edatum = %(end_date)s::timestamp,
    name = %(title)s,
    beschreibung = %(description)s,
    art = %(type)s
    WHERE id = %(id)s"""

    def __init__(self):
        Appointment.__init__(self)
        self.type = 0
        self.store_sql = self.STORE_CASE_APPOINTMENT_SQL
        self.delete_sql = self.DELETE_CASE_APPOINTMENT_SQL


class CaseReminder(CaseAppointment):

    def __init__(self):
        CaseAppointment.__init__(self)
        self.type = 1

# Tags


class TagsReminderOverview(AppointmentOverview):
    GET_CASES_WITH_TAGS = """
    SELECT m.id
    FROM master_tbl_view m
    JOIN ka_status_tbl_view s ON s.master_id = m.id
    JOIN nm_tags_cases_view t ON t.case_id = m.id
    WHERE s.status = 1
    AND m.bearbeiter_id = %(editor)s
    AND t.tag_id = %(tags)s
    """

    GET_CASES_WITH_TAGS_ADMIN = """
    SELECT m.id
    FROM master_tbl_view m
    JOIN ka_status_tbl_view s ON s.master_id = m.id
    JOIN nm_tags_cases_view t ON t.case_id = m.id
    WHERE s.status = 1
    AND t.tag_id = %(tags)s
    """

    def __init__(self):
        self.appointment_list = []
        user = session['USER_AUTHORIZED']
        reminder_tags = g.mpuls_config.get('overview_tags', 'tag_ids')

        # Return if there are no "important" tags defined
        if not reminder_tags:
            return

        cases = MpulsCaseOverview()

        # 1. Load the user's cases so we can check their tags
        sm = g.case_factory.get_state_machine()
        cases.search({"own": user.id, "standin": user.id,
                      "state": [1, 2, 3],
                      "phase": sm.get_search_item_ids()})

        h.tag_appointment = True

        # Make appointments for all cases that have the "important"-tags
        for tag in reminder_tags:
            case_list = {}
            with db.cursor() as cur:
                if user.isAdmin():
                    cur.execute(self.GET_CASES_WITH_TAGS_ADMIN,
                                {'tags': int(tag)})
                else:
                    cur.execute(self.GET_CASES_WITH_TAGS,
                                {'editor': self.getUserId(), 'tags': int(tag)})
                result = cur.fetchall()

                for r in result:
                    case_list[r[0]] = {}

            cases_with_tags_set = set(case_list)
            cases_with_tags = [case for case in cases.getDatasets()
                               if case.id in cases_with_tags_set]

            log.debug('cases with tags: %s' % case_list)
            for case in cases_with_tags:
                appointment = self.makeAppointment(case,
                                                   g.taggingConf.get_desc(tag),
                                                   g.taggingConf.get_name(tag))
                self.appointment_list.append(appointment)

    def makeAppointment(self, case, description, title):
        factory = CaseAppointmentFactory()
        appointment_type = 2

        appointment = factory.createByData(None,
                                           title,
                                           None,
                                           None,
                                           description,
                                           case.id,
                                           appointment_type)
        return appointment

    def getUserId(self):
        """
        Return the User-ID. This function can be overwritten by applications
        that have a "structure" part where the user that owns the cases is not
        the user logged in.
        """
        user = session['USER_AUTHORIZED']
        return user.id


# MAXSAVETIME

class MaxSaveTimeReminderOverview(AppointmentOverview):

    GET_OLDFINISHED_CASES = """
    SELECT m.id, s.zugriff,
    getLastCaseAppointment(m.id) as last_date
    FROM master_tbl_view m
    JOIN ka_status_tbl_view s ON s.master_id = m.id
    WHERE s.status <> 5 AND s.status <> 4 AND s.status <> 3
    AND m.bearbeiter_id = %%(editor)s
    AND extract(day from now()-s.zugriff) > %%(maxdays)s
    AND phase in %%(phases)s
    AND (%s)"""

    GET_OLDFINISHED_CASES_4_ADMIN = """
    SELECT m.id, s.zugriff,
    getLastCaseAppointment(m.id) as last_date
    FROM master_tbl_view m
    JOIN ka_status_tbl_view s ON s.master_id = m.id
    WHERE s.status <> 5
    AND extract(day from now()-s.zugriff) > %%(maxdays)s
    AND phase in %%(phases)s
    AND (%s)"""

    GET_OLDACTIVE_CASES = """
    SELECT m.id, s.zugriff,
    getLastCaseAppointment(m.id) AS last_date
    FROM (
        SELECT * FROM ka_status_tbl_view
        WHERE zugriff < now() - interval '1 day' * %%(maxdays)s
        AND status <> 5 AND status <> 4 AND status <> 3) s
    JOIN master_tbl_view m ON m.id = s.master_id
    WHERE phase in %%(phases)s
    AND (%s)
    AND m.bearbeiter_id = %%(editor)s"""

    GET_OLDACTIVE_CASES_4_ADMIN = """
    SELECT m.id, s.zugriff,
    getLastCaseAppointment(m.id) AS last_date
    FROM (
        SELECT * FROM ka_status_tbl_view
        WHERE zugriff < now() - interval '1 day' * %%(maxdays)s
        AND status <> 5) s
    JOIN master_tbl_view m ON m.id = s.master_id
    WHERE phase in %%(phases)s
    AND (%s)"""

    def __init__(self):
        self.appointment_list = []
        agency = Agency()
        maxage = agency.getMaxSavetime()
        cases = MpulsCaseOverview()
        user = session['USER_AUTHORIZED']

        # 0. Load ids and some additional info on cases which either are active
        # but seems to be orphaned or are finished and not edited for
        # longer than maxage.
        # orphaned means that all last appointments, aids or the
        # modification date is older than maxage
        oldAndActive = self.getOldAndActiveCases(maxage, user,
                                                 agency.getReminderCondition())
        oldAndFinished = self.getOldAndFinishedCases(
            maxage, user, agency.getReminderCondition())

        # 1. Load the user's cases so we can check the maximum savetime on them
        sm = g.case_factory.get_state_machine()
        cases.search({"own": user.id, "standin": user.id, "state": [1, 2, 3],
                      "phase": sm.get_search_item_ids()})

        # 2. filter cases which seem to be orphaned and create reminders
        oldAndActiveSet = set(oldAndActive)
        old_cases = [case for case in cases.getDatasets()
                     if case.id in oldAndActiveSet]
        log.debug('oldcases: %s' % old_cases)
        for case in old_cases:
            info = oldAndActive[case.id]
            last = info['last_date']
            appointment = self.makeAppointment(case, last, maxage,
                                               finished=False)
            self.appointment_list.append(appointment)

        # 3. filter cases which are finished and not edited and create
        # reminders
        oldAndFinishedSet = set(oldAndFinished)
        old_cases = [case for case in cases.getDatasets()
                     if case.id in oldAndFinishedSet]
        for case in old_cases:
            info = oldAndFinished[case.id]
            access = info['access']
            appointment = self.makeAppointment(case, access, maxage,
                                               finished=True)
            self.appointment_list.append(appointment)

    def getOldAndFinishedCases(self, days, user, remindercondition):
        list = {}
        phases = tuple(g.case_factory.get_state_machine().get_end_state_ids())
        with db.cursor() as cur:
            if user.isAdmin():
                sql = self.GET_OLDFINISHED_CASES_4_ADMIN % remindercondition
                cur.execute(sql, {'maxdays': days,
                                  'phases': phases})
            else:
                sql = self.GET_OLDFINISHED_CASES % remindercondition
                cur.execute(sql, {'maxdays': days,
                                  'editor': user.id,
                                  'phases': phases})
            result = cur.fetchall()
            for r in result:
                list[r[0]] = {'access': r[1], 'last_date': r[2]}
            return list

    def getOldAndActiveCases(self, days, user, remindercondition):
        list = {}
        phases = tuple(g.case_factory.get_state_machine()
                       .get_active_state_ids())
        with db.cursor() as cur:
            if user.isAdmin():
                sql = self.GET_OLDACTIVE_CASES_4_ADMIN % remindercondition
                cur.execute(sql, {'maxdays': days,
                                  'editor': user.id,
                                  'phases': phases})
            else:
                sql = self.GET_OLDACTIVE_CASES % remindercondition
                cur.execute(sql, {'maxdays': days,
                                  'editor': user.id,
                                  'phases': phases})
            result = cur.fetchall()
            for r in result:
                list[r[0]] = {'access': r[1], 'last_date': r[2]}
            return list

    def makeAppointment(self, case, last_date, maxage, finished):
        factory = CaseAppointmentFactory()
        type = 2
        title = u"%s, %s" % (case.last_name, case.first_name)
        age = datetime.now() - last_date
        mydate = datetime.now() - timedelta(age.days - int(maxage))
        if finished:
            desc = (ungettext('Attention! The maximum save duration is '
                      'exceeded for %s day for this case document!',
                      'Attention! The maximum save duration is '
                      'exceeded for %s days for this case document!',
                             age.days - int(maxage))
                    %  (age.days - int(maxage)))
        else:
            desc = (_(u"Inactive case document? Please check,"
                      u" if the case will still be edited."))
        return factory.createByData(None, title, mydate, None, desc,
                                    case.id, type)

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf-8:
