# -*- coding: utf-8 -*-

import logging
import re

from pylons import app_globals as g
from mpulsweb.lib.translation import _

from formed.instance.backends.common import RuleEvalContext
from formed.instance.tree import SetDataException
from formed.meta.data import PageNode

from mpulsweb.lib.helpers import tag, literal, get_formular_url, getRole, \
    deprecated
from mpulsweb.lib.helper.case import walk_it
from mpulsweb.lib.exceptions import MpulsException

# name of the dbfield for phase
PHASEFIELD = "phase"
# name of flags
REQUIRED = re.compile('required')
PHASEDATE = re.compile('time')

# CHANGEMODES
# Can be configured json config
PC_MANUAL = '0'
PC_HALFAUTO = '1'
PC_FULLAUTO = '2'

log = logging.getLogger(__name__)


@deprecated("Please use a StateMachine instead (get_successor_state_ids)")
def get_phase_successors(p, phaselist=None):
    suc = []
    if phaselist is None:
        phaselist = PhaseFactory().build()
    for phase in phaselist:
        if p == phase.id:
            suc.append(phase.start.id)
            suc.append(phase.end.id)
            for ppid in phase.end.suc:
                pp = phaselist.get_phase(ppid)
                suc.extend(get_phase_successors(pp.phase, phaselist))
    return suc


class ConsistenceCheckException(MpulsException):

    def __init__(self, value):
        self.value = value


class PhaseAnonymizableCheckException(ConsistenceCheckException):

    def __init__(self, value):
        self.value = value


class PhaseFactory(object):

    @deprecated("Please use a StateMachine instead")
    def build(self):
        '''Build phaselist from metadata'''
        return PhaseList()

    @deprecated("Please use a StateMachine instead")
    def load(self, case_id):
        '''Build phaselist from metadata and load current state from instance
        tree'''
        return PhaseList(case_id)


class PhaseList(list):

    def __init__(self, case_id=None):
        """ If case_id is not None the current state of the case is loaded from
        the instance tree """

        self.suc_phase_parts = {}
        self.pre_phase_parts = {}

        self._build()

        if case_id is not None:
            self._load(case_id)

    def _get_suc_phase_part(self, phase_id, suc_conf, pairs_conf):
        """ Returns a list of start phase parts ids for the phase """

        ps = []
        s = suc_conf.get(phase_id)
        for p in s:
            start = pairs_conf.get(p)[0]
            ps.append(int(start))
        return ps

    def _build(self):
        """ Build successors and predessesors of the phases """

        pairs_conf = g.mpuls_config.get('phases', 'pairs')[0]
        phase_ids = g.mpuls_config.get('phases', 'phases')
        suc_conf = g.mpuls_config.get('phases', 'successors')[0]
        phases_desc = g.mpuls_config.get('phases', 'description')[0]

        for p in phase_ids:
            pp = pairs_conf.get(p)
            if len(pp) == 2:
                start_phase_part = int(pp[0])
                end_phase_part = int(pp[1])
                self.suc_phase_parts[start_phase_part] = [end_phase_part]
                self.suc_phase_parts[end_phase_part] = self._get_suc_phase_part(
                    p, suc_conf, pairs_conf)

        for k, v in self.suc_phase_parts.iteritems():
            for p in v:
                self.pre_phase_parts[p] = [k]

        for p in phase_ids:
            pair = pairs_conf.get(p)
            start_id = int(pair[0])
            end_id = int(pair[1])
            desc = phases_desc.get(p)
            start = StartPhasePart(start_id, p,
                                   self.suc_phase_parts.get(start_id, []),
                                   self.pre_phase_parts.get(start_id, []))
            end = EndPhasePart(end_id, p,
                               self.suc_phase_parts.get(end_id, []),
                               self.pre_phase_parts.get(end_id, []))
            self.append(Phase(int(p), start, end, desc))

    def _load(self, case_id):
        """ Loads the current state from the instance tree """

        factory = g.get_instance_factory()
        it = factory.loadInstanceTreeByIdentifier(case_id)
        for p in self:
            start = p.start
            end = p.end
            start.load(it)
            end.load(it)

    def get_unknown_phase_part(self):
        p = EndPhasePart(-1, None, [1], [])
        p.phase = -1
        return p

    def get_unknown_phase(self):
        return Phase(-1, self.get_unknown_phase_part(), None, _("Unkown"))

    def get_current_phase_id(self):
        '''Returns id of current active phasepart'''
        p = self.get_current_phase()
        return p.id

    def get_current_phase(self):
        '''Returns current active phasepart'''
        for phase in self:
            p1 = phase.getStart()
            if p1.is_active():
                return p1
            p2 = phase.getEnd()
            if p2.is_active():
                return p2

        # Return dummy phase for unknown phase
        return self.get_unknown_phase_part()

    def get_current(self):
        '''Returns the current phase'''
        for phase in self:
            p1 = phase.getStart()
            if p1.is_active():
                return phase
            p2 = phase.getEnd()
            if p2.is_active():
                return phase

        log.warn("No current phase found returning Unkonwn dummy phase")
        return self.get_unknown_phase()

    def get(self, id):
        """
        Returns the Phase instance in this PhaseList corresponding to the id
        """
        for p in self:
            if p.id == id:
                return p
        return None

    def get_phase(self, id):
        '''Returns the PhasePart of the id'''
        for p in self:
            if id == p.start.id:
                return p.start
            if id == p.end.id:
                return p.end

        # Return dummy phasepart for unknown phasepart id
        if id == -1:
            return self.get_unknown_phase_part()

    def get_max_phase_id(self):
        return max(self.get_reachable_path())

    def get_reachable_path(self, phasepart_id=0, only_reachable=True):
        '''Returns id of max reachable phase which is valid starting from
        phasepart_id'''

        path = [-1]  # default unknown phase
        if not self.get_phase(phasepart_id).is_ok() and only_reachable:
            return path
        for pp in self.suc_phase_parts.get(phasepart_id, []):
            if not self.get_phase(pp).is_ok() and only_reachable:
                continue
            path.extend(self.get_reachable_path(pp, only_reachable))
            path.append(pp)
        path.append(phasepart_id)
        return path

    def get_path(self, phasepart_id, only_reachable=False):
        '''Returns a list of phaseparts id from start to the given
        phasepart_id'''
        path = []
        for pp in self.pre_phase_parts.get(phasepart_id, []):
            # Ignore phase which are not reachable
            if not self.get_phase(pp).is_ok() and only_reachable:
                continue
            path.extend(self.get_path(pp))
            path.append(pp)
        path.append(phasepart_id)
        return path

    def is_on_path(self, find_id, pathend_id, only_reachable=False):
        return find_id in self.get_path(pathend_id, only_reachable)

    def is_on_reachable_path(self, find_id, start_id=0, only_reachable=True):
        """ Returns True if id of phase part (find_id) is on the reachable path
        """
        return find_id in self.get_reachable_path(start_id, only_reachable)

    def is_valid(self):
        '''Returns true if all phaseparts up to the current phase are ok.
        This means all requierd fields are filled out and all checks for
        rules are ok. Else return False
        '''
        try:
            self.check_consistence()
            return True
        except ConsistenceCheckException:
            return False

    def check_consistence(self, fields=None, phase=None):
        CASE_NOT_CONSISTENT = _("The phase '%s' is in a non consitent \
                                state and has missing required fields: %s")
        if phase is not None:
            current_phase = phase
        else:
            current_phase = self.get_current_phase_id()
        path = self.get_path(str(current_phase))
        for phases in self:
            p1, p2 = phases.getStart(), phases.getEnd()
            if p1.id <= current_phase and p1.id in path:
                if not p1.is_ok(fields, check_setting_phase=False):
                    errors = []
                    errors.append('<ol>')
                    for e in p1.errors:
                        errors.append('<li>%s</li>' % e)
                    errors.append('</ol>')
                    raise ConsistenceCheckException(CASE_NOT_CONSISTENT
                                                    % (phases.description,
                                                       "".join(errors)))
            if p2.id <= current_phase and p2.id in path:
                if not p2.is_ok(fields, check_setting_phase=False):
                    errors = []
                    errors.append('<ol>')
                    for e in p2.errors:
                        errors.append('<li>%s</li>' % e)
                    errors.append('</ol>')
                    raise ConsistenceCheckException(CASE_NOT_CONSISTENT
                                                    % (phases.description,
                                                       "".join(errors)))


class Phase(object):

    def __init__(self, id, start, end, description):
        self.id = id
        self.start = start
        self.end = end
        self.description = description

    def getStart(self):
        return self.start

    def getEnd(self):
        return self.end

    def getDuration(self):
        t1 = self.start.getTime()
        t2 = self.end.getTime()
        if not t1 or not t2:
            return None
        return t2 - t1

    def getDescription(self):
        return literal(self.description)

    def isRunning(self):
        return self.start.is_active()


class PhasePart(object):

    def __init__(self, id, phase_id, suc, pre):
        self.id = id
        self.it = None
        self.phase = phase_id
        self.active = False
        self.date = None
        self.datefield = None
        self.suc = suc
        self.pre = pre
        self.fields = []
        self.errors = []

        # If id is 0 (which should be the start phasepart of the first phase,
        # add "unknown" phase as predesessor)
        if self.id == 0:
            self.pre.append(-1)

        # Get phase relevat dates
        for p in g.formedTree.walk():
            if self._is_in_flag(p, "time"):
                self.datefield = p.getName()

        if self.datefield is None and self.phase != -1:
            log.info("Phase %s has no associated datefield", id)

    def getDescription(self):
        desc = [g.mpuls_config.get('phases',
                                   'description')[0][str(self.phase)]]
        if (g.mpuls_config.is_enabled("phases", "display_with_running_state")
                and self.phase != -1):
            if isinstance(self, StartPhasePart):
                running = _('Running')
            else:
                running = _('Finished')
            desc.append(running)
        return " ".join(desc)

    def getIcon(self):
        icon = g.mpuls_config.get('phases', 'icons')[0].get(str(self.phase))
        return icon or self.getDescription()[0].lower()

    def load(self, it):
        formed = it.getFormEd()
        self.it = it
        # Check if the current phasepart is active
        case_id = it.getRootNode().getIdentifier()
        realid = "%s:%s" % (PHASEFIELD, case_id)
        field = it.getItem(realid)
        self.active = field.getValue() == self.id

        for p in walk_it(it.getRootNode()):
            meta = p.getMeta()
            if isinstance(meta, PageNode):
                page = meta.getName()

            name = meta.getName()
            identifier = p.getIdentifier()
            # Check if this fields is a phase relvant required field

            if self._is_in_flag(meta, "required"):
                # Only add warnigs and rules with "required" in der name
                rules = [rule for rule in formed.name2markingrules.get(name, ())
                         if rule.isWarning() and REQUIRED.match(rule.getName())]
                self.fields.append(Field(identifier, page, rules, it))

        # Load phase relvant datefield
        if self.datefield is not None:
            realid = "%s:%s" % (self.datefield, case_id)
            field = it.getItem(realid)
            self.date = field.getValue()

    def _is_in_flag(self, meta, flag):
        flags = meta.get_parsed_flags()

        time = flags[flag]
        for p in time:
            sp = "phase_%s" % self.id
            if p == sp:
                return True

        return False

    def is_neighbor(self, phase):
        return (phase in self.suc) or (phase in self.pre)

    def getLinks(self):
        case_id = self.it.getRootNode().getIdentifier()
        pages = {}
        key2page = {}
        tree_order = []
        order = []
        p2 = {}
        if len(self.fields) > 0:
            for field in [field for field in self.fields]:
                if field.page not in p2:
                    tree_order.append(field.page)
                    p2[field.page] = []
                p2[field.page].append(field)
        for p, f in p2.iteritems():
            if p == 'ee':
                continue  # (hidden)
            realid = "%s:%s" % (p, case_id)
            item = self.it.getItem(realid)
            key = item.getMeta().getDescription()
            key2page[p] = key
            pages[key] = f
        # Ordering based of treestructure
        for o in tree_order:
            order.append(key2page.get(o))
        return pages, order

    def getTime(self):
        return self.date

    def is_active(self):
        return self.active

    def is_anonymizable(self):
        return self.id not in g.mpuls_config.get('phases', 'notanonymizable')

    def is_ok(self, fields=None, check_setting_phase=True):

        # 1. Simulate switching the phase
        # Setting values will trigger evaluation error rules. This way we can
        # check logic rules which are modeled as errors.
        errors = []
        if check_setting_phase and self.it:
            case_id = self.it.getRootNode().getIdentifier()
            realid = "%s:%s" % (PHASEFIELD, case_id)
            try:
                self.it.setData({realid: str(self.id)})
            except SetDataException:
                return False

        # 2. Check rules
        for field in self.fields:
            if fields is None:
                # No fields provided -> checkall
                if not field.is_ok():
                    errors.append(field.label)
            elif field.name in fields:
                if not field.is_ok():
                    errors.append(field.label)
        self.errors = errors
        return len(errors) == 0

    def get_phase_name(self):
        return g.mpuls_config.get('phases', 'description')[0].get(self.phase)

    def __repr__(self):
        return "<PhasePart 0%x ok=%r active=%r id=%r name=%r " \
               "description=%r>" % (id(self), self.is_ok(), self.is_active(),
                                    self.id, self.get_phase_name(),
                                    self.getDescription())


class StartPhasePart(PhasePart):
    pass


class EndPhasePart(PhasePart):
    pass


class Field(object):

    def __init__(self, id, page, rules, it):
        self.id = id
        self.page = page
        self.it = it
        self.label = None
        self.name = None
        self.conditions = []

        field = self.it.getItem(self.id)
        meta = field.getMeta()

        self.name = meta.getName()
        self.label = meta.getDescription()

        for rule in rules:
            self.conditions.append(Condition(field, rule, it))

    def get_url(self):
        role = getRole()
        if role == 'Verwaltung':
            action_arg = 'edit'
        else:
            action_arg = 'select'
        return get_formular_url(self.it.getFormEd(), self.id, self.name,
                                self.page, action_arg=action_arg)

    def get_link(self):
        out = []
        if self.is_ok():
            style = "ok"
            atitle = _('Field is OK and requires no further editing')
        else:
            style = "error"
            atitle = _('Field is not OK and requires further editing')
        out.append(tag("a", class_=style,
                       href=self.get_url(),
                       title_=atitle))
        out.append('%s' % self.label)
        out.append('</a>')
        return "".join(out)

    def has_relevant_rules(self):
        '''Returns true if the field has relevant rules. Else False e.g if the
        required field is in a condition which is not true'''
        for c in self.conditions:
            if c.is_relevant():
                return True
        return False

    def is_ok(self):
        for c in self.conditions:
            if not c.is_ok():
                return False
        return True

    def __repr__(self):
        return "<Field 0%x ok=%r id=%r page=%r name=%r>" % (
            id(self), self.is_ok(), self.id, self.page, self.name)


class Condition(object):

    def __init__(self, field, rule, it):
        self.field = field
        self.rule = rule
        self.it = it

    def is_ok(self):
        '''Returns True if expr is valid for the current value in field
        identified with name'''
        return self.evaluate()

    def is_relevant(self):
        id = self.field.getIdentifier()
        context = RuleEvalContext(self.it, id)
        return self.it.rule_available(self.rule, context)

    def evaluate(self):
        id = self.field.getIdentifier()
        context = RuleEvalContext(self.it, id)
        if not self.is_relevant():
            return True
        expr = self.rule.getExpr()
        for result, variables in expr.evaluate_instances(context):
            if not result and id in variables:
                return False
            return True

    def __repr__(self):
        return "<Condition 0%x ok=%r rule=%r>" % (
            id(self), self.is_ok(), self.rule)

# vim:set ts=4 sw=4 si et sta sts=4 :
