# -*- coding: utf-8 -*-
#
# Copyright (c) 2007 - 2011, 2013, 2014 by Intevation GmbH
#
# FormEd is free software under the terms of the
# GNU General Public License Version 3. See file
# LICENSE coming with the source of FormEd for details.
#
# Authors:
# Torsten Irlaender   <torsten.irlaender@intevation.de>
# Sascha L. Teichmann <teichmann@intevation.de>
# Frank Koormann      <frank.koormann@intevation.de>
# Björn Ricks         <bjoern.ricks@intevation.de>
#

import sys
import codecs
from time import gmtime, strftime

from formed.meta import data
from formed.meta.nodecomponents import Node, Leaf, \
VISIT_IGNORE_CHILDREN, VISIT_CONTINUE
from formed.meta.structure import structure_path
from formed.export.rules import ExportSQLForVirtualFields

VIEW_WHERE_CLAUSE = \
"""WHERE
    master_tbl.bearbeiter_id = getuserid()
    OR master_tbl.id IN (
        SELECT nm.master_id from nm_benutzer_master_tbl nm
        WHERE nm.benutzer_id IN (
            SELECT nm2.gruppe_id from nm_benutzer_gruppe_tbl nm2
            WHERE nm2.benutzer_id = getuserid()
        )
    )
    OR pg_has_role('admin','MEMBER');
"""

FORBID_DELETE_RULE = \
"""CREATE OR REPLACE RULE %s_delete_rule AS ON DELETE TO %s_view
    DO INSTEAD NOTHING;
"""

FORBID_INSERT_RULE = \
"""CREATE OR REPLACE RULE %s_insert_rule AS ON INSERT TO %s_view
    DO INSTEAD NOTHING;
"""

FORBID_UPDATE_RULE = \
"""CREATE OR REPLACE RULE %s_update_rule AS ON UPDATE TO %s_view
    DO INSTEAD NOTHING;
"""

UPDATE_RULE = \
"""CREATE OR REPLACE RULE %s_update_rule AS ON UPDATE TO %s_view
DO INSTEAD UPDATE %s SET
"""

GRANT_SELECT = \
"""GRANT SELECT ON %s TO GROUP %s;
"""

GRANT_INSERT = \
"""GRANT INSERT ON %s TO GROUP %s;
"""

GRANT_UPDATE = \
"""GRANT UPDATE ON %s TO GROUP %s;
"""

ANON_FUNCTION_HEADER = \
"""--
-- Function to anonymize a case and all of its repeat groups
--
CREATE OR REPLACE FUNCTION anonymize_case(mid int)
RETURNS void AS
$$
BEGIN
    PERFORM anonymize_uuid(mid);
"""

ANON_UPDATE = \
"""    UPDATE %s_view SET
        %s
    WHERE %s;

"""

ANON_FUNCTION_FOOTER = \
"""END;
$$ LANGUAGE plpgsql VOLATILE SECURITY DEFINER;

REVOKE ALL ON FUNCTION anonymize_case(int) FROM public;
GRANT EXECUTE ON FUNCTION anonymize_case(int) TO GROUP :admin_group;

"""

EVALUATION_VIEW= 'master_tbl_eval_total_view'

def removeTbl(s):
    return s.replace("_tbl", "")


def sqlQuote(name):
    return name.replace("'", "''")


def sqlName(name):
    return name.replace("-", "_")


class EvaluateColumnsCollector(object):

    def __init__(self, root=None):
        self.columns = []
        self.root = root

    def visitor(self, nc):
        if isinstance(nc, Node):
            if isinstance(nc, (data.RadioNode, data.ChoiceNode)):
                self.checkEvaluate(nc)
                return VISIT_IGNORE_CHILDREN
            if isinstance(nc, data.RepeatNode):
                return VISIT_IGNORE_CHILDREN
        elif (isinstance(nc, Leaf)
              and not isinstance(nc, (data.RuleLeaf,
                                      data.ExternalChoiceListLeaf,
                                      data.InfoLeaf))):
            self.checkEvaluate(nc)

        return VISIT_CONTINUE

    def checkEvaluate(self, nc):
        flags = nc.getFlags()
        if flags.find('evaluation') >= 0:
            self.columns.append(sqlName(nc.getName()))


class RelationAnonymizer(object):

    def __init__(self, raw_table, table, joins):
        self.name = table
        anonym = []
        self.joins = joins

        for c in raw_table.columns:
            if not c.func:
                # no set needed
                continue
            if c.func == "suppress_value":
                anonym.append('%s = %s' % (c.name, c.anon_value))
            else:
                args = [x.strip() for x in c.func.split(':')]
                if len(args) > 1:
                    anonym.append("%s = %s(%s, %s)"
                                  % (c.name, args[0], c.name,
                                     ", ".join(args[1:])))
                else:
                    anonym.append("%s = %s(%s)" % (c.name, args[0], c.name))

        if not anonym: # empty -> not sets
            self.processed = True
            self.set_columns = ""
        else:
            self.processed = False
            self.set_columns = ",\n        ".join(anonym)

    def writeUpdate(self, pfeil, where_clause="master_id = mid"):
        if self.processed:
            return
        self.processed = True
        if len(self.joins) == 2:
            where_clause = \
                ("""%s_id IN (SELECT id FROM %s WHERE master_id = mid)"""
                 % (self.joins[0][:-4], self.joins[0]))
        elif len(self.joins) > 2:
            # TODO: x->y->master_tbl
            # should result in
            #  x_id IN (SELECT x.id FROM x JOIN y
            #           ON x.y_id = y.id AND y.master_id = mid)
            raise NotImplementedError
        pfeil(ANON_UPDATE % (self.name, self.set_columns, where_clause))


class CaseAnonymizer(object):

    def __init__(self):
        self.anonymizers = []

    def addAnoymizer(self, anonymizer):
        self.anonymizers.append(anonymizer)

    def _findMaster(self):
        for rel in self.anonymizers:
            if rel.name == 'master_tbl':
                return rel
        return None

    def writeFunction(self, pfeil):
        master = self._findMaster()
        if master is None:
            print >> sys.stderr, "master_tbl not found"
            return
        pfeil(ANON_FUNCTION_HEADER)
        pfeil("    -- anonymize master table\n")
        master.writeUpdate(pfeil, "id = mid")

        pfeil("    -- anonymize repeat groups\n")
        for anonymizer in self.anonymizers:
            anonymizer.writeUpdate(pfeil)

        pfeil(ANON_FUNCTION_FOOTER)


class Column(object):

    def __init__(self, name, datatype, accessible=True, func=None,
                 anon_value='NULL'):
        self.name = name
        self.datatype = datatype
        self.accessible = accessible
        self.func = func
        self.anon_value = anon_value

    def definition(self):
        return "%s %s" % (self.name, self.datatype)


class VirtualColumn(Column):

    pass


class Table(object):

    def __init__(self, name, parent=None, root=None):
        self.name = name
        self.columns = []
        self.constraints = []
        self.inserts = []
        self.depends = set()
        self.parent = parent
        self.root = root

    def addDependency(self, dependency):
        self.depends.add(dependency)

    def appendColumn(self, column):
        self.columns.append(column)

    def appendConstraint(self, constraint):
        self.constraints.append(constraint)

    def appendInsert(self, ins):
        self.inserts.append(ins)

    def getName(self):
        return self.name

    def getSchema(self):
        out = ["CREATE TABLE %s (\n    " % self.name]
        out.append(",\n    ".join([column.definition()
                                   for column in self.columns
                                   if not isinstance(column, VirtualColumn)]))
        if self.constraints:
            out.append(",\n    ")
            out.append(",\n    ".join(self.constraints))

        out.append("\n);")
        return "".join(out)

    def getInserts(self):
        if not self.inserts:
            return ""
        out = []
        for ins in self.inserts:
            out.append("INSERT INTO %s %s;" % (self.name, ins))
        return "\n".join(out)

    def equalContent(self, other):
        # XXX: This is a bit to simple for general case
        if len(self.inserts) != len(other.inserts):
            return False
        ains = [a.lower() for a in self.inserts]
        bins = [b.lower() for b in other.inserts]
        ains.sort()
        bins.sort()
        return ains == bins


class SchemaCreator(object):

    def __init__(self, root):
        self.root = root
        self.tableStack = []
        self.tables = []

    def findSameContent(self, table):
        same_content = None
        for t in self.tables:
            if t.name == table.name:
                if t.equalContent(table):
                    return t
                else:
                    raise ValueError("Existing table %r does not have the same"
                                     " contents as the new table named %r"
                                     " with inserts %r"
                                     % (t.name, table.name, table.inserts))
            if t.equalContent(table):
                same_content = t
        return same_content

    def add_choice_table(self, table):
        other = self.findSameContent(table)
        if other:
            into = other.name
        else:
            into = table.name
            self.tables.append(table)
        return into

    def _create_enumerated_node(self, enumerated,
                                keyvalue_table_key_type="INTEGER"):
        name = sqlName(enumerated.getName())
        table_name = sqlName(enumerated.get_table_name())

        kvs = enumerated.getValueMapping()
        if not kvs:
            raise ValueError("enumerated %r contains no values" % name)

        max_length = max([len(v) for v in kvs.itervalues()]) + 1

        enumTable = Table(table_name)
        enumTable.appendColumn(Column('id',
                                      '%s PRIMARY KEY NOT NULL'
                                      % (keyvalue_table_key_type,)))
        enumTable.appendColumn(Column('value',
                                      'VARCHAR(%d) NOT NULL' % max_length))

        for k in sorted(kvs.iterkeys()):
            enumTable.appendInsert("(id, value) VALUES (%d, '%s')"
                                   % (k, sqlQuote(kvs[k])))

        into = self.add_choice_table(enumTable)

        table = self.tableStack[-1]

        default_value = enumerated.findDefault()
        if default_value is None:
            raise ValueError("Enumerated node %r has no default value")
        default_clause = " DEFAULT %d" % (default_value,)

        table.appendColumn(Column(name,
                                  "INTEGER" + default_clause,
                                  func=enumerated.getFunction(),
                                  anon_value=default_value))

        table.appendConstraint("FOREIGN KEY (%s) REFERENCES %s (id)"
                               % (name, into))
        table.addDependency(into)


    def createSchema(self):
        table = Table("master_tbl", root=self.root)
        table.appendColumn(Column("id", "SERIAL PRIMARY KEY NOT NULL", False))
        table.appendColumn(Column("uuid_id", "CHAR(36)", False,
                                  "suppress_value"))
        table.appendColumn(Column("bearbeiter_id", "int NOT NULL", False))

        self.maybe_add_fkz_column(table)

        table.appendConstraint("UNIQUE (uuid_id)")
        self.tableStack = [table]
        self._createRecursive(self.root)
        self.tables.append(self.tableStack.pop())

    def maybe_add_fkz_column(self, table):
        """Add column 'fkz' to table, if it wouldn't be added automatically.

        The master_tbl.fkz column is a special case because there is
        code related to the evaluation server that requires it, and
        therefore it must be present in all mpuls applications. Some
        meta trees explicitly define such a field (WASKA, WASKO), others
        do not, so it must be added in some cases.

        This method adds the fkz column only if the meta tree doesn't
        contain an fkz field that would end up in master_tbl and that is
        a text field. If the meta tree describes an fkz field for the
        master_tbl which is not text, that's likely an error, so a
        warning is printed to stderr.
        """
        for nc in self.root.walk():
            if nc.getName() == "fkz":
                # if nc is not directly in the master_tbl, i.e. if it is
                # contained in a repeat group, ignore it.
                if len(structure_path(nc)) > 1:
                    continue

                if isinstance(nc, data.TextLeaf):
                    return
                print >> sys.stderr, ("Tree contains node component named 'fkz'"
                                      " which is not a TextLeaf. Column 'fkz'"
                                      "may be duplicated.")
                break

        table.appendColumn(Column("fkz", "VARCHAR(60)", False))

    def topological(self):
        free, bound = [], []
        for t in self.tables:
            if not t.depends:
                free.append(t)
            else:
                bound.append(t)

        out = []
        while free:
            f = free.pop()
            out.append(f)
            tmp = []
            for i in range(len(bound)):
                t = bound[i]
                try:
                    t.depends.remove(f.name)
                    if not t.depends:
                        tmp.insert(0, i)
                except:
                    pass
            for i in tmp:
                free.append(bound.pop(i))

        if bound:
            print >> sys.stderr, \
                "WARNING: cyclic dependencies on %d tables." % len(bound)
            for t in bound:
                deps = ", ".join(sorted(t.depends))
                print >> sys.stderr, "  %s: %s" % (t.name, deps)
            # append them for debugging purposes
            out.extend(bound)
        return out

    def _createRecursive(self, node):
        for child in node.children:
            self._createChild(child)

    def _create_simple_column(self, widget, sql_type, column_class=Column,
                              **kw):
        table = self.tableStack[-1]
        table.appendColumn(column_class(sqlName(widget.getName()), sql_type,
                                        func=widget.getFunction(), **kw))

    def _createDate(self, date):
        self._create_simple_column(date, "DATE")

    def _createBool(self, bool):
        name = sqlName(bool.getName())
        table = self.tableStack[-1]
        kvs = {-1: 'Keine Angabe', 0: 'Nein', 1: 'Ja'}
        max_length = max([len(v) for v in kvs.itervalues()]) + 1

        choiceTable = Table(name + '_tbl')
        choiceTable.appendColumn(Column('id', 'SERIAL PRIMARY KEY NOT NULL'))
        choiceTable.appendColumn(Column('value',
                                        'VARCHAR(%d) NOT NULL' % max_length))

        for k in sorted(kvs.iterkeys()):
            choiceTable.appendInsert("(id, value) VALUES (%d, '%s')"
                                     % (k, sqlQuote(kvs[k])))

        into = self.add_choice_table(choiceTable)

        table = self.tableStack[-1]
        table.appendColumn(Column(name, 'INTEGER DEFAULT -1',
                                  func=bool.getFunction()))
        table.appendConstraint("FOREIGN KEY (%s) REFERENCES %s (id)"
                               % (name, into))
        table.addDependency(into)

    def _createText(self, text):
        length = text.getMaxLength()
        if length:
            sql_type = "VARCHAR(%d)" % length
        else:
            sql_type = "VARCHAR"
        self._create_simple_column(text, sql_type)

    def _createVirtual(self, text):
        # FIXME: datatype is used for anonymization. Return NULL::TEXT
        # here but virtual may represent all kind of datatypes
        self._create_simple_column(text, "TEXT", column_class=VirtualColumn,
                                   accessible=False)

    def _createTextArea(self, text):
        self._create_simple_column(text, "TEXT")

    def _createInteger(self, integer):
        self._create_simple_column(integer, "INTEGER")

    def _createNumeric(self, numeric):
        self._create_simple_column(numeric, "NUMERIC")

    def _createChoice(self, choice):
        if choice.getMultiple():
            raise ValueError("No support for multiple choices.")
        self._create_enumerated_node(choice, keyvalue_table_key_type="SERIAL")

    def _createRadioGroup(self, radio):
        self._create_enumerated_node(radio)

    def _createRepeatNode(self, group):
        name = sqlName(group.getName())
        parent = self.tableStack[-1]
        table = Table("%s_tbl" % name, parent, group)
        table.appendColumn(Column("id", "SERIAL PRIMARY KEY NOT NULL",
                                  False))
        table.appendColumn(Column("%s_id" % removeTbl(parent.name),
                                  "INTEGER NOT NULL", False))
        table.appendColumn(Column("uuid_id", "CHAR(36)", False,
                                  "suppress_value"))
        table.appendConstraint("FOREIGN KEY (%s_id) REFERENCES %s (id)"
                               " ON DELETE CASCADE"
                               % (removeTbl(parent.name), parent.name))
        table.appendConstraint("UNIQUE (uuid_id)")
        table.addDependency(parent.name)
        self.tableStack.append(table)

        for child in group.children:
            self._createChild(child)

        self.tables.append(self.tableStack.pop())

    def _createChild(self, child):
        if isinstance(child, Leaf):
            if isinstance(child, data.NumericLeaf):
                self._createNumeric(child)
            elif isinstance(child, data.IntLeaf):
                self._createInteger(child)
            elif isinstance(child, data.DateLeaf):
                self._createDate(child)
            elif isinstance(child, data.TextLeaf):
                self._createText(child)
            elif isinstance(child, data.TextAreaLeaf):
                self._createTextArea(child)
            elif isinstance(child, data.BoolLeaf):
                self._createBool(child)
            elif isinstance(child, data.VirtualLeaf):
                self._createVirtual(child)
        elif isinstance(child, Node):
            if isinstance(child, data.RepeatNode):
                self._createRepeatNode(child)
            elif isinstance(child, data.RadioNode):
                self._createRadioGroup(child)
            elif isinstance(child, data.ChoiceNode):
                self._createChoice(child)
            else:
                self._createRecursive(child)


class ExportAsSQL(object):

    def __init__(self, document):
        self.document = document
        self.root = document.root

    @staticmethod
    def getDescription():
        return "Export document as SQL statements"

    def printRepeats(self, tables):
        arr = []
        for t in tables:
            root = t.root
            if not root:
                continue
            parent = t.parent
            joins = []
            while parent:
                joins.append(parent.name)
                parent = parent.parent
            linear = list(root.firstLevel(lambda c:
                                          isinstance(c, (data.RepeatNode,
                                                         data.PageNode))))
            print "%s:" % t.name
            arr.append((t, joins, linear))
            if joins:
                print "\tjoins: %s" % " -> ".join(joins)
            if linear:
                print "\tlinear: %s" % ", ".join([p.getName() for p in linear])
        return arr

    def doExport(self, path, just_views=False, write_eval_total_view=False,
                 transaction=True):
        tableGrants = ""
        creator = SchemaCreator(self.root)
        creator.createSchema()
        tables = creator.topological()
        arr = self.printRepeats(tables)

        caseAnonymizer = CaseAnonymizer()

        out = codecs.open(path, "wb", "UTF-8")
        try:
            pfeil = out.write
            pfeil(u'--\n-- Schema generated by Libformed\n-- %s\n'
                  % strftime("%a, %Y-%m-%d %H:%M:%S +0000", gmtime()))
            pfeil('-- # Revision: $Revision: 8dcc83c778ca $\n--\n')
            pfeil('-- # tables: %d\n--\n' % len(tables))

            if transaction:
                pfeil('BEGIN;')

            if not just_views:
                pfeil('\n--\n-- DROP TABLES\n--\n')
                for t in reversed(tables):
                    pfeil('-- DROP TABLE %s CASCADE;\n' % t.name)

                pfeil('\n--\n-- CREATE TABLES\n--\n')
                for table in tables:
                    pfeil("%s\n" % table.getSchema())

                pfeil('\n--\n-- INSERT DATA\n--\n')
                for table in tables:
                    inserts = table.getInserts()
                    if inserts:
                        pfeil("%s\n\n" % inserts)
                        tableGrants = tableGrants \
                            + (GRANT_SELECT % (table.getName(), ':cm_group')) \
                            + (GRANT_SELECT % (table.getName(), ':admin_group'))

            pfeil('\n--\n-- GRANTS ON TABLES\n--\n')
            pfeil(tableGrants)

            # generate ro-views for master table and repeat groups
            pfeil("--\n--\n-- views for master table and repeat groups\n--\n")
            virtual_sql_export = ExportSQLForVirtualFields(self.document)
            for raw_table, joins, linear in arr:
                oldjoins = joins
                table = raw_table.name
                tab = sqlName(table)
                pfeil("CREATE OR REPLACE VIEW %s_view\nAS SELECT \n" % tab)

                str_columns = []
                for column in raw_table.columns:
                    if isinstance(column, VirtualColumn):
                        sql = virtual_sql_export.doExport(column.name)
                        if sql:
                            str_columns.append("     ( %s ) as %s"
                                               % (sql, column.name))
                        else:
                            str_columns.append("     -- missing definition for"
                                               " %s" % column.name)
                            str_columns.append("     NULL:VARCHAR as %s"
                                               % column.name)
                    else:
                        str_columns.append("     %s.%s AS %s"
                                           % (tab, column.name, column.name))
                pfeil(",\n".join(str_columns))
                pfeil("\nFROM %s\n" % tab)

                njoins = joins
                oldTable = table

                join_path = []
                while njoins:
                    found = False
                    for t in arr:
                        raw_table2, joins, linear2 = t
                        nname = raw_table2.name
                        if nname == njoins[0]:
                            found = True
                            break
                    if not found:
                        break

                    njoins = joins

                    self._write_inner_join(pfeil, oldTable, nname)
                   # pfeil("    INNER JOIN %s\n" % nname)
                   # pfeil("    ON         %s.%s_id = %s.id\n"
                   #       % (oldTable, removeTbl(nname), nname))
                    join_path.append((oldTable, nname))
                    oldTable = nname
                pfeil(VIEW_WHERE_CLAUSE)
                pfeil("\n")


                pfeil(UPDATE_RULE % (tab, tab, tab))
                pfeil(",\n".join(["    %s = NEW.%s" % (c.name, c.name)
                                  for c in raw_table.columns if c.accessible]))
                pfeil("\n")
                pfeil("WHERE id = NEW.id;\n\n")

                pfeil(FORBID_DELETE_RULE % (tab, tab))
                pfeil("\n")
                pfeil(FORBID_INSERT_RULE % (tab, tab))
                pfeil("\n")

                vn = "%s_view" % tab
                pfeil(GRANT_SELECT % (vn, ':cm_group'))
                pfeil(GRANT_UPDATE % (vn, ':cm_group'))
                pfeil(GRANT_SELECT % (vn, ':admin_group'))
                pfeil("\n")

                # anonymization view
                self._writeAnonymizationView(pfeil, raw_table, tab, join_path)

                caseAnonymizer.addAnoymizer(RelationAnonymizer(raw_table,
                                                               table, oldjoins))

            caseAnonymizer.writeFunction(pfeil)

            # Create Master Table Evaluation View
            if write_eval_total_view:
                self._writeEvalTotalView(pfeil)

            if transaction:
                pfeil('COMMIT;\n')
        finally:
            out.close()

    def _write_inner_join(self, pfeil, oldTable, nname):
        pfeil("    INNER JOIN %s\n" % nname)
        pfeil("    ON         %s.%s_id = %s.id\n"
              % (oldTable, removeTbl(nname), nname))

    def _writeEvalTotalView(self, pfeil):
        # NOTE: The master_tbl_eval_total_view defined here must be the
        # same as the one defined in db-admin's sql/install/14_evaluation.sql
        pfeil("-- Create a default (not very performant) evaluation view.\n")
        pfeil("-- This is based on master_tbl_view so that the restrictions\n")
        pfeil("-- based on the user-id implemented by master_tbl_view are\n")
        pfeil("-- also applied to %s. It suffices\n" % EVALUATION_VIEW)
        pfeil("-- to restrict the view to entries that are not marked for\n")
        pfeil("-- deletion.\n")

        pfeil("CREATE OR REPLACE VIEW %s AS\n" % EVALUATION_VIEW)
        pfeil("    SELECT m.*\n")
        pfeil("    FROM master_tbl_view m JOIN ka_status_tbl s\n")
        pfeil("    ON m.id = s.master_id\n")
        pfeil("    WHERE s.status <> 3;\n")
        pfeil("\n")
        pfeil(GRANT_SELECT % (EVALUATION_VIEW, ':cm_group'))
        pfeil(GRANT_SELECT % (EVALUATION_VIEW, ':admin_group'))
        pfeil("\n")

    def _writeAnonymizationView(self, pfeil, raw_table, table, path):
        anonym = []

        for c in raw_table.columns:
            if not c.func:
                anonym.append("%s.%s AS %s" % (table, c.name, c.name))
            else:
                if c.func == "suppress_value":
                    if c.anon_value == 'NULL':
                        datatype = c.datatype.split()[0] # a bit hackish
                        anonym.append('NULL::%s AS %s' % (datatype, c.name))
                    else:
                        anonym.append('%s AS %s' % (c.anon_value, c.name))
                else:
                    args = [x.strip() for x in c.func.split(':')]
                    if len(args) > 1:
                        anonym.append("%s(%s, %s) AS %s"
                                      % (args[0], c.name, ", ".join(args[1:]),
                                         c.name))
                    else:
                        anonym.append("%s(%s) AS %s"
                                      % (args[0], c.name, c.name))
        if table == "master_tbl":
            pfeil("CREATE OR REPLACE VIEW %s_anonym_view\nAS SELECT\n"
                  "    %s FROM %s_view master_tbl\n"
                  "    INNER JOIN ka_status_tbl s ON master_tbl.id = s.master_id\n"
                  "    WHERE  s.status <> 3;\n\n"
                  % (table, ",\n    ".join(anonym), table))
        else:
            pfeil("CREATE OR REPLACE VIEW %s_anonym_view\nAS SELECT\n"
                  "    %s FROM %s\n"
                  % (table, ",\n    ".join(anonym), table))
            for oldTable, nname in path:
                if nname == "master_tbl":
                    pfeil("    INNER JOIN master_tbl_anonym_view\n")
                    pfeil("    ON   %s.master_id = master_tbl_anonym_view.id\n"
                          % oldTable)
                else:
                    self._write_inner_join(pfeil, oldTable, nname)
            pfeil(" ;\n\n")

        vn = "%s_anonym" % table
        pfeil(FORBID_DELETE_RULE % (vn, vn))
        pfeil("\n")
        pfeil(FORBID_INSERT_RULE % (vn, vn))
        pfeil("\n")

        vn = "%s_anonym_view" % table
        pfeil(GRANT_SELECT % (vn, ':cm_group'))
        pfeil(GRANT_SELECT % (vn, ':admin_group'))
        pfeil("\n")

# vim:set ts=4 sw=4 si et sta sts=4:
