# -*- coding: utf-8 -*-
#
# Copyright (c) 2010, 2011, 2012, 2014 by Intevation GmbH
#
# FormEd is free software under the terms of the
# GNU General Public License Version 3. See file
# LICENSE coming with the source of FormEd for details.
#
# Authors:
# Bernhard Herzog     <bernhard.herzog@intevation.de>

"""A common interface to access different types of instance trees."""

import logging
import datetime

from formed.meta.data import DateLeaf
from formed.instance.backends.common import rule_available


log = logging.getLogger(__name__)


def as_name(name_or_meta):
    if isinstance(name_or_meta, basestring):
        name = name_or_meta
    else:
        name = name_or_meta.getName()
    return name


class FormData(object):

    """Base class for common interfaces for different types of instance trees.
    """

    def __init__(self, formed):
        self.formed = formed

    def push_node(self, node):
        """Push the repeat node on the stack.
        After this call names of items contained in the repeat group
        refer to this particular instance.

        The default implementation does nothing.
        """

    def pop_node(self):
        """Remove the last repeat node pushed onto stack.
        The default implementation does nothing.
        """

    def get_meta_node(self, name):
        """Return the meta node for the given name."""
        return self.formed.findByName(name)

    def get_meta_node_from_item_name(self, item_name):
        """Return the meta node for the instance item given by item_name.

        This method must be overwritten in derived classes.  The default
        implementation always raises NotImplementedError.
        """
        raise NotImplementedError

    def get_value(self, meta):
        """Return the value for the given meta item or name.
        The parameter meta may be the formed meta node for the item or
        its name as a string.

        This method must be overwritten in derived classes.  The default
        implementation always raises NotImplementedError.
        """
        raise NotImplementedError

    def get_referenced_meta_node(self, nc):
        """Return the meta node referenced by the ReferenceNode nc"""
        refname = nc.getReference()
        if refname:
            refnc = self.get_meta_node(refname)
            if not refnc:
                raise ValueError("ReferenceNode %r references an unkown Node"
                                 " %s", nc, refname)
            return refnc
        raise ValueError("ReferenceNode %r does not reference a Node", nc)

    def get_repeat_group(self, meta):
        """Return all instances of the repeat group indicated by meta.
        The parameter meta may be the formed meta node for the repeat
        group or its name as a string.  The return value is a sequence.

        This method must be overwritten in derived classes.  The default
        implementation always raises NotImplementedError.
        """
        raise NotImplementedError

    def get_ordered_repeat_group(self, meta):
        class SortDummy:

            def __init__(self, rg, values):
                self.rg = rg
                self.values = values

            def __repr__(self):
                return "SortDummy(%r, %r)" % (self.rg, self.values)

        orderby = meta.getOrderBy()
        instances = []

        for rg in self.get_repeat_group(meta):
            self.push_node(rg)
            try:
                values = dict()
                for field, direction in orderby:
                    values[field] = self.get_value(field)
                instances.append(SortDummy(rg, values))
            finally:
                self.pop_node()

        for field, direction in reversed(orderby):
            field_meta = self.get_meta_node(field)
            if isinstance(field_meta, DateLeaf):
                def sort_key(rg):
                    date = rg.values[field]
                    if date is None:
                        # put empty fields to the end
                        if direction == "asc":
                            return datetime.date.max.toordinal()
                        else:
                            return -1
                    elif isinstance(date, basestring):
                        # somebody inserted a wrong date which should be
                        # rendered now with an ordering. Therefore always render
                        # rgs with errors on the top of the rg list.
                        if direction == "desc":
                            return datetime.date.max.toordinal()
                        else:
                            return -1
                    else:
                        return date.toordinal()
            else:
                def sort_key(rg):
                    if isinstance(direction, list):
                        return direction.index(rg.values[field])
                    else:
                        return rg.values[field]
            instances.sort(key=sort_key, reverse=(direction == "desc"))
        return [rg.rg for rg in instances]

    def get_id(self, meta):
        """Return the form identifier for the item given the meta item or name.
        The parameter meta may be the formed meta node for the item or
        its name as a string.

        This method must be overwritten in derived classes.  The default
        implementation always raises NotImplementedError.
        """
        raise NotImplementedError

    def get_rg_id(self, rg):
        """Return the form identifier for the repeat group rg.
        This method must be overwritten in derived classes.  The default
        implementation always raises NotImplementedError.
        """
        raise NotImplementedError

    def has_errors(self, meta):
        """Return whether any errors are known for the item given by meta.
        The parameter meta may be the formed meta node for the item or
        its name as a string.

        The default implementation always returns False.  Derived
        classes should override this method to return a more correct
        value.
        """
        return False

    def full_item_name(self, meta):
        """Return the full name of the instance item for the meta item or name.
        The parameter meta may be the formed meta node for the item or
        its name as a string.

        This method must be overwritten in derived classes.  The default
        implementation always raises NotImplementedError.
        """
        raise NotImplementedError

    def get_variable_provider(self):
        """Return a VariableProvider instance.
        This method must be overwritten in derived classes.  The default
        implementation always raises NotImplementedError.
        """
        raise NotImplementedError

    def evaluate_conditional(self, meta):
        """Evaluate the conditional node meta and return the result.
        If the conditional does not have an expression, the method
        returns false.
        """
        expr = meta.getExpr()
        if expr is not None:
            return expr.evaluate(self.get_variable_provider())
        return False

    def evaluate_expression(self, meta):
        """Evaluate the expression node meta and return the result.
        If the expression does not have an expression, the method
        raises a ValueError
        """
        expr = meta.getExpr()
        if expr is not None:
            return expr.evaluate(self.get_variable_provider())
        return ValueError("Unable to evaluate expression %r", meta)

    def evaluate_rules(self, meta, warnings_only=False):
        """Evaluate rules for the field indicated by the meta node.
        The rules for a field are all rules that indicate that they mark
        that field.

        Normally, all rules are evaluated. However, if the warnings_only
        parameter is true (default False), only warnings are evaluated.
        Warning rules are all rules that are warnings or notices.

        The return value is a tuple with three sequences, each
        containing the rules that evaluated to false. The first sequence
        contains warning rules, the second notice rules and the last
        contains error rules.
        """

        formed = self.formed
        rules = formed.name2markingrules.get(meta.getName(), ())
        if not rules:
            return ([], [], [])

        instance_name = self.full_item_name(meta)

        warnings = []
        notices = []
        errors = []

        context = self.get_variable_provider()
        for rule in rules:
            if warnings_only and not rule.isWarning():
                continue
            if not rule_available(formed, rule, context):
                continue
            for result, variables in rule.getExpr().evaluate_instances(context):
                if not result and instance_name in variables:
                    if rule.isError():
                        collection = errors
                    elif rule.isNotice():
                        collection = notices
                    elif rule.isWarning():
                        collection = warnings
                    else:
                        log.warning("Unknown rule mark type %s. Rule %s will"
                                    "be ignored." % (rule.getMark(),
                                                     instance_name))
                        continue
                    collection.append(rule)

        return (warnings, notices, errors)

    def evaluate_warnings(self, meta):
        """Evaluate all warning rules for the field indicated by the meta node.
        The return value is a tuple with two sequences, each containing
        the rules that evaluated to false. The first sequence contains
        warning rules, the second notice rules.
        This method is a special case of evaluate_rules. See the
        doc-string of evaluate_rules for more details.
        """
        return self.evaluate_rules(meta, warnings_only=True)[:2]
