# -*- coding: utf-8 -*-
# Copyright 2009 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

from collections import defaultdict
import logging

from formed.meta.data import RelevantVisitor, RuleLeaf, ExternalChoiceListLeaf,\
    ConditionalNode, PageNode, RootNode, RepeatNode, ReferenceNode, CaseNode,\
    ExprLeaf, InfoLeaf, ChoiceNode, RadioNode, Leaf, EnumeratedNode, \
    VirtualLeaf, SaveLeaf


log = logging.getLogger(__name__)


class DocumentValidationError(Exception):

    pass


def conditional_parents(nc):
    """Return a list with all ConditionalNode ancestors of NodeComponen nc"""
    # ignore the last item of the return value of nc.pathToRoot()
    # because that is nc itself.
    return [item for item in nc.pathToRoot()[:-1]
            if isinstance(item, ConditionalNode)]


def containing_page(nc):
    """Return the page node containing nc"""
    for item in nc.pathToRoot():
        if isinstance(item, PageNode):
            return item
    return None


def containing_repeat_group(nc):
    """Return the innermost repeat group containing nc.
    If nc is not part of a repeat group, the function returns None.
    """
    for item in reversed(nc.pathToRoot()):
        if isinstance(item, RepeatNode):
            return item
    return None


class Document(object):

    """The FormEd meta tree as a whole.

    Fully initialized instances, i.e. instances whose init method has
    been called, have the following public instance variables:

    name2nc

        Maps node component names to the corresponding NodeComponent
        instance. All node components that occur in the XML file and
        have a unique name are contained here, except for the following:
        LogbookNode, CaseNode, RootNode and all children of RadioNode
        and ChoiceNode instances.

    name2page

        Maps node component names to the PageNode instance that contains
        the node component. Node components that are not contained in
        pages are not included in this mapping. Membership tests can
        therefore be used to determine whether a given node component is
        contained in a page.

    name2rules

        Maps node component names to a list of all rules that depend on
        the node component.

    name2markingrules

        Maps node component names to a list of all rules that refer to
        the node in the mark attribute (see RuleLeaf.getMarkedName).

    name2conditionals

        Maps node component names to a list of all conditionals directly
        or indirectly containing the node component. For node components
        that are themselves conditionals, the node component itself is
        not included in that list.

    name2ref

        Maps node component name to a list of reference nodes that refers to
        this node.

    page_fields

        Maps page node components to a list of node components
        representing the potentially user editable fields on that page.

        NOTE: This is somewhat experimental, and precisely which types
        of nodes are added to the list may change. In particular, how
        reference nodes are handled may change. For a reference node, it
        might be good not to use the reference node itself, but the node
        it references. At the moment, reference nodes are added to the
        list as reference nodes, so users can decide for themselves how
        to treat them.
    """

    def __init__(self, root=None):
        if root is None:
            self.root = RootNode()
        else:
            self.root = root
        self.root.document = self
        self.name2nc = None

    def init(self, validate_tree=True):
        self.buildIndices()
        if validate_tree:
            self.validate_tree()

    def buildIndices(self):
        name2nc = {}
        name2rules = defaultdict(list)
        name2markingrules = defaultdict(list)
        name2conditionals = {}
        name2page = {}
        name2ref = defaultdict(list)
        name2rgtype = {}
        page_fields = defaultdict(list)

        def callback(nc):
            if isinstance(nc, RuleLeaf):
                expr = nc.getExpr()
                for var in expr.getDependencies():
                    name2rules[var].append(nc)
                marked_name = nc.getMarkedName()
                if marked_name:
                    name2markingrules[marked_name].append(nc)

            name = nc.getName()
            page = containing_page(nc)
            if name:
                if name in name2nc:
                    log.warning("name %r used multiple times", name)
                name2nc[name] = nc
                name2conditionals[name] = conditional_parents(name2nc[name])
                if page is not None:
                    name2page[name] = page

                    # Determine type of a potential rg where the field is in.
                    rg = containing_repeat_group(page)
                    if rg is not None:
                        # page is within RG -> standalone rg
                        name2rgtype[name] = 2
                    else:
                        # page is not in RG -> inline rg or no rg at all
                        rg = containing_repeat_group(nc)
                        if rg is not None:
                            name2rgtype[name] = 1
            else:
                log.warning("ignore unnamed: %r", nc)

            if isinstance(nc, ReferenceNode):
                refname = nc.getReference()
                if refname == nc.getName():
                    log.warning("Reference %r references itself",
                                nc.getName())
                else:
                    name2ref[refname].append(nc)

            if (page is not None
                and (isinstance(nc, (Leaf, EnumeratedNode, ReferenceNode))
                     and not isinstance(nc, (ExprLeaf, InfoLeaf, VirtualLeaf,
                                             RuleLeaf, SaveLeaf)))):
                page_fields[page.getName()].append(nc)

        visitor = RelevantVisitor(callback)
        self.visit(visitor.visitor)

        self.name2nc = name2nc
        self.name2rules = name2rules
        self.name2markingrules = name2markingrules
        self.name2conditionals = name2conditionals
        self.name2page = name2page
        self.name2ref = name2ref
        self.name2rgtype = name2rgtype
        self.page_fields = page_fields

    def validate_tree(self):
        """Check the tree for inconsistencies and print warnings.
        Currently this checks whether all dependencies of expressions in
        rules or conditionals are actually defined in the tree.
        """
        for item in self.findAllByClass(RuleLeaf):
            self.validate_expr_dependencies(item)
        for item in self.findAllByClass(ConditionalNode):
            self.validate_expr_dependencies(item)
        for item in self.findAllByClass(VirtualLeaf):
            self.validate_expr_dependencies(item)
        self.validate_info_leafs()
        self.validate_external_choice_leafs()
        self.validate_choice_options()
        self.validate_enumeration_defaults()
        self.validate_repeat_group_digests()
        self.validate_tags()
        self.validate_references()

    def validate_info_leafs(self):
        """Check if any info leaf has a value attribute and also contains child
        elements. Afterwards validate its expression children.
        """
        for node in self.findAllByClass(InfoLeaf):
            if node.getValue() and node.getChildren():
                log.error("Info %r has value and child elements to specify"
                          " the text content" % node)
            for item in node.getChildren():
                if isinstance(item, ExprLeaf):
                    self.validate_expr_dependencies(item)

    def validate_expr_dependencies(self, node):
        """Check whether all dependencies of the expression in node exist.
        The node must be have a getExpr method that returns an Expr object.
        For each dependency that doesn't exist in the formed tree, a
        message is printed to stderr.
        """
        for dependency in node.getExpr().getDependencies():
            if dependency not in self.name2nc:
                if isinstance(node, RuleLeaf):
                    kind = "rule"
                elif isinstance(node, ExprLeaf):
                    kind = "expr"
                elif isinstance(node, VirtualLeaf):
                    kind = "virtual"
                else:
                    kind = "conditional"
                log.warning("%s %r has unknown dependency %r",
                            kind, node.getName(), dependency)

    def validate_external_choice_leafs(self):
        """Force loading of external choice data.
        If the file referenced by an ExternalChoiceListLeaf instance
        cannot be loaded an exception will be raised.
        """
        for external in self.findAllByClass(ExternalChoiceListLeaf):
            # calling getChildren forces loading of the external choice
            # data.
            external.getChildren()

    def validate_choice_options(self):
        """Check that the options of choice nodes have unique values."""
        # FIXME: we should also check radio nodes in the same way
        for choice in self.findAllByClass(ChoiceNode):
            # We use a set instead of a list to gather the descriptions
            # of the options with the same value because we actually do
            # have cases where one options is actually listed twice with
            # the same value and descriptoion intentionally. For
            # instance, 'Deutschland' in a list of countries so that
            # this very commonly used value is listed once at the top
            # and once in the alphabetically ordered list of countries.
            found = defaultdict(set)
            for value, description in choice.get_options():
                found[value].add(description)

            for value, descriptions in found.iteritems():
                if len(descriptions) > 1:
                    log.warning("%r defines option value %r with multiple"
                                " descriptions: %s",
                                choice, value,
                                ", ".join(repr(desc) for desc in descriptions))

    def validate_enumeration_defaults(self):
        """Check that all enumerated nodes define a default value."""
        # Note: another property that would be worthwhile to check here
        # is that the default value must be one of the choices. We don't
        # actually need to do that, though, because it must be one by
        # construction. The default value is determined by adding a
        # checked attribute to the option that is to be the default
        # value, after all
        nodes_missing_defaults = [node
                                  for node in self.findAllByClass((ChoiceNode,
                                                                   RadioNode))
                                  if node.findDefault() is None]
        if nodes_missing_defaults:
            for node in nodes_missing_defaults:
                log.error("%r has no default value", node)
            raise DocumentValidationError(
                "Enumerated nodes without default values: %s"
                % (", ".join(node.getName()
                             for node in nodes_missing_defaults),))


    def validate_repeat_group_digests(self):
        """Check repeat group digest specifications.
        The fields referred to in digest specifications must exist and
        refer to fields directly contained in the repeat group.
        """
        for rg in self.findAllByClass(RepeatNode):
            for field_name in rg.getDigest():
                self.validate_digest_field(rg, field_name, "display field")
            orderby = rg.getOrderBy()
            for field_name, direction in orderby:
                self.validate_digest_field(rg, field_name, "orderby field")
            if len(orderby) > 1 and rg.getAttribute("reverse"):
                log.warning("Repeat group %r has multiple order fields and"
                            " a 'reverse' attribute", rg.getName())

    def validate_digest_field(self, rg, field_name, field_type):
        """Checks whether field_name refers to an existing field in rg.
        The field_type parameter is used in warning messages that are
        generated if the checks fail and can be used to distinguish
        between e.g. fields that are shown in the digest and fields that
        are used for ordering to make it easier to relate the warnings
        to the XML file.
        """
        if field_name:
            field_nc = self.name2nc.get(field_name)
            if field_nc is None:
                log.warning("Repeat group %r digest refers to unknown %s %r",
                            rg.getName(), field_type, field_name)
            else:
                if containing_repeat_group(field_nc) is not rg:
                    log.warning("Repeat group %r digest refers to %s %r which"
                                " is not contained in the repeat group",
                                rg.getName(), field_type, field_name)

    def validate_tags(self):
        seen_tags = defaultdict(list)
        for rule in self.findAllByClass(RuleLeaf):
            tag = rule.getTag()
            if tag and tag != "none":
                seen_tags[tag].append(rule)

        for tag, rules in seen_tags.iteritems():
            if len(rules) > 1:
                log.warning("Tag %r used multiple times: %s", tag,
                            ", ".join(rule.getName() for rule in rules))

    def validate_references(self):
        for reference in self.findAllByClass(ReferenceNode):
            refname = reference.getReference()
            # check if reference links to itself
            if refname == reference.getName():
                log.warning("Reference %r references itself",
                            reference.getName())
                return
            if refname not in self.name2nc:
                log.warning("Reference %r does not reference a valid Node with"
                            " name %r", reference.getName(), refname)
                return
            refnc = self.name2nc[refname]
            if isinstance(refnc, ReferenceNode):
                log.warning("Reference %r references another Referecence %r",
                            reference.getName(), refnc.getName())
            elif isinstance(refnc, RuleLeaf):
                log.warning("Reference %r references Rule %r",
                            reference.getName(), refnc.getName())
            elif isinstance(refnc, CaseNode):
                log.warning("Reference %r references Case %r",
                            reference.getName(), refnc.getName())
            # check if referenced element is child of a PageNode
            # and implicit check if referenced group contains a page
            if refname not in self.name2page:
                log.warning("Reference %r does not reference a child of a page",
                            reference.getName())
            # check if referenced element is parent of reference
            parents = reference.pathToRoot()
            if refnc in parents:
                log.warning("Reference %r references a parent Node %r",
                            reference.getName(), refnc.getName())
            # check if referenced element has a parent repeat group and
            # reference is also a child of this repeat group
            current = refnc.parent
            repeat = None
            while current is not None and repeat is None:
                if isinstance(current, RepeatNode):
                    repeat = current
                current = current.parent
            if repeat:
                current = reference.parent
                while current is not None and current is not repeat:
                    current = current.parent
                if not current:
                    log.warning("Reference %r references child Node %r of "
                                "RepeatNode %r but the Reference is "
                                "outside of this RepeatNode",
                                reference.getName(), refnc.getName(),
                                repeat.getName())

    def getVersion(self):
        return self.root.getVersion()

    def getRoot(self):
        return self.root

    def findByName(self, name, linear_lookup=False):
        if self.name2nc:
            try:
                return self.name2nc[name]
            except KeyError:
                if not linear_lookup:
                    return None

        if self.root:
            for nc in self.root.walk():
                if name == nc.getName():
                    return nc
        return None

    def findByClassAndName(self, clazz, name=None):
        if self.root:
            for nc in self.root.walk():
                if isinstance(nc, clazz) and (not name or nc.getName() == name):
                    return nc
        return None

    def findAllByClass(self, clazz):
        if self.root:
            for nc in self.root.walk():
                if isinstance(nc, clazz):
                    yield nc

    def visit(self, visitor):
        if self.root:
            self.root.visit(visitor)

    def walk(self):
        if self.root:
            for nc in self.root.walk():
                yield nc

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf-8 :
