from cgi import escape as cgi_escape


def escape(s, quote=None):
    if s is None:
        return ""
    return cgi_escape(s, quote)

def NA(text):
    if not text:
        return '-/-'
    return text

def nl_to_br(text):
    """Replace newlines in text with <br> tags"""
    text = text.replace('\r\n', '<br>')
    text = text.replace('\n', '<br>')
    text = text.replace('\r', '<br>')
    return text


def tag(tagname, _close=False,
        _booleans=('checked', 'disabled', 'multiple', 'selected'),
        _ignore_empty=('placeholder'),
        **attrs):
    """Create an opening HTML tag with properly escaped attributes.

    The tagname parameter gives the name of the tag.  The attributes and
    their values are given as keyword parameters.  Most attributes have
    string values.  Boolean attributes are treated as described below.
    The values of string attributes are converted to strings if they are
    not aleady strings (byte-strings or unicode) using these rules: None
    is converted to the empty string, other values are converted using
    the str function.  After conversion the values of string attributes
    will be properly escaped as attribute values when inserted into the
    tag.

    To ignore empty ("") attributes, these can be added to the
    _ignore_empty list (which contains 'placeholder' as default only)
    which will cause them to be excluded from the output if empty.

    Boolean flags are treated differently: if the value of such an
    attribute is true, the attribute is included in the tag without a
    value, if the value is false, the attribute is omitted.  The boolean
    attributes handled in this way are given by the _booleans parameter,
    which is a sequence of attribute names.  Its default value is
    ('checked', 'disabled', 'multiple', 'selected')

    To allow for attribute names that are also python keywords, a single
    trailing underscore character is removed from an attribute name if
    it is preset, so that e.g. the class attribute can be specified
    using the class_ keyword parameters.

    A few attribute names are not valid Python identifiers because they
    contain a minus-sign (e.g. accept-charset).  However, since there
    are no attribute names that contain underscores, underscores in a
    keyword argument will be converted to minus-signs to determine the
    attribute name.  For instance the accept-charset attribute can be
    given as accept_charset to this function.

    Examples:
    >>> from formed.renderer.helpers import tag
    >>> tag("input", type="text", name="title", disabled=False)
    '<input type="text" name="title">'
    >>> 
    """
    out = ["<%s" % tagname]
    for name, value in attrs.items():
        if name.endswith("_"):
            name = name[:-1]
        name = name.replace("_", "-")
        if name in _booleans:
            if value:
                out.append(' %s="%s"' % (name, name))
        elif name in _ignore_empty and value == "":
            # Ignore certain empty-string attributes.
            pass
        else:
            if value is None:
                value = ""
            elif not isinstance(value, basestring):
                value = str(value)
            out.append(' %s="%s"' % (name, escape(value, True)))
    # ?:
    if _close:
        out.append(">")
    else:
        out.append(">")
    return "".join(out)

def folding_helper_functions():
    """Render the javascript folding helper functions:
    - toggle_folding(id): Toggle folding of element identified by id
    - init_folding(id, display): Init the folding of element identified by
        id with display status.
    The init_folding is needed to support also browsers with javascript
    disabled: In these cases all folding sections has to be unfolded.
    """

    out = []
    out.append("""
        <script type="text/javascript">
        <!--
            function toggle_folding(id) {
               var e = document.getElementById(id);
               var v = document.getElementById("folded."+id);
               var l = document.getElementById("label_"+id);
               if(e.style.display == 'block') {
                  e.style.display = 'none';
                  v.value= 'none';
                  if (l != null) {
                      l.className = l.className.replace( /(?:^|\s)unfolded(?!\S)/g , 'folded' );
                  }
               }
               else {
                  e.style.display = 'block';
                  v.value= 'block';
                  if (l != null) {
                      l.className = l.className.replace( /(?:^|\s)folded(?!\S)/g , 'unfolded' );
                  }
               }
            }
        //-->
        </script>
        """)
    out.append("""
        <script type="text/javascript">
        <!--
            function init_folding(id) {
               var e = document.getElementById(id);
               var v = document.getElementById("folded."+id);
               var l = document.getElementById("label_"+id);
               if (v == null) {
                    // If no hidden variable is found create one
                    // and set to default: none, i.e. folded
                    var v = document.createElement('INPUT');
                    v.id = 'folded.'+id;
                    v.name = 'folded.'+id;
                    v.type = 'hidden';
                    v.value = 'none';
                    var f = document.getElementsByTagName('FORM');
                    f[0].appendChild(v);
               }
               if(v.value == 'block') {
                  e.style.display = 'block';
                  if (l != null) {
                      l.className = l.className.replace( /(?:^|\s)folded(?!\S)/g , 'unfolded' );
                  }
               }
               else {
                  e.style.display = 'none';
                  v.value= 'none';
               }
            }
        //-->
        </script>
        """)
    return "".join(out)

def folding_init(list):
    """Initialise folding for a list of ID, DISPLAY tuples:
    - ID: Id of the element to be initialised
    - DISPLAY: css display state (usually 'block' or 'none')
    """

    out = []
    if len(list):
        out.append("""
          <script type="text/javascript">
          <!-- // The initialisation is needed to display the section in
               // non-javascript environments: Sections are visible by
               // default and are hidden if needed
            window.onload = function() {\n""")

        for ID in list:
            out.append("""
                init_folding("%s");
              """ % ID)
        out.append("""
            }
          //-->
          </script>\n""")

    return "".join(out)

