"""Basic HTML rendering support"""

import os
import re
import logging
import gettext
import datetime
from collections import defaultdict

from formed.meta import UNKNOWN_INT, UNKNOWN_NUMERIC, UNKNOWN_DATE
from formed.meta.nodecomponents import NodeComponent
from formed.meta.data import PageNode, GroupNode, RepeatNode, ConditionalNode, \
    RadioNode, ChoiceNode, ExternalChoiceListLeaf, MatrixNode, \
    TextLeaf, TextAreaLeaf, DateLeaf, NumericLeaf, IntLeaf, BoolLeaf, \
    InfoLeaf, SaveLeaf, RuleLeaf, ReferenceNode, ExprLeaf, VirtualLeaf, FileLeaf
from formed.meta.document import conditional_parents
from formed.instance.simplified import StructureVariableProvider, \
    StructureFormData
from formed.instance.formdata import FormData, as_name
from formed.instance.backends.common import rule_available, RuleEvalContext
from formed.renderer.helpers import escape, NA, nl_to_br, tag, \
                folding_helper_functions, folding_init


log = logging.getLogger(__name__)


SET_MODIFICATION_ATTRS = {"onchange": "setModification();"}

i18n_base_dir = None


def set_i18n_basedir(basedir):
    """Set the base directory for libformed's translation files.
    This function is only useful in special circumstances, such as when
    libformed is packaged via py2exe.
    """
    global i18n_base_dir
    i18n_base_dir = basedir


def load_catalog(languages):
    """Return message catalog for the languages given by the list languages"""
    global i18n_base_dir
    if not i18n_base_dir:
        i18n_path = os.path.join(os.path.dirname(__file__), "..", "i18n")
        if os.path.isdir(i18n_path):
            i18n_base_dir = i18n_path
    return gettext.translation("libformed", i18n_base_dir, languages,
                               fallback=True)


def hidden_bool_tag(id):
    """Return the hidden HTML input tag for the plain boolean field given by id.

    This hidden input field will have a name derived from id by
    prepending the id with 'bool-'.  The hidden field can be used to
    determine which booleans are present to detect which have been
    unckecked by the user.
    """
    return tag("input", type="hidden", name="bool-" + id)


def apply_hidden_booleans(form_result, value=""):
    """Set the boolean values in form results using the hidden fields.

    In HTML-Forms, check boxes that are not checked when the user
    submits the form are not included in the form results.  They need to
    be included the form results for mpuls, though, so that their values
    are updated properly.  This function uses the hidden fields created
    by hidden_bool_tag to determine which booleans are present in the
    form and adds all booleans which have a hidden field but are not
    included in form_result to form_result with the value given by the
    optional value parameter (default: empty string).  The hidden fields
    are removed from form_result.

    The form_result parameter should be a dictionary mapping HTML form
    field names to values.
    """
    for key in [key for key in form_result.keys() if key.startswith("bool-")]:
        del form_result[key]
        fieldid = key[5:]
        if fieldid not in form_result:
            form_result[fieldid] = value

def extract_hidden_folding(form_result):
    """Extract (hidden) input variables covering folding

    The folding feature uses hidden input fields to store and hand over
    the state of folding groups. These fields are not part of the form
    definition and have to be stripped from the form_result before these are
    stored to the instance tree.

    The function return a list of input fields (key,value)-tuples
    starting with 'folded.', which in turn are deleted from the form_result.
    """

    hidden_folding = []
    for key in [key for key in form_result.keys() if key.startswith("folded.")]:
        hidden_folding.append((key, form_result[key]))
        del form_result[key]
    return hidden_folding

def is_required(flags):
    return flags.find("required:") >= 0


def is_evaluation_relevant(flags):
    return flags.find("evaluation:") >= 0


class Text:

    def __init__(self, txt):
        self.txt = txt

    def render(self, target, weight=None):
        return self.txt


class Div(Text):

    def render(self, target, width=""):
        return '<td width="%s">%s</td>\n' % ("%s%%" % width, self.txt)


class Item(Text):

    def render(self, target, width="100"):
        if width:
            return '<td width="%s">%s</td>\n' % ("%s%%" % width, self.txt)
        else:
            return '<td>%s</td>\n' % (self.txt)


class Container:

    def __init__(self, target):
        self.target = target
        self.items = []

    def append(self, item, weight=1.0):
        self.items.append((item, weight))

    def render(self):
        out = []
        if self.items:
            out.append('<table>')
            out.append('<tr>')
            for col, w in self.items:
                data = col.render(self.target, w or "")
                if not data:
                    data = "<td></td>"
                out.append(data)
            out.append('</tr>')
            out.append('</table>')
        return "".join(out)


class ChoiceContainer(Container):

    def render(self):
        return "".join([col.render(self.target, w or "")
                        for col, w in self.items])


def repeat_id(full_id):
    """Helper function for TreeFormData"""
    return full_id.split(":", 1)[-1]


class TreeFormData(FormData):

    """FormData class for the full instance tree"""

    def __init__(self, instance, shadow=None, errors=None, node=None):
        """Initialize the form data object.

        The parameter instance is the instance tree.  The parameter
        shadow is a dictionary mapping full instance item names to
        values which will be consulted before the instance tree when
        data is fetched, so that the values in the shadow dict shadow
        the ones from the instance tree.  The parameter errors is a
        dictionary mapping item ids to error messages associated with
        the item.

        The parameter node is used as the initial node on the node
        stack. It should be either a node of the instance tree, or a
        meta node which is then used to find the corresponding instance
        node, or None in which case the root node will be used. The
        default is None.
        """
        super(TreeFormData, self).__init__(instance.getFormEd())
        self.instance = instance
        self.shadow = shadow if shadow is not None else dict()
        self.errors = errors if errors is not None else dict()
        if node is None:
            node = instance.getRootNode()
        elif isinstance(node, NodeComponent):
            case_id = instance.getRootNode().getIdentifier()
            node = instance.getItem(as_name(node) + ":" + unicode(case_id))
        self.node_stack = [node]

    def push_node(self, node):
        self.node_stack.append(node)

    def pop_node(self):
        self.node_stack.pop()

    def get_value(self, meta):
        full_name = self.get_id(meta)
        if full_name in self.shadow:
            return self.shadow[full_name]
        return self.instance.getItem(full_name).getValue()

    def get_repeat_group(self, meta):
        return self.instance.getItem(self.get_id(meta)).getChildren()

    def get_id(self, meta):
        return (as_name(meta) + ":"
                + repeat_id(unicode(self.node_stack[-1].getIdentifier())))

    def get_rg_id(self, rg):
        return rg.getIdentifier()

    def has_errors(self, meta):
        return self.get_id(meta) in self.errors

    def get_errors(self, meta):
        message = self.errors.get(self.get_id(meta))
        if message:
            return (message,)
        return ()

    def get_meta_node_from_item_name(self, item_name):
        return self.get_meta_node(item_name.split(":")[0])

    def full_item_name(self, name):
        return self.get_id(name)

    def form_name_from_full_name(self, full_name):
        return full_name

    def get_variable_provider(self):
        reference = self.node_stack[-1].getIdentifier()
        if isinstance(reference, int):
            # in most cases, the return value of getIdentifier is a
            # string. The only case where it's an int, is if the node is
            # the root node in which case getIdentifier returns the case
            # id as an int. When used as a reference item string, it
            # needs to be prefixed with a colon, so that the code in
            # formed/instance/backends/common.py will parse it
            # correctly.
            reference = ":" + unicode(reference)
        return RuleEvalContext(self.instance, reference)


class UrlFactory(object):

    def url(self, controller, action, id=None):
        """Return the URL for the given controller, action and id.
        This method must be overwritten in derived classes.  The default
        implementation always raises NotImplementedError.
        """
        raise NotImplementedError

    def icon(self, name):
        """Return the URL for the icon indicated by name.
        This method must be overwritten in derived classes.  The default
        implementation always raises NotImplementedError.
        """
        raise NotImplementedError


class GermanFormatter(object):

    """Format values using the conventions used in Germany.

    Ideally, the formatting should be based on the current locale, or be
    set depending on the preferences of the user (which is not the same
    in a web-application), but the german conventions are used in many
    places in libformed, especially the user input parsers in
    formed.meta.semantic, and the applications using it, and therefore
    this is the default formatter used in the renderer.
    """

    def format_numeric(self, meta, value):
        """Format the decimal object value with comma as decimal separator."""
        return (u"%.2f" % value).replace(".", ",")

    def format_date(self, meta, date):
        """Format the date object value using the format 'DD.MM.YYYY'."""
        return u"%02d.%02d.%d" % (date.day, date.month, date.year)


class RendererContext(object):

    def __init__(self, node, container_class):
        self.node = node
        containers = [container_class(c) for c in node.getParsedContainers()]
        containers.append(container_class(None))
        self.containers = containers
        self.target_dict = dict((t.target, t) for t in containers)

    def add(self, target, content):
        weight = None
        if target:
            m = re.match(r"(\w+)(?:|:(.+))$", target)
            if m:
                target = m.group(1)
                try:
                    weight = m.group(2)
                except (TypeError, ValueError):
                    pass

        self.find_container(target).append(content, weight)

    def find_container(self, target):
        container = self.target_dict.get(target)
        if container is None:
            container = self.containers[-1]
        return container

    def render(self):
        return "".join(c.render() for c in self.containers)


class ConditionalContext(object):

    """Renderer context for conditionals.

    Conditionals are not treated like the normal containers and the
    children only refer to targets in parents of the conditional. To
    support that, this context can be used while rendering a conditional
    and it will delegate the adds to a parent context and simply render
    to None to indicate that nothing has been renderer.
    """

    def __init__(self, parent_context):
        self.context = parent_context

    def add(self, *args):
        self.context.add(*args)

    def render(self):
        return None


class BaseRenderer(object):

    """Base class for renderers that use a FormData object and a UrlFactory.
    """

    def __init__(self, data, view_root, url_factory, read_only=True,
                 languages=None):
        self.data = data
        self.view_root = view_root
        self.url_factory = url_factory
        self.formatter = GermanFormatter()
        self.read_only = read_only
        self.translation = load_catalog(languages)

    def url(self, controller, action, id=None):
        return self.url_factory.url(controller, action, id=id)

    def icon(self, name):
        return self.url_factory.icon(name)

    def gettext(self, text):
        return self.translation.ugettext(text)

    def ungettext(self, singular, plural, n):
        return self.translation.ungettext(singular, plural, n)

    def format_numeric(self, nc, value):
        """Return the numeric value as a formatted string.
        If value is None, the method returns an empty string. If the
        value is already a string (should only happen if the user tried
        to input values that cannot be parsed as a float) the method
        returns that string. If the value is UNKNOWN_NUMERIC the method
        returns the localized "unknown" string. Otherwise the value
        should be a decimal object and the method formats the value with
        the format_numeric method of the renderer's formatter instance.
        """
        if value is None:
            formatted = ""
        elif isinstance(value, basestring):
            formatted = value
        elif value == UNKNOWN_NUMERIC:
            formatted = self.gettext("unknown")
        else:
            formatted = self.formatter.format_numeric(nc, value)
        return formatted

    def format_date(self, nc, value):
        """Return the date value as a formatted string.
        If value is None, the method returns an empty string. If the
        value is already a string (should only happen if the user tried
        to input values that cannot be parsed as a float) the method
        returns that string. If the value is UNKNOWN_DATE the method
        returns the localized "unknown" string. Otherwise the value
        should be a date object and the method formats the date with the
        format_date method of the renderer's formatter instance.
        """
        if value is None:
            formatted = ""
        elif isinstance(value, basestring):
            formatted = value
        elif value == UNKNOWN_DATE:
            formatted = self.gettext("unknown")
        else:
            formatted = self.formatter.format_date(nc, value)
        return formatted

    def render_annotation_link(self, nc, kind):
        return [tag("a", target="_blank",
                    href=self.url(controller="annotations",
                                  action=kind, id=nc.getName()),
                    class_=kind),
                tag("img", src=self.icon("formular/%s.png" % kind),
                           width=12, height=12, alt=kind),
                '</a>']

    def render_annotation_links(self, nc):
        out=[]
        if nc.hasHelp():
            out.extend(self.render_annotation_link(nc, "help"))
        if is_required(nc.getFlags()):
            out.extend(self.render_annotation_link(nc, "required"))
        if is_evaluation_relevant(nc.getFlags()):
            out.extend(self.render_annotation_link(nc, "evaluation"))
        return out

    def render_page_headings(self, all=False):
        """Render the HTML page headings for the form.
        Normally, the description of the renderer's view_root is used as
        the heading together with links for any annotations available
        for the view_root.  However, if the parameter all is given and
        true (default: False), the descriptions of the parents of the
        view_root are also rendered as headings, ordered from the
        formed-root to the view_root.
        """
        if all:
            # omit first two items of pathToRoot because those are the
            # root node and the case node.
            nodes = self.view_root.pathToRoot()[2:]
        else:
            nodes = [self.view_root]

        out = []
        for depth, node in enumerate(nodes):
            heading = "h%d" % (depth + 1,)
            out.append("<%s>%s %s</%s>"
                       % (heading, escape(node.getDescription()),
                          "".join(self.render_annotation_links(node)),
                          heading))
        return "\n".join(out)


class HtmlRenderer(BaseRenderer):

    def __init__(self, data, view_root, url_factory, read_only=True,
                 languages=None):
        super(HtmlRenderer, self).__init__(data, view_root, url_factory,
                                           read_only=read_only,
                                           languages=languages)
        self.tabindex = 3
        self.warnings = defaultdict(list)
        self.notices = defaultdict(list)
        self.errors = defaultdict(list)
        self.widget_order = []
        self.save_next = False

        self.init_folding = []

        # currently the root of the view must be a page
        assert isinstance(self.view_root, PageNode)

    def next_tabindex(self):
        tabindex = self.tabindex
        self.tabindex += 1
        return str(tabindex)

    def render(self, show_next=True, hidden_fields=()):
        out = []
        self.show_next = show_next

        if not self.read_only:
            out.append(self.render_form_tag())
            out.extend(self.render_submit_buttons_top(show_next))

        for name, value in hidden_fields:
            out.append(tag("input", type="hidden", name=name, id=name, value=value))
        out.extend(self.render_page(RendererContext(self.view_root, Container),
                                    self.view_root))

        if not self.read_only:
            out.extend(self.render_submit_buttons_bottom(show_next))
            out.append("\n</form>")

        if len(self.init_folding):
            out.append(folding_init(self.init_folding))

        return "".join(out)

    def render_form_tag(self):
        return tag("form", id="formularpage",
                   action=self.url(controller="formularpage",
                                   action="save"),
                   enctype="multipart/form-data",
                   accept_charset="UTF-8",
                   method="POST")

    def render_submit_buttons_top(self, show_next=True):
        return self.render_submit_buttons(show_next, position="Top")

    def render_submit_buttons_bottom(self, show_next=True):
        return self.render_submit_buttons(show_next, position="Bottom")

    def render_submit_buttons(self, show_next, position='', from_tag=False):
        out = []
        if not from_tag:
            out.append(tag("div", class_="widget formbuttons"))
        out.append(tag("input", type="submit", name=("save%s" % position),
                       class_="button", value=self.gettext('Save'),
                       tabindex=self.next_tabindex(),
                       id="formSave%s" % position))
        if show_next:
            out.append("&nbsp;")
            out.append(tag("input", type="submit", name="savenext",
                           class_="button",
                           value=self.gettext('Save and proceed'),
                           tabindex=self.next_tabindex(),
                           id="formSaveNext%s" % position))
        if not from_tag:
            out.append('</div>')
        return out

    def render_page(self, context, nc):
        parts = []
        parts.append(self.render_children(RendererContext(nc, Container), nc))
        return parts

    def render_children(self, context, nc, child_renderer=None):
        if child_renderer is None:
            child_renderer = self.render_nc

        for child in nc.getChildren():
            if isinstance(child, (RuleLeaf, VirtualLeaf)):
                continue
            content = child_renderer(context, child)
            if content is not None:
                context.add(child.getTarget(), content)
        return context.render()

    def render_nc(self, context, nc):
        self.check_rules(nc)
        if isinstance(nc, TextLeaf):
            return self.render_text(context, nc)
        if isinstance(nc, TextAreaLeaf):
            return self.render_text_area(context, nc)
        if isinstance(nc, DateLeaf):
            return self.render_date(context, nc)
        if isinstance(nc, NumericLeaf):
            return self.render_numeric(context, nc)
        if isinstance(nc, IntLeaf):
            return self.render_int(context, nc)
        if isinstance(nc, BoolLeaf):
            return self.render_bool(context, nc)
        if isinstance(nc, InfoLeaf):
            return self.render_info(context, nc)
        if isinstance(nc, SaveLeaf):
            return self.render_save(context, nc)
        if isinstance(nc, ChoiceNode):
            return self.render_choice(context, nc)
        if isinstance(nc, RadioNode):
            return self.render_radio(context, nc)
        if isinstance(nc, PageNode):
            return self.render_page(context, nc)
        if isinstance(nc, MatrixNode):
            return self.render_radio_matrix(context, nc)
        if isinstance(nc, RepeatNode):
            return self.render_inline_repeat_group(context, nc)
        if isinstance(nc, GroupNode):
            return self.render_group(context, nc)
        if isinstance(nc, ConditionalNode):
            return self.render_conditional(context, nc)
        if isinstance(nc, ReferenceNode):
            return self.render_reference(context, nc)
        if isinstance(nc, FileLeaf):
            return self.render_file(context, nc)
        else:
            raise ValueError("Cannot render nc %r" % nc)

    def render_text(self, context, nc):
        value = self.data.get_value(nc)
        id = self.data.get_id(nc)

        out = []
        out.extend(self.render_label(context, nc, id))
        out.extend(self.render_annotation_links(nc))
        out.append('<br>')

        out.append(tag("input", name=id, id=id, value=value,
                       class_=self.error_css_class(context, nc),
                       type="text", size=nc.getSize(),
                       placeholder=nc.getPlaceholder(),
                       maxlength=nc.getMaxLength(), disabled=self.read_only,
                       **SET_MODIFICATION_ATTRS))
        out.append('<br>')
        return Item("".join(out))

    def render_text_area(self, context, nc):
        value = self.data.get_value(nc)
        id = self.data.get_id(nc)
        cols = nc.getCols() or "0"
        rows = nc.getRows() or "0"

        out = []
        out.extend(self.render_label(context, nc, id))
        out.extend(self.render_annotation_links(nc))
        out.append('<br>')

        if not self.read_only:
            out.append(tag("textarea", id=id, name=id, rows=rows, cols=cols,
                           placeholder=nc.getPlaceholder(),
                           class_=self.error_css_class(context, nc),
                           **SET_MODIFICATION_ATTRS))
            out.append(escape(value))
            out.append("</textarea>")
            out.append("<br>\n")
        else:
            out.append('<p class="readonly">%s</p>'
                       % (NA(nl_to_br(escape(value)))))
        return Item("".join(out))

    def render_choice(self, context, nc):
        value = self.data.get_value(nc)
        id = self.data.get_id(nc)

        out = []
        out.extend(self.render_label(context, nc, id))
        out.extend(self.render_annotation_links(nc))
        out.append('<br>')

        out.append(tag("select", id=id, name=id, size=nc.getSize() or "1",
                       multiple=bool(nc.getMultiple()), disabled=self.read_only,
                       class_=self.error_css_class(context, nc),
                       **SET_MODIFICATION_ATTRS))
        out.append(self.render_choice_children(context, nc, value))
        out.append('</select>')

        out.append('<br>')
        return Item("".join(out))

    def render_choice_children(self, context, nc, value):
        def child_renderer(context, nc):
            return self.render_choice_nc(context, nc, value)

        return self.render_children(RendererContext(nc, ChoiceContainer),
                                    nc, child_renderer)

    def render_choice_nc(self, context, nc, value):
        if isinstance(nc, BoolLeaf):
            return self.render_choice_bool(context, nc, value)
        if isinstance(nc, GroupNode):
            return self.render_choice_group(context, nc, value)
        if isinstance(nc, ExternalChoiceListLeaf):
            return self.render_external_choice_list(context, nc, value)
        if isinstance(nc, ConditionalNode):
            return self.render_choice_conditional(context, nc, value)
        raise ValueError("Cannot render choice nc %r" % nc)

    def render_choice_group(self, context, nc, value):
        out = []
        if not self.read_only:
            out.append(tag("optgroup", label=nc.getDescription()))
            out.append("\n")
        out.append(self.render_choice_children(context, nc, value))
        if not self.read_only:
            out.append("</optgroup>\n")
        return Text("".join(out))

    def render_choice_bool(self, context, nc, value):
        selected = value == nc.getValue()
        out = []
        if not self.read_only and not nc.isInvisible() or selected:
            out.append(tag("option", selected=selected, value=nc.getValue()))
            out.append(escape(nc.getDescription()))
            out.append("</option>")

        return Text("".join(out))

    def render_external_choice_list(self, context, nc, value):
        # the children are sorted by target
        out = []
        for child in sorted(nc.getChildren(), reverse=True,
                            key=lambda child: child.getTarget()):
            child_value = child.getValue()
            out.append(tag("option", value=child_value,
                           selected=(child_value == value)))
            out.append(escape(child.getDescription()))
            out.append("</option>")
        return Text("\n".join(out))

    def render_choice_conditional(self, context, nc, value):
        return self.do_render_conditional(context, nc,
                                          self.render_choice_children, value)

    def render_date(self, context, nc):
        id = self.data.get_id(nc)
        formatted = self.format_date(nc, self.data.get_value(nc))
        size = "10"
        maxlength = "10"

        out = []
        out.extend(self.render_label(context, nc, id))
        out.extend(self.render_annotation_links(nc))
        out.append('<br>')

        css_class = " ".join(["datefield", self.error_css_class(context, nc)])

        out.append(tag("input", name=id, id=id, value=formatted,
                       type="text", size=size, maxlength=maxlength,
                       disabled=self.read_only, class_=css_class,
                       placeholder=nc.getPlaceholder(),
                       **SET_MODIFICATION_ATTRS))
        out.append('<br>')
        return Item("".join(out))

    def render_numeric(self, context, nc):
        id = self.data.get_id(nc)
        formatted = self.format_numeric(nc, self.data.get_value(nc))

        out = []
        out.extend(self.render_label(context, nc, id))
        out.extend(self.render_annotation_links(nc))
        out.append('<br>')

        out.append(tag("input", name=id, id=id, value=formatted, type="text",
                       placeholder=nc.getPlaceholder(),
                       class_=self.error_css_class(context, nc),
                       disabled=self.read_only, **SET_MODIFICATION_ATTRS))

        out.append('<br>')
        return Item("".join(out))

    def render_int(self, context, nc):
        id = self.data.get_id(nc)
        value = self.data.get_value(nc)
        if value == UNKNOWN_INT:
            value = self.gettext("unknown")

        out = []
        out.extend(self.render_label(context, nc, id))
        out.extend(self.render_annotation_links(nc))
        out.append('<br>')

        out.append(tag("input", name=id, id=id, value=value, type="text",
                       placeholder=nc.getPlaceholder(),
                       class_=self.error_css_class(context, nc),
                       disabled=self.read_only, **SET_MODIFICATION_ATTRS))

        out.append('<br>')
        return Item("".join(out))

    def render_bool(self, context, nc):
        id = self.data.get_id(nc)
        value = self.data.get_value(nc)

        out = []
        out.append(tag("input", type="checkbox",
                       class_=" ".join(["checkbox",
                                        self.error_css_class(context, nc)]),
                       id=id, name=id, value="1", disabled=self.read_only,
                       checked=(value == 1), tabindex=self.next_tabindex(),
                       **SET_MODIFICATION_ATTRS))
        if not self.read_only:
            out.append(hidden_bool_tag(id))
        out.extend(self.render_label(context, nc, id))
        out.extend(self.render_annotation_links(nc))
        out.append("<br>")
        if out[-1] != "<br>":
            out.append("<br>")
        return Item("".join(out))

    def render_radio(self, context, nc):
        id = self.data.get_id(nc)
        value = self.data.get_value(nc)

        out = []
        css_class, messages = self.error_info(context, nc)
        if messages and not self.read_only:
            out.append(tag("div", class_=css_class))

        out.extend(self.render_label(context, nc, id))
        out.extend(self.render_annotation_links(nc))
        # br tag with a newline to be more like the old renderer.
        out.append('<br>\n')

        for child in nc.getChildren():
            if child.isInvisible():
                continue
            child_value = child.getValue()

            # the child name is usually only unique within the radio
            # node, so prefix the id used in the HTML form with the
            # radio instance id so that it hopefully is unique on the
            # page.
            radio_id = id + "." + child.getName()

            out.append(tag("input", type="radio", id=radio_id, name=id,
                           value=child_value, disabled=self.read_only,
                           checked=(child_value == value),
                           tabindex=self.next_tabindex(),
                           class_=self.error_css_class(context, nc),
                           **SET_MODIFICATION_ATTRS))
            out.append(" ")
            out.append(tag("label", for_=radio_id))
            out.append(tag("a", name="f_" + radio_id))
            out.append(escape(child.getDescription()))
            out.append("</a>")
            out.append("</label>")
            if nc.getLayout() != "horizontal":
                out.append('<br>\n')
        if nc.getLayout() == "horizontal":
            out.append('<br>\n')
        out.append('</div>')
        return Item("".join(out))

    def render_radio_matrix(self, context, nc):
        legend = []

        out = []
        out.append('<table class="radiomatrix">')

        # write table header
        children = nc.getChildren()[0].getChildren()
        nr_of_children = len(children) + 1

        out.append('<thead>')
        out.append('<tr>')
        out.append(tag("th", class_="label", colspan=nr_of_children))
        out.extend(self.render_label(context, nc, ""))
        out.extend(self.render_annotation_links(nc))
        out.append('</th>')
        out.append('</tr>')

        out.append('<tr>')
        out.append('<th></th>')
        for hcol in children:
            alternative = hcol.getAlternative()
            description = hcol.getDescription()
            if not alternative:
                header = description
            else:
                header = alternative
                legend.append("<strong>%s</strong>:%s"
                              % (escape(alternative), escape(description)))
            out.append('<th>%s</th>' % escape(header))
        out.append('</tr>')
        out.append('</thead>')

        # write table body
        out.append('<tbody>')
        for num, row in enumerate(nc.getChildren()):
            if isinstance(row, RuleLeaf):
                continue
            self.check_rules(row)
            id = self.data.get_id(row)
            out.append('<tr>')
            out.append(tag("td", class_=("label %s"
                                         % (num % 2 and "hl" or ""))))
            out.extend(self.render_label(context, row, id))
            out.extend(self.render_annotation_links(row))
            out.append('</td>')
            value = self.data.get_value(row)
            for col in row.getChildren():
                col_value = col.getValue()
                out.append(tag("td", class_=(num % 2 and "hl" or "")))
                out.append(tag("input", type="radio", name=id,
                               value=col_value, disabled=self.read_only,
                               checked=(value == col_value),
                               tabindex=self.next_tabindex(),
                               class_=self.error_css_class(context, nc),
                               **SET_MODIFICATION_ATTRS))
                out.append('</td>')
            out.append("</tr>")

        # write legend
        out.append('<tr>')
        out.append(tag("td", colspan=nr_of_children))
        out.append('</td></tr>')
        if legend:
            out.append('<tr>')
            out.append(tag("td", class_="legend", colspan=nr_of_children))
            out.append(" ".join(legend))
            out.append('</td>')
            out.append('</tr>')
        out.append('<tr>')
        out.append(tag("td", class_="label", colspan=nr_of_children))
        out.append('</td>')
        out.append('</tr>')

        out.append('</tbody>')
        out.append('</table>')
        out.append('<br>')

        return Item("".join(out))

    def render_info(self, context, nc):
        if nc.isInvisible():
            return
        out = []
        text = ""
        for item in nc.get_content():
            if isinstance(item, ExprLeaf):
                value = self.data.evaluate_expression(item)
                if isinstance(value, basestring):
                    text += value
                elif isinstance(value, datetime.date):
                    text += self.format_date(item, value)
                # todo: check for numeric values and format them
                else:
                    text += str(value)
            elif isinstance(item, basestring):
                text += item
            else:
                log.error("InfoLeaf %r contains not supported child %r",
                          nc, item)
        if not text:
            return Text("")

        out.append(tag("div", class_="info"))
        out.append(tag("p"))
        out.append(tag("img", src=self.icon("formular/info.png"),
                       width="12", height="12"))
        out.append(escape(text))
        out.append('</p>')
        out.append('</div>')
        return Item("".join(out))

    def render_save(self, context, nc):
        if nc.isInvisible():
            return
        if self.read_only:
            return
        id = self.data.get_id(nc)
        out = []
        out.append(tag("a", class_="anchor", id=id))
        out.extend(
            self.render_submit_buttons(show_next=False, position=id,
                                       from_tag=True))
        out.append('</a>')
        return Item("".join(out))

    def render_group(self, context, nc):
        description = nc.getDescription()

        out = []
        if nc.isInvisible():
            if description:
                out.append("<!-- %s -->\n" % escape(description))
        else:
            out.append("<fieldset>\n")
            if description:
                out.append("<legend>%s %s</legend>\n"
                           % ("".join(self.render_label(context, nc,
                                                        self.data.get_id(nc))),
                              "".join(self.render_annotation_links(nc))))
                if nc.isFolding():
                    out.append(folding_helper_functions())
                    out.append("<div id='%s' style='display:block'>\n" \
                                % nc.getName() )
                    self.init_folding.append('%s' % nc.getName())
                else:
                    out.append("<div>\n")
            out.append(self.render_children(RendererContext(nc, Container), nc))
            out.append("</div></fieldset>\n")
        return Item("".join(out))

    def render_conditional(self, context, nc):
        return self.do_render_conditional(ConditionalContext(context), nc,
                                          self.render_children)

    def do_render_conditional(self, context, nc, render_children, *args):
        result = None
        editable = self.data.evaluate_conditional(nc)
        if editable or not nc.isInvisible():
            old_read_only = self.read_only
            try:
                self.read_only = self.read_only or not editable
                # Uncomment when conditional description should actually
                # be rendered by default in mpuls applications. See
                # issue1859
                #
                # if not editable:
                #     self.render_conditional_description(context, nc)
                result = render_children(context, nc, *args)
            finally:
                self.read_only = old_read_only
        if result is not None:
            result = Text(result)
        return result

    def render_conditional_description(self, context, nc):
        descr = nc.getDescription()
        if descr:
            out = []
            out.append(tag("div", class_="cond_desc"))
            out.append(escape(descr))
            out.append('</div>')
            context.add(nc.getTarget(), Text("".join(out)))

    def render_reference(self, context, nc):
        return self.render_nc(context, self.data.get_referenced_meta_node(nc))

    def render_inline_repeat_group(self, context, nc):
        description = nc.getDescription()
        out = []
        if not nc.isInsideFolding():
            out.append('<fieldset>')
            out.append('<legend>%s %s</legend>'
                   % (escape(description),
                      "".join(self.render_annotation_links(nc))))
        out.append('<table class="repeatgroup">')
        id = self.data.get_id(nc)
        out.append(tag("a", class_=self.error_css_class(context, nc),
                       name="%s" % id))
        out.extend(self.render_repeat_list(context, nc))
        out.append('</table>')
        if not nc.isInsideFolding():
            out.append('</fieldset>')
        return Item("".join(out))

    def render_repeat_list(self, context, nc):
        show_expr = nc.getShowControls()
        min_children = nc.getMin()
        max_children = nc.getMax()
        child_rgs = self.data.get_ordered_repeat_group(nc)
        len_child_rgs = len(child_rgs)

        show_delete_button = len_child_rgs > min_children
        show_add_button = max_children is None or len_child_rgs < max_children
        if show_expr:
            var_provider = self.data.get_variable_provider()
            show_controls = show_expr.evaluate(var_provider)
            show_delete_button = show_delete_button and show_controls & 2
            show_add_button = show_add_button and show_controls & 1

        out = []
        if len_child_rgs > 0:
            for rg in child_rgs:
                self.data.push_node(rg)
                try:
                    out.append('<tr>')
                    out.append('<td class="container">')
                    out.append(self.render_children(
                        RendererContext(nc, Container), nc))
                    out.append('</td>')

                    out.append('<td class="vcontrol" width="17">')
                    if not self.read_only and show_delete_button:
                        out.append(
                            self.render_inline_rg_delete_button(
                                self.data.get_rg_id(rg)))
                    out.append('</td>')
                    out.append('</tr>')
                finally:
                    self.data.pop_node()
        else:
            out.extend(self.render_empty_repeat_list())

        if not self.read_only and show_add_button:
            out.append('<tr>')
            out.append('<td class="hcontrol" style="text-align: right;"'
                       ' colspan="2">')
            out.extend(self.render_inline_rg_add_rg(self.data.get_id(nc)))
            out.append(
                self.render_inline_rg_create_button(self.data.get_id(nc)))
            out.append("</td>")
            out.append("</tr>")
        elif self.read_only and len_child_rgs == 0:
            out.extend(self.render_add_rg_info_text())
        else:
            out.append('<tr><td class="hcontrol" style="text-align: right;"'
                       ' colspan="2">&nbsp;</td></tr>')
        return out

    def render_empty_repeat_list(self):
        out = []
        return out

    def render_add_rg_info_text(self):
        out = []
        out.append('<tr><td class="hcontrol" style="text-align: right;"'
                   ' colspan="2">%s</td></tr>'
                   % self.gettext("To add a RG, edit the document"))
        return out

    def render_inline_rg_delete_button(self, rg_id):
        """Return HTML-fragment for the delete button for inline repeat groups.
        """
        return tag("input", name="del_rg|%s" % rg_id,
                   type="submit", class_="pseudo_imagebutton del_rg",
                   value="",
                   style=('background-image:url("%s")'
                          % self.icon("delete_active_16.png")),
                   title=self.gettext('Delete RG'))

    def render_inline_rg_create_button(self, rg_id):
        """Return HTML-fragment for the create button for inline repeat groups.
        """
        return tag("input", name="add_rg|%s" % rg_id,
                   type="submit", class_="pseudo_imagebutton add_rg",
                   value="",
                   style=('background-image:url("%s")'
                          % self.icon("new_16.png")),
                   title=self.gettext('Add new RG'))

    def render_inline_rg_add_rg(self, rg_id):
        """Return text to add a new rg"""
        out = []
        out.append('<span class="add_rg">')
        out.append(tag("input", name="add_rg|%s" % rg_id,
                   type="submit", class_="add_rg",
                   style="background-color: transparent; border-width: 0;",
                   value=self.get_inline_rg_add_text()))
        out.append("</span>")
        return out

    def get_inline_rg_add_text(self):
        return self.gettext('Add new RG')

    def get_tooltip_height(self, context, nc):
        css_class, messages = self.error_info(context, nc)

        for msg in messages:
            if self.error_css_class(context, nc) == "warning tooltip":
                if len(msg) > 26:
                    tlheight = (int(len(msg) / 26) + 1) * 26 * 1.2
                else:
                    tlheight = 50
            elif self.error_css_class(context, nc) == "error tooltip":
                if len(msg) > 65:
                    tlheight = (int(len(msg) / 57) + 1) * 57 * 1.2
                else:
                    tlheight = 50
            return tlheight

    def render_tooltip(self, context, nc, out):
        height = self.get_tooltip_height(context, nc)
        if self.error_css_class(context, nc) == "warning tooltip":
            out.append(tag("span", class_="warningtip",
                           style="height:" + str(height) + "px"))
            out.append(tag("img", src=self.icon("infotool.png"),
                           alt="Warnung",
                           width="48", height="48"))
            out.append("<em>")
            out.append("Hinweis!")
            out.append("</em>")
            out.extend(self.render_errors(context, nc))
            out.append("</span>")
        elif self.error_css_class(context, nc) == "error tooltip":
            out.append(tag("span", class_="errortip",
                           style="height:" + str(height) + "px"))
            out.append(tag("img", src=self.icon("alerttool.png"),
                           alt="Warnung",
                           width="48", height="48"))
            out.append("<em>")
            out.append("Warnung!")
            out.append("</em>")
            out.extend(self.render_errors(context, nc))
            out.append("</span>")

    def render_label(self, context, nc, id):
        out = []
        out.append(tag("label", for_=id))
        if isinstance(nc, GroupNode) and nc.isFolding():
            out.append(tag("a",
                       class_="folded " + self.error_css_class(context, nc),
                       name="f_%s" % id,
                       id="label_%s" % nc.getName(),
                       onClick="toggle_folding('%s');" % nc.getName()))
        else:
            out.append(tag("a", class_=self.error_css_class(context, nc),
                       name="f_%s" % id))
        out.append(escape(nc.getDescription()))
        self.render_tooltip(context, nc, out)
        out.append("</a>")
        out.append("</label>")
        return out

    def render_file(self, context, nc):
        value = self.data.get_value(nc)
        id = self.data.get_id(nc)

        out = [tag("div", class_="file")]

        out.extend(self.render_label(context, nc, id))
        out.extend(self.render_annotation_links(nc))
        out.extend("<br/>")

        if value:
            out.append(tag("span", class_="attachment"))
            out.append(escape(value))
            out.append("</span>")
        else:
            out.append(tag("input", name=id, id=id,
                           class_=self.error_css_class(context, nc),
                           type="file", disabled=self.read_only,
                           **SET_MODIFICATION_ATTRS))
        return Item("".join(out))

    def render_errors(self, context, nc):
        """Render error hints for the item indicated by nc.
        The return value is a sequence of strings with an error or
        warning message if there are any.
        """
        css_class, messages = self.error_info(context, nc)

        out = []
        if messages:
            out.extend([", ".join(escape(msg) for msg in messages)
                        ])
        return out

    def error_info(self, context, nc):
        """Return error information for the item indicated by nc.

        The return value should be a tuple (CSS-CLASS, MESSAGES) where
        CSS-CLASS is a string with the css class to use when rendering
        the span with the messages. Typical classes are 'error' and
        'warning'. It should normally be the same as the class returned
        by the error_css_class method. MESSAGES is a list if strings
        containing the messages. If there are no errors, the MESSAGES
        list should be empty.
        """
        return ("error", self.get_errors(nc))

    def error_css_class(self, context, nc):
        if self.has_errors(nc):
            return "error tooltip"
        elif self.has_warnings(nc):
            return "warning tooltip"
        return ""

    def check_warnings(self, nc):
        """
        This method is deprecated. Please use check_rules instead.
        """
        self.check_rules(nc)

    def check_rules(self, nc):
        """Check node for warnings, notices and errors by evaluating the
        relevant rules.

        Found warnings, notices and rules are stored in the corresponding member
        dict as a list for the requested node.
        """
        full_name = self.data.full_item_name(nc)

        # add all checked widgets to warning_order, even if they have no
        # warnings or no warning rules so that the widget_order list can
        # be used to order the errors too.
        self.widget_order.append(full_name)

        warnings, notices, errors = self.data.evaluate_rules(nc)

        self.warnings[full_name].extend(rule.getValue() for rule in warnings)
        self.notices[full_name].extend(rule.getValue() for rule in notices)
        self.errors[full_name].extend(rule.getValue() for rule in errors)

    def has_warnings(self, nc):
        return bool(self.warnings.get(self.data.full_item_name(nc)))

    def get_warnings(self, nc):
        return self.warnings.get(self.data.full_item_name(nc), ())

    def has_errors(self, nc):
        """
        Returns True if the meta node has an error

        An error is present if either a rule with a error mark is invalid or the
        provided data has an error caused by a rule violation.
        """
        full_name = self.data.full_item_name(nc)
        return bool(self.errors[full_name]) or self.data.has_errors(nc)

    def get_errors(self, nc):
        """
        Returns a list of error messages for the meta node nc
        """
        messages = self.data.get_errors(nc)
        errors = self.errors.get(self.data.full_item_name(nc), [])
        return list(messages) + list(errors)


class RepeatGroupRenderer(BaseRenderer):

    """Renderer for repeat group digests

    Repeat groups that contain entire form pages have a digest page that
    shows an overview of the existing instances. This overview is a list
    with one entry for each instance and buttons to delete or create new
    instances. The entries for the individual repeat group show a subset
    of the fields of the repeat group.

    Which fields are shown in an item for a repeat group and in which
    order the repeat groups are shown is determined by the digest and
    orderby attributes of the meta node of the repeat group.

    The parameter digests is not used and is only there for backwards
    compatibility reasons. There is still code that passes it to the
    constructor, although no mpuls application actually uses it anymore.
    """

    def __init__(self, data, view_root, url_factory, read_only=True,
                 digests=None, languages=None):
        super(RepeatGroupRenderer, self).__init__(data, view_root, url_factory,
                                                  read_only=read_only,
                                                  languages=languages)
        self.warnings = {}
        self.notices = {}
        self.errors = {}
        self.widget_order = []

        # the root of the repeat group renderer must be a repeat node
        assert isinstance(self.view_root, RepeatNode)

    def render(self, show_next=False, hidden_fields=()):
        out = []
        out.append('<div class="widget">')
        out.append('<table class="list" style="margin:0px;">')
        out.append('<tr><th class="table_width_small"></th>')
        out.append('<th colspan>%s</th>' % escape(self.gettext('Description')))
        out.append('<th class="actions">%s</th>'
                   % escape(self.gettext('Actions')))
        out.append("</tr>")

        # view_root is the RepeatNode and the first child must be the
        # page that's repeated.
        page = self.view_root.getChildren()[0]
        assert isinstance(page, PageNode)

        show_controls = 3  # show both buttons
        show_expr = self.view_root.getShowControls()
        if show_expr:
            var_provider = self.data.get_variable_provider()
            show_controls = show_expr.evaluate(var_provider)

        fields = self.view_root.getDigest()
        instances = self.data.get_ordered_repeat_group(self.view_root)
        for num, rg in enumerate(instances):
            out.append(tag("tr", class_=(num % 2 and "hl" or "")))
            out.append('<td align="center">')
            out.append('<strong>')
            out.append('(%s)' % (num + 1))
            out.append('</strong>')
            out.append('</td>')
            out.append('<td>')
            self.data.push_node(rg)
            try:
                select_url = self.url(controller="navigation",
                                      action="select_branch",
                                      id=self.data.get_id(page))
                out.append(self.render_digest(rg, fields, select_url))
            finally:
                self.data.pop_node()
            out.append('</td>')
            out.append('<td class="actions hcontrol">')
            out.append(tag("a", href=select_url))
            out.append(tag("img", src=self.icon("open_active_22.png"),
                           border="0", alt=self.gettext('view'),
                           title=self.gettext('view'),
                           width="22", height="22"))
            out.append("</a>")
            if not self.read_only and show_controls & 2:
                out.append(tag("a",
                               href=self.url(controller="repeatgroup",
                                             action="delete",
                                             id=self.data.get_rg_id(rg))))
                out.append(tag("img", src=self.icon("delete_active_22.png"),
                               border="0", alt=self.gettext('delete'),
                               title=self.gettext('delete'),
                               width="22", height="22"))
                out.append("</a>")
            out.append('</td>')
            out.append('</tr>')

        out.append('<tr class="background_slight">')
        out.append('<td></td>')
        out.append('<td style="text-align:right;">')
        if not self.read_only and show_controls & 1:
            out.append(tag("a",
                           href=self.url(controller="repeatgroup",
                                         action="new",
                                         id=self.data.get_id(self.view_root))))
            out.append(self.gettext('Add new RG'))
            out.append('</a>')
        out.append('</td>')
        out.append('<td class="actions">')
        if not self.read_only and show_controls & 1:
            out.append(tag("a",
                           href=self.url(controller="repeatgroup",
                                         action="new",
                                         id=self.data.get_id(self.view_root))))

            out.append(tag("img", src=self.icon("new_22.png"),
                           border="0", alt=self.gettext('Add new RG'),
                           title=self.gettext('Add new RG'),
                           width="22", height="22"))
            out.append("</a>")
        out.append('</td>')
        out.append('</tr>')
        if not instances and self.read_only:
            out.extend(self.render_add_rg_info_text())
        out.append('</table>')
        out.append('</div>')
        out.append('<div>%s</div>'
                   % (escape(self.ungettext('%d Item found', '%d Items found',
                                            len(instances))
                             % len(instances))))

        return "\n".join(out)

    def render_add_rg_info_text(self):
        out = []
        out.append('<tr class="background_slight">')
        out.append('<td/>')
        out.append('<td style="text-align:right;">')
        out.append(escape(self.gettext("To add a RG, edit the document")))
        out.append('</td>')
        out.append('<td/>')
        return out

    def render_digest(self, rg, fields, select_url):
        out = []
        out.append('<table class="rgdigest">')
        for field in fields:
            description, value = self.get_field_info(field)
            if description is None:
                continue
            out.append('<tr>')
            out.append('<td class="label">')
            out.append(tag("a", href=select_url))
            out.append(escape(description))
            out.append('</a>')
            out.append('</td>')
            out.append('<td>')
            out.append(tag("a", href=select_url))
            out.append('%s' % escape(NA(value)))
            out.append('</a>')
            out.append('</td>')
            out.append('</tr>')
        out.append('</table>')
        return "".join(out)

    def get_field_info(self, field):
        """Return description and formatted value for the given digest field.
        If the item referred to by the field would be invisible in the
        rendered repeat group instance due to visibility conditionals,
        the method returns (None, None) to indicate that it shouldn't be
        rendered in the digest either.
        """
        nc = self.data.get_meta_node(field)

        # check whether the item would be visible according to the
        # conditionals containing it.
        for conditional in conditional_parents(nc):
            if (conditional.isInvisible()
                    and not self.data.evaluate_conditional(conditional)):
                return None, None

        return (nc.getDescription(), self.format_nc_value(nc))

    def format_nc_value(self, nc):
        """Return the formatted value for the meta node nc.
        If nc is a ReferenceNode, the value is that of the referenced
        field.
        """
        if isinstance(nc, ReferenceNode):
            nc = self.data.get_referenced_meta_node(nc)
        value = self.data.get_value(nc)
        if isinstance(nc, DateLeaf):
            value = self.format_date(nc, value)
        elif isinstance(nc, NumericLeaf):
            value = self.format_numeric(nc, value)
        if isinstance(nc, ChoiceNode):
            value = nc.getValueMapping().get(value, "?")
        return value


class ErrorRenderer(BaseRenderer):

    def __init__(self, data, meta_tree, errors, warnings, notices, widget_order,
                 url_factory, languages=None):
        super(ErrorRenderer, self).__init__(data, None, url_factory,
                                            languages=languages)
        self.meta_tree = meta_tree
        self.errors = errors
        self.warnings = warnings
        self.notices = notices
        self.widget_order = widget_order

    def url(self, controller, action, id=None):
        return self.url_factory.url(controller, action, id=id)

    def icon(self, name):
        return self.url_factory.icon(name)

    def in_widget_order(self, items):
        """Generate the items in the dictionary items in widget order.
        The dictionary should map widget ids to errors or warnings.
        The widget order is defined by the widget_order parameter given
        to the constructor when the ErrorRenderer was instantiated.
        """
        copy = items.copy()
        for id in self.widget_order:
            value = copy.pop(id, None)
            if value is not None:
                yield (id, value)
        if copy:
            # This shouldn't happen, but if it does happen, log a debug
            # message and yield the rest of the items in an arbitrary
            # order, so that all are at least shown to the user.
            log.debug("ErrorRenderer: not all items occur in the widget_order"
                      " list")
            for key, value in copy.iteritems():
                yield (key, value)

    def render(self):
        out = []
        out.extend(self.render_items(self.errors,
                                     self.gettext('Formular errors'),
                                     self.icon("info/alert.png")))
        out.extend(self.render_items(self.notices,
                                     self.gettext('Formular notices'),
                                     self.icon("info/info.png"),
                                     css_class='warning-box',
                                     show_undo=False))
        out.extend(self.render_items(self.warnings,
                                     self.gettext('Formular warnings'),
                                     self.icon("info/info.png"),
                                     css_class='warning-box',
                                     show_undo=False))
        return "".join(out)

    def render_items(self, items, title, title_icon, css_class="",
                     show_undo=True, icon_width=48, icon_height=48,
                     item_url_function=None):
        items = dict((key, value) for key, value in items.items() if value)
        if not items:
            return ""

        if item_url_function is None:
            item_url_function = self.error_item_url

        out = []
        out.append(tag("div", class_=("widget dialog %s" % css_class)))
        out.append('<h1>')
        out.append(tag("img", src=title_icon, alt='',
                       width=icon_width, height=icon_height))
        out.append(escape(title))
        out.append('</h1>')
        out.append('<ol>')

        for k, values in self.in_widget_order(items):
            element = self.data.get_meta_node_from_item_name(k)
            description = element.getDescription()
            # the values in the items dict may either be strings or
            # lists of strings.
            if not isinstance(values, list):
                values = [values]
            for v in values:
                out.append('<li>')
                if show_undo:
                    out.append(tag("a",
                                   href=self.url(controller="formularpage",
                                                 action="undo", id=k)))
                    out.append(tag("img", src=self.icon("formular/undo.png"),
                                   border="0", alt="undo",
                                   title=self.gettext('Restore last'
                                                      ' valid value'),
                                   _close=True))
                    out.append('</a>')
                out.append(tag("a", class_="error", href=item_url_function(k)))
                out.append("%s:</a>" % escape(description))
                out.append(" %s" % escape(v))
                out.append("</li>")

        out.append('</ol>')
        out.append('</div>')
        return out

    def error_item_url(self, full_name):
        """Return the URL for a link to the form element given by full_name"""
        return ("#f_%s" % self.data.form_name_from_full_name(full_name))
