#!/bin/sh

set -e

usage()
{
cat << EOF
usage: $0 options source

This script will install the distribution under
/home/mpuls/releases/<name>-<rev>. It will further setup a virtual
environment. Database setup is not done!

OPTIONS:
   -h       Show this message
   -n       Name of the release (Defaults to name of archived source-revision)
   -v       Setup a virtual environment for python and install the packages in
            it (Default: No).
   -a       Make the installed package the active package by setting the
            "current" sybolic link (Default: No).
   -p DIR   Change ownership of the sub-directory DIR to postgres and give
            postgres write permissions there. Default: db-admin.
            This will only be done if DIR is not the empty string and
            this script is run by root.
EOF
}

# DEFAULTS
VERBOSE=0
INSTALL_BASE_PATH=/home/mpuls

NAME=
BUILDDIR=/tmp
COMPRESS=0
ACTIVATE=0
VIRTUALPYTHON=0
POSTGRESDIR=db-admin

# Parse Options
while getopts "hvan:r:p:" OPTION; do
    case $OPTION in
        h)
            usage
            exit 1
            ;;
        a)
            ACTIVATE=1
            ;;
        n)
            NAME=$OPTARG
            ;;
        v)
            VIRTUALPYTHON=1
            ;;
        p)
            POSTGRESDIR=$OPTARG
            ;;
        ?)
            # Will be set if the optstring begins with a ":". In this case
            # all options not defined in the optstring will OPTION set to
            # "?". Otherwise the shell reports an error on the unknown
            # option.
            usage
            exit
            ;;
    esac
done

if ! id -u mpuls >/dev/null 2>&1; then
    # Create mpuls use if run by root
    if [ "`id -u`" -eq 0 ]; then
        echo "Adding system user mpuls using /bin/bash as shell";
        adduser --system --home $INSTALL_BASE_PATH mpuls --shell /bin/bash
    else
        echo "Error! mpuls user doesn't exist!";
        exit 1;
    fi
fi


# Check if installation base dir exists
if [ ! -d $INSTALL_BASE_PATH ]; then
    echo "Error! Installation directory ${INSTALL_PATH} does not exist!";
    exit 1;
fi


if [ -z $NAME ]; then
    echo "Error! Please provide name of the distribution!";
    usage;
    exit 1;
fi

DISTRIBUTION=`readlink -f ../`
DISTRIBUTION_BASENAME=`basename $DISTRIBUTION`

# Try to figure out names based on package name
DISTRIBUTION_NAME=${NAME}

INSTALL_APP_PATH=$INSTALL_BASE_PATH/$DISTRIBUTION_NAME

# Check if installation distribution dir exists
if [ ! -d $INSTALL_APP_PATH ]; then
    echo "Cannot find installation directory ${INSTALL_APP_PATH}." \
        "Creating it ..."
    mkdir ${INSTALL_APP_PATH}
    (cd ${INSTALL_APP_PATH} && mkdir releases)
fi

# For convenience, change into the releases dir for the rest of the script.
cd ${INSTALL_APP_PATH}/releases

# Copy all files.
echo "Copying file from ${DISTRIBUTION} to ${INSTALL_APP_PATH}/releases ...";
rm -rf ${DISTRIBUTION_BASENAME}
cp -r ${DISTRIBUTION} .

if [ $ACTIVATE = 1 ]; then
    rm -f current
    ln -s ${DISTRIBUTION_BASENAME} current
else
    echo "Not activating the installation / setting the 'current' symlink.";
fi

# Set permissions if run by root
if [ "`id -u`" -eq 0 ]; then
    echo "Setting permissions ...";
    chown mpuls: -R ${DISTRIBUTION_BASENAME}

    if [ -n "$POSTGRESDIR" ] && id -g postgres >/dev/null 2>&1; then
        echo "Changing group permission of ${DISTRIBUTION_BASENAME}/$POSTGRESDIR to postgres"
        chgrp postgres -R ${DISTRIBUTION_BASENAME}/$POSTGRESDIR
        chmod g+w -R ${DISTRIBUTION_BASENAME}/$POSTGRESDIR
    fi
fi


setup_egg_info () {
    dir="${DISTRIBUTION_BASENAME}/$1"
    if [ -d "$dir" ]; then
        (cd "$dir" && python setup.py egg_info)
    fi
}

# Setup environment
if [ $VIRTUALPYTHON = 1 ]; then
    echo "Setting up virtual environment ...";
    (
        cd ${DISTRIBUTION_BASENAME}/maintenance
        sh setup-environment.sh -n ${DISTRIBUTION_NAME} -g -d .. .. 1>/dev/null
    )
else
    echo "Not setting up a virtual environment";
    setup_egg_info application
    setup_egg_info base
    setup_egg_info libformed
    setup_egg_info libmpuls
fi

echo Install finished in ${INSTALL_APP_PATH}/releases/${DISTRIBUTION_BASENAME}!
